/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.ars.ars.mgmt.commands.ArSCmdInterface;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.mgmt.simulation.SimulationManager;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.scheduler.TaskScheduler;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.ars.ars.persistence.ContainerDataIterator;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.DAVAppBase;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import java.io.File;
import java.lang.ref.WeakReference;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Zentrale Archiv-Verwaltung.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveManager extends DAVAppBase {

	@SuppressWarnings("ALL")
	private static long SYS_TIME_OFFSET = 0;            // nur für testzwecke

	@SuppressWarnings("ALL")
	public static long DATAINDEX_OFFSET = 0;            // nur für testzwecke

	/**
	 * Nur für Testzwecke; falls wahr, werden nur Datensätze mit Daten archiviert. Es gibt dann keine Datensätze folgender Art: "keine Quelle", "keine Daten",
	 * "pot. Datenlücke". Manche Testverfahren werden einfacher, wenn ausschliesslich Datensätze archiviert werden, die auch gesendet wurden.
	 */
	@SuppressWarnings("ALL")
	public static boolean ONLY_DATA = false;

	private static final int DEFAULT_CMD_PORT = 4242;

	private static final String P_PRINT_PARAMS = "-nurParametrierungAusgeben";

	private static final String P_PRINT_PARAMS_FILE = "-parametrierungAusgabeDatei";

	private static final String P_PERS_PATH = "-persistenzVerzeichnis";

	private static final String P_CHECK_PERS_DIR = "-persistenzVerzeichnisPruefen";

	private static final String P_REMOTE_CONTROL = "-kommandoPort";

	private static final String P_BACKUP = "-backup";

	private static final String P_MAX_QUERIES_PER_APPLICATION = "-maxAnfragenProApplikation";

	private static final int DEFAULT_MAX_QUERIES_PER_APPLICATION = 5;

	public static final String P_CHECK_PERS_DIR_DIR = "Verzeichnis";

	public static final String P_CHECK_PERS_DIR_HDR = "Header";

	public static final String P_CHECK_PERS_DIR_ALL = "Alles";

	public static final String P_REMOTE_CONTROL_NORC = "deaktiviert";

	public static final String P_REMOTE_CONTROL_DEFAULT = "default";

	private PersistenceManager persistenceMgr;

	private final DataIdentTree dataIdentTree = new DataIdentTree();

	private InQueuesMgr inQueuesMgr;

	private ArchivConfig archivConfig;

	private TaskScheduler taskScheduler;

	private SimulationManager simulationMgr;

	// Aufrufparameter:
	private boolean printParamsAndExit;

	private File printParamsFile;

	private String checkPersDirMode;
	private Path archivePath;

	private int maximumQueriesPerApplication;

	private int cmdPortNum = -1;

	private boolean terminated;

	private boolean runningProperly;

	private final Object _terminatedLock = new Object();

	/** Enthält die Anzahl aktiver Archivanfragen je Applikation. Es sind nur für die Applikationen Einträge in der Map enthalten für die mindestens eine aktive Archivanfrage in der Verarbeitung ist. */
	private final TreeMap<SystemObject, Integer> _applicationToArchiveQueryCount;

	/** Thread für Shutdown-Hook */
	private Thread _shutDownHookThread;

	/**
	 * Anzahl Online-Archivqueues/Buffer
	 */
	private int _numOfArchQueuesOnline = -1;

	/**
	 * Anzahl Threads fürs Index-Schließen beim Beenden
	 */
	private int _numCloseIndexThreads = -1;

	/**
	 * Anzahl Threads für Persistenz-Prüfung und Wiederherstellung
	 */
	private int _numCheckPersistenceThreads = -1;

	/**
	 * Anzahl Threads fürs Löschen
	 */
	private int _numDeleteThreads = -1;

	/**
	 * Anzahl Threads fürs Backup-Indizieren
	 */
	private int _numBackupThreads = -1;

	/**
	 * Minimum Nanosekunden-Wartezeit zwischen Anmeldungen
	 */
	private long _minWaitNanosPerSubscription;

	/**
	 * Maximum Nanosekunden-Wartezeit zwischen Anmeldungen
	 */
	private long _maxWaitNanosPerSubscription;

	/**
	 * Anzahl gleichzeitiger Anmeldungen
	 */
	private int _slidingWindowSize;

	/**
	 * Gesamtkapazität Datensätze in Online-Queues
	 */
	private int _totalCapacityOfOnlineQueues = -1;

	/**
	 * Gesamtkapazität Datensätze in Nachgefordert-Queues
	 */
	private int _totalCapacityOfRequestedQueues = -1;

	/**
	 * System.exit() überspringen (für Unit-Tests)
	 */
	private boolean _skipSystemExit;

	/**
	 * Maximale Anzahl geladener Indexe
	 */
	private long _indexCacheMaxCount;

	/**
	 * Maximale Anzahl Bytes pro Index als Cache
	 */
	private int _indexCacheMaxSize;
	private boolean _useDynamicDeletionTime;

	/**
	 * Pid des Objekt für das Archivsystem (vom Typ typ.archiv, in der Regel die lokale AOE)
	 */
	private String _archiveObjectPid;

	public ArchiveManager(String[] args) throws Exception {
		// Initialer DisconnectHandler wird später in der run()-Methode ersetzt
		super(args, "Archivsystem", () -> {
			Debug.getLogger().error("Verbindung zum Datenverteiler wurde frühzeitig terminiert, Archivsystem wird beendet");
			System.exit(2);
		});
		try {
			// Abhängigkeiten prüfen, damit der Anwendet sofort merkt, dass etwas falsch läuft wenn er eine zu alte Kernsoftware verwendet.
			Class.forName("de.bsvrz.dav.daf.util.FileAccess");
		}
		catch(ClassNotFoundException ignored){
			Debug.getLogger().error("Das ArchivSystem benötigt die Kernsoftware in Mindestversion 3.9.6");
			System.exit(3);
		}
		Thread.setDefaultUncaughtExceptionHandler(new ExceptionHandler());
		_applicationToArchiveQueryCount = new TreeMap<>();
		logger = Debug.getLogger();
		parseArguments(argList);
		if(!printParamsAndExit) persistenceMgr = new PersistenceManager(this, archivePath);
		archivConfig = new ArchivConfig(this);
	}

	/** 
	 * Gibt das Archiv-Objekt vom Typ typ.archiv zurück
	 * @return das Archiv-Objekt
	 */
	public final ConfigurationObject getArchiveObject() {
		ConfigurationObject object = null;
		if(_archiveObjectPid.isEmpty()) {
			 object = getConfigAuth();
		}
		else {
			DataModel dataModel = getDataModel();
			if(dataModel != null) {
				SystemObject systemObject = dataModel.getObject(_archiveObjectPid);
				if(systemObject instanceof ConfigurationObject) {
					object = (ConfigurationObject) systemObject;
				}
				else {
					throw new IllegalArgumentException("Das Archiv-Objekt vom typ.archiv konnte nicht bestimmt werden. Aufrufargument -archiv=" + _archiveObjectPid + " prüfen.");
				}
			}
		}
		if(object == null) {
			throw new IllegalArgumentException("Das Archiv-Objekt vom typ.archiv konnte nicht bestimmt werden. Eventuell besteht noch keine Datenverteiler-Verbindung.");
		}
		return object;
	}

	@Nullable
	@Override
	public DataModel getDataModel() {
		ClientDavInterface connection = getDavCon();
		if(connection == null) return null;
		return connection.getDataModel();
	}

	public boolean wasTerminated() {
		synchronized(_terminatedLock) {
			return terminated;
		}
	}

	/** @return aktuelle Systemzeit (die zu Testzwecken mit einem Offset versehen werden kann) */
	public static long archiveTime() {
		return System.currentTimeMillis() + SYS_TIME_OFFSET;
	}

	/**
	 * Verstellt die Systemzeit um offset Sekunden. NUR FÜR TESTZWECKE!!!
	 *
	 * @param offsetSeconds Spanne in Sekunden
	 */
	public static void setSystemTimeOffset(long offsetSeconds) {
		SYS_TIME_OFFSET = offsetSeconds * 1000;
	}


	private void parseArguments(ArgumentList argLst) {
		// Kommandozeilenparameter:
		CacheManager.getInstance().init(argList);
		printParamsAndExit = argLst.hasArgument(P_PRINT_PARAMS);
		printParamsFile = argLst.hasArgument(P_PRINT_PARAMS_FILE) ? argLst.fetchArgument(P_PRINT_PARAMS_FILE).asFile() : null;
		checkPersDirMode = argLst.hasArgument(P_CHECK_PERS_DIR) ? argLst.fetchArgument(P_CHECK_PERS_DIR).asString() : "";
		archivePath = argLst.hasArgument(P_PERS_PATH) ? Paths.get(argLst.fetchArgument(P_PERS_PATH).asString()) : null;
		String cmdPort = argLst.hasArgument(P_REMOTE_CONTROL) ? argLst.fetchArgument(P_REMOTE_CONTROL).asString() : null;
		final String backupProps = argLst.hasArgument(P_BACKUP) ? argLst.fetchArgument(P_BACKUP).asString() : null;
		maximumQueriesPerApplication =  argLst.hasArgument(P_MAX_QUERIES_PER_APPLICATION) ? argLst.fetchArgument(P_MAX_QUERIES_PER_APPLICATION).intValueBetween(1, Integer.MAX_VALUE): DEFAULT_MAX_QUERIES_PER_APPLICATION;

		// Archiv-Objekt setzen
		_archiveObjectPid = argLst.fetchArgument("-archiv=").asString();
		
		// Parameter für Archivqueues
		_numOfArchQueuesOnline = argLst.fetchArgument("-archivierungsThreads=11").intValueBetween(1, Integer.MAX_VALUE);
		
		// Der folgende Prameter ist nicht mehr benutzt, wird aber dennoch abgefragt damit es bei Updates zu keinen 
		// Problemen durch "unbekannte" Parameter kommt
		argLst.fetchArgument("-nachforderungsThreads=3").intValueBetween(1, Integer.MAX_VALUE);
		
		_totalCapacityOfOnlineQueues = argLst.fetchArgument("-archivierungsPuffer=200000").intValueBetween(getNumOfArchQueuesOnline() * 250, Integer.MAX_VALUE);
		_totalCapacityOfRequestedQueues = argLst.fetchArgument("-nachforderungsPuffer=20000").intValueBetween(250, Integer.MAX_VALUE);

		// Parameter für Thread-Anzahl
		_numCloseIndexThreads = argLst.fetchArgument("-indexUpdateThreads=" + getNumOfArchQueuesOnline()).intValueBetween(1, Integer.MAX_VALUE);
		_numCheckPersistenceThreads = argLst.fetchArgument("-pruefThreads=" + getNumOfArchQueuesOnline()).intValueBetween(1, Integer.MAX_VALUE);
		_numDeleteThreads = argLst.fetchArgument("-lzvThreads=1").intValueBetween(1, Integer.MAX_VALUE);
		_numBackupThreads = argLst.fetchArgument("-sicherungsThreads=" + getNumOfArchQueuesOnline()).intValueBetween(1, Integer.MAX_VALUE);

		// Parameter für Anmeldeverzögerung
		double minPerSecond = argLst.fetchArgument("-minAnmeldungenProSekunde=1").doubleValueBetween(1, Double.POSITIVE_INFINITY);
		double maxPerSecond = argLst.fetchArgument("-maxAnmeldungenProSekunde=Infinity").doubleValueBetween(1, Double.POSITIVE_INFINITY);
		_slidingWindowSize = argLst.fetchArgument("-maxAnmeldungenGleichzeitig=1000").intValueBetween(1, 1_000_000);

		// Begrenzung der offenen Containerdateien
		ContainerDataIterator.setOpenFileLimit(argLst.fetchArgument("-maxOffeneDateien=500").intValueBetween(1, Integer.MAX_VALUE));

		PersistenceManager.setDeleteBrokenContainers(argLst.fetchArgument("-defekteDateienVerschieben=false").booleanValue());
		
		_useDynamicDeletionTime = argLst.fetchArgument("-dynamischVorhalten=false").booleanValue();
		
		_indexCacheMaxCount = argLst.fetchArgument("-maxIndexCacheAnzahl=100").longValueBetween(1, Long.MAX_VALUE);
		_indexCacheMaxSize = argLst.fetchArgument("-maxIndexBytes=16384").intValueBetween(1, Integer.MAX_VALUE);
		
		// Max/Min-Vertauschung absichtlich wegen Kehrwert.
		_minWaitNanosPerSubscription = (long) (1_000_000_000.0 / maxPerSecond);
		_maxWaitNanosPerSubscription = (long) (1_000_000_000.0 / minPerSecond);

		// Konsistenzprüfung Kommandozeilenparameter:
		if(checkPersDirMode.length() > 0 && archivePath == null) quitError("Ungültige Parameterkombination: " + P_CHECK_PERS_DIR + " erfordert " + P_PERS_PATH);

		_skipSystemExit = argLst.fetchArgument("-noExit=false").booleanValue();

		if(checkPersDirMode.length() > 0 && printParamsAndExit) {
			quitError("Ungültige Parameterkombination: " + P_CHECK_PERS_DIR + " und " + P_PRINT_PARAMS + " nicht gleichzeitig möglich");
		}

		if(checkPersDirMode.length() > 0 && !(checkPersDirMode.equalsIgnoreCase(P_CHECK_PERS_DIR_DIR) || checkPersDirMode.equalsIgnoreCase(P_CHECK_PERS_DIR_HDR)
		                                      || checkPersDirMode.equalsIgnoreCase(P_CHECK_PERS_DIR_ALL))) {
			quitError(
					"Ungültiger Parameterwert: Für " + P_CHECK_PERS_DIR + " sind folgende Werte möglich: [" + P_CHECK_PERS_DIR_DIR + ", " + P_CHECK_PERS_DIR_HDR
					+ ", " + P_CHECK_PERS_DIR_ALL + "]"
			);
		}

		if(!printParamsAndExit && printParamsFile != null) {
			quitError("Ungültige Parameterkombination: " + P_PRINT_PARAMS_FILE + " nur in Verbindung mit " + P_PRINT_PARAMS + " möglich");
		}

		if((!printParamsAndExit || checkPersDirMode.length() > 0) && archivePath == null) {
			quitError("Ungültige Parameterkombination: " + P_PERS_PATH + " nicht vorhanden");
		}

		if(cmdPort != null && !cmdPort.equalsIgnoreCase(P_REMOTE_CONTROL_NORC)) {
			if(cmdPort.equals(P_REMOTE_CONTROL_DEFAULT)) {
				cmdPortNum = DEFAULT_CMD_PORT;
			}
			else {
				try {
					cmdPortNum = Integer.parseInt(cmdPort);
				}
				catch(NumberFormatException e) {
					quitError("Ungueltiger Wert fuer " + P_REMOTE_CONTROL + ": '" + cmdPort + "'");
				}
			}
		}

		if(!(checkPersDirMode.length() > 0 || printParamsAndExit)) {
			// Wenn nur das PV geprüft werden oder die Parametrierung ausgegeben werden soll, brauchen wir kein Backup-Modul
			try {
				BackupTask.setBackupModul(backupProps);
			}
			catch(Exception e) {
				quitError(e.getMessage());
			}
		}
	}

	public void run() throws Exception {
		if(wasTerminated()) {
			logger.error("Das Archivsystem wurde bereits beendet.");
			throw new IllegalStateException("Das Archivsystem wurde bereits beendet.");
		}
		if(executeSpecialModeNoDAV()) {
			quit();
		}
		else {

			if(executeSpecialModeWithDAV())	// startet den DAV
			{
				quit();
			}
			else {
				// ##########################################################################################
				// hier gehts richtig los:
				logger.info("Archivsystem im Normalmodus gestartet");

				// Persistenzverzeichnis überprüfen (evtl anderes ArS angemeldet?)
				String msg = persistenceMgr.createLockFile();
				if(msg != null) {
					quitError(msg);
				}
				else {
					setDisconnectHandler(() -> quitError("Verbindung zum Datenverteiler wurde terminiert, Archivsystem wird kontrolliert beendet"));
					_shutDownHookThread = new Thread(new ShutdownHook(this));
					Runtime.getRuntime().addShutdownHook(_shutDownHookThread);
					// Keine Sperrdatei vorhanden, eine angelegt, wir können starten
					runningProperly = true;

					QueueParameters parameters = new QueueParameters(
							getNumOfArchQueuesOnline(),
							_totalCapacityOfOnlineQueues,
							_totalCapacityOfRequestedQueues,
							_minWaitNanosPerSubscription,
							_maxWaitNanosPerSubscription,
							_slidingWindowSize
					);

					inQueuesMgr = new InQueuesMgr(this, dataIdentTree, parameters);
					taskScheduler = new TaskScheduler(this);
					simulationMgr = new SimulationManager(this);

					// Als erstes das Kommando-Interface starten
					if(cmdPortNum != -1) new ArSCmdInterface(this, cmdPortNum).start();

					// Loesch-/Nachfordern-/Sichern-Uhren starten:
					if(!wasTerminated()) taskScheduler.start();

					// PersistenceManager ist der einzige der einen Startup-Check braucht,
					// alle anderen Module haben keine persistenten Daten.
					if(!persistenceMgr.startupProcedure()) {
						quitError("Fehler im Persistenzverzeichnis. Archivsystem kann nicht gestartet werden.");
					}
					else {
						enableExplicitApplicationReadyMessage();
						connectToDavQuit();	// Bis hierher sollte alles ohne DAV gehen

						if(!wasTerminated()) inQueuesMgr.getObjectsFromDav();	// jeder Task darf benoetigte Objekte holen
						if(!wasTerminated()) inQueuesMgr.startAllTasks();		// startet alle Tasks (mit Ausnahme von ArchivConfig)
						if(!wasTerminated()) inQueuesMgr.subscribeSettings();	// auf alle Parameter anmelden

						// mit der Anmeldung der zu archivierenden Daten gehts los...
						if(!wasTerminated()) archivConfig.startSubscribeArchiveParams(inQueuesMgr, dataIdentTree);

						if(!wasTerminated()) simulationMgr.start();				//Simulationsverwaltung starten
					}
				}
			}
		}
	}

	public void connectToDavQuit() {
		try {
			connectToDav();
		}
		catch(Exception e) {
			quitError("Es konnte keine Verbindung zum Datenverteiler aufgebaut werden.\n" + e.getMessage());
		}
	}

	private boolean executeSpecialModeNoDAV() throws InterruptedException {
		if(checkPersDirMode.length() > 0) {
			logger.info("Archivsystem zur Prüfung des Persistenzverzeichnisses gestartet");
			persistenceMgr.checkPersistenceDir(checkPersDirMode);
			return true;
		}
		return false;
	}

	private boolean executeSpecialModeWithDAV() {
		if(printParamsAndExit) {
			connectToDavQuit();
			logger.info(
					"Archivsystem zur Parameterausgabe " + (printParamsFile == null ? "auf der Konsole" : ("in '" + printParamsFile.getAbsolutePath() + "'"))
					+ " gestartet"
			);
			try {
				archivConfig.printArchiveParams(printParamsFile);
			}
			catch(Exception e) {
				logger.error("Parameterausgabe fehlgeschlagen: " + e.getMessage());
			}
			return true;
		}
		return false;
	}

	/** Beendet alle Tasks, meldet App vom DAV ab. */
	private void cleanUp() {
		/*
		Der folgende synchronized-Block sorgt dafür, dass alle Threads die (u.U. gleichzeitig/kurz nacheinander) cleanUp() aufrufen,
		immer solange warten, bis das cleanUp() (also das saubere Herunterfahren des Archivsystems) auch wirklich beendet wurde.

		Das heißt der erste Thread führt den CleanUp-Code aus, alle weiteren Threads warten dann solange bis der erste Thread fertig ist.

		Das ist wichtig für den folgenden Fall:
		  - Jemand beendet das Archivsystem über die Telnet-Schnittstelle oder die Datenverteiler-Verbindung bricht ab
		    - cleanUp() wird ausgeführt
		  - Zusätzlich terminiert jemand den Archivsystem-Prozess indem ein SIGTERM, Ctrl+C o. ä. gesendet wird
		    - jetzt führt der ShutdownHook ebenfalls cleanUp() aus
		      - wenn der Shutdown-Hook-Thread vorzeitig beendet wird, bevor das cleanUp() des ursprünglichen Threads abgeschlossen ist,
		       sorgt das dafür, dass sich die Applikation sofort beendet, da die JVM den Prozess nach Bearbeitung aller
		       ShutdownHooks sofort tötet.
		       --> Es ist wichtig, dass der Shutdown-Hook-Thread im cleanUp() solange wartet bis das cleanUp() im ursprünglichen
		       Thread fertig ist.
		*/
		synchronized(_terminatedLock) {
			final boolean wasTerminated;
			wasTerminated = terminated;
			terminated = true;
			if(wasTerminated) return;
			try {
				// Alle periodischen Tasks stoppen:
				logger.info("Beende Task Scheduler...");
				if(taskScheduler != null) taskScheduler.terminate();
				logger.info("Task Scheduler beendet");

				// Archivdaten abmelden
				logger.info("Melde Archivdaten ab...");
				if(archivConfig != null) archivConfig.unsubscribeArchiveData();
				logger.info("Archivdaten abgemeldet");

				// Archivparametrierung abmelden
				logger.info("Melde Archivparametrierung ab...");
				if(archivConfig != null) archivConfig.unsubscribeArchiveParams();
				logger.info("Archivparametrierung abgemeldet");

				//Simulationsverwaltung herunterfahren
				logger.info("Melde Simulation ab...");
				if(simulationMgr != null) simulationMgr.terminateTask();
				logger.info("Simulation abgemeldet");

				// Beenden aller (u.a. der ArchiveTasks) Tasks
				logger.info("Beende Tasks...");
				if(inQueuesMgr != null) inQueuesMgr.stopAllTasks();
				logger.info("Tasks beendet");

				if(persistenceMgr != null) {
					persistenceMgr.saveUnsubscriptionTime();
				}

				// Alle anderen abonnierten Objekte abmelden (nach dem Beenden der ArchivQueryTasks, weil nach dem
				// abgebrochenem take() das Totengloeckchen über den DAV gesendet werden muss):
				logger.info("Melde sonstige Parameter ab...");
				if(getDavCon() != null) ConnectionManager.unsubscribeAll(getDavCon());
				logger.info("Sonstige Parameter abgemeldet");

				// Quittungen erst nach Leeren der Archivwarteschlange abmelden
				logger.info("Melde Quittungen ab...");
				if(archivConfig != null) archivConfig.unsubscribeArchiveDataAck();
				logger.info("Quittungen abgemeldet");

				// Warten bis SettingsManager fertig
				logger.info("Warte auf SettingsManager...");
				if(archivConfig != null) {
					archivConfig.terminateTask();
					archivConfig.join();
				}
				logger.info("SettingsManager beendet");

				logger.info("Starte Finalisierung...");
				if(persistenceMgr != null && runningProperly) {
					persistenceMgr.prepareShutdown();
					persistenceMgr.deleteLockFile(); // Persistenzverzeichnis freigeben
				}
				logger.info("Finalisierung beendet (Indexe geschrieben, Lock-Datei geloescht)");
			}
			catch(Exception e) {
				logger.warning("Fehler beim Herunterfahren", e);
			}
			disconnectFromDav();

			persistenceMgr = null;
			inQueuesMgr = null;
			archivConfig = null;
			taskScheduler = null;
			simulationMgr = null;
		}
	}

	/** Faehrt das ArS herunter und beendet die VM. */
	public void quit() {
		logger.info("Archivsystem wird beendet.");
		cleanUp();	// setzt terminated=true
		logger.info("Bye.");
		if(!_skipSystemExit) {
			System.exit(0);
		}
	}

	/** Faehrt das ArS herunter ohne die VM zu beenden (dies ist bei Fernsteuerung des ArS aus JUnit-Tests notwendig). */
	public void quitNoExit() {
		// Für den Fall, dass ein Test das ArS mehrfach beendet...
		if(!wasTerminated()) {
			logger.info("Archivsystem wird beendet - Java-VM laeuft weiter");
			cleanUp();	// setzt terminated=true
			logger.info("Bye.");
		}
	}

	/** Faehrt das ArS mit einer Fehlermeldung herunter und beendet die VM. */
	public void quitError(String msg) {
		logger.error(msg);
		logger.info("Archivsystem wird wegen eines Fehlers beendet.");
//		dumpThreads();
		if(inQueuesMgr != null) inQueuesMgr.setFastExit(true);
		cleanUp();
		logger.info("Bye (with error). " + msg);
		if(!_skipSystemExit) {
			System.exit(1);
		}
		else {
			throw new RuntimeException(msg);
		}
	}

	private void dumpThreads() {
		StringBuilder out = new StringBuilder();
		out.append("Full thread dump\n");
		Map<Thread, StackTraceElement[]> allStackTraces = Thread.getAllStackTraces();
		for(Map.Entry<Thread, StackTraceElement[]> entry : allStackTraces.entrySet()) {
			out.append("\n");
			Thread thread = entry.getKey();
			out.append(String.format("\"%s@%d\"%s prio=%d tid=0x%x%n", thread.getName(), thread.getId(), thread.isDaemon()?" daemon" : "", thread.getPriority(), thread.getId()));
			StackTraceElement[] trace = entry.getValue();
			for (StackTraceElement traceElement : trace) {
				out.append("\t  at ").append(traceElement).append("\n");
			}
		}
		logger.info(out.toString());
	}

	/** Faehrt das ArS mit einer Fehlermeldung herunter. */
	public void quitErrorNoExit(String msg) {
		if(inQueuesMgr != null) inQueuesMgr.setFastExit(true);
		if(!wasTerminated()) {
			logger.error(msg);
			logger.info("Archivsystem wird wegen eines Fehlers beendet.");
			dumpThreads();
			cleanUp();
			logger.info("Bye (with error).");
		}
	}

	/**
	 * Startpunkt des Archivsystems.
	 *
	 * @param args Kommandozeilenparameter
	 */
	public static void main(String[] args) {
		try {
			ArchiveManager am = new ArchiveManager(args);
			am.run();
		}
		catch(Exception ex) {
			ex.printStackTrace();
			System.exit(4);
		}
	}

	public ArchivConfig getArchivConfig() {
		return archivConfig;
	}

	public DataIdentTree getDataIdentTree() {
		return dataIdentTree;
	}

	public InQueuesMgr getInQueuesMgr() {
		return inQueuesMgr;
	}

	public PersistenceManager getPersMgr() {
		return persistenceMgr;
	}

	public TaskScheduler getTaskScheduler() {
		return taskScheduler;
	}


	/** @return Liefert den SimulationsManager. */
	public SimulationManager getSimulationMgr() {
		return simulationMgr;
	}

	public int getMaximumQueriesPerApplication() {
		return maximumQueriesPerApplication;
	}

	/** @return Liefert die original Command Line Argumente als unmodifiableList von String Objekten. */
	public List<String> getOriginalCommandLineArguments() {
		List<String> result = new ArrayList<>();
		Collections.addAll(result, argList.getInitialArgumentStrings());
		return Collections.unmodifiableList(result);
	}

	public int increaseArchiveQueryCountForApplication(SystemObject application) {
		synchronized(_applicationToArchiveQueryCount) {
			final Integer queryCount = _applicationToArchiveQueryCount.get(application);
			final int newCount;
			if(queryCount == null) {
				newCount = 1;
			}
			else {
				newCount = queryCount + 1;
			}
			_applicationToArchiveQueryCount.put(application, newCount);
			//System.out.println("increaseArchiveQueryCountForApplication: newCount = " + newCount);
			return newCount;
		}
	}

	public int decreaseArchiveQueryCountForApplication(SystemObject application) {
		synchronized(_applicationToArchiveQueryCount) {
			final Integer queryCount = _applicationToArchiveQueryCount.get(application);
			final int newCount;
			if(queryCount == null) {
				throw new IllegalStateException("Anzahl Archivanfragen für Applikation " + application + " darf nicht kleiner als 0 werden");
			}
			else {
				newCount = queryCount - 1;
			}
			if(newCount == 0) {
				_applicationToArchiveQueryCount.remove(application);
			}
			else {
				_applicationToArchiveQueryCount.put(application, newCount);
			}
			//System.out.println("decreaseArchiveQueryCountForApplication: newCount = " + newCount);
			return newCount;
		}
	}

	public int getArchiveQueryCountForApplication(SystemObject application) {
		synchronized(_applicationToArchiveQueryCount) {
			final Integer queryCount = _applicationToArchiveQueryCount.get(application);
			return queryCount == null ? 0 : queryCount;
		}
	}

	/**
	 * Anzahl Threads fürs Index-Schließen beim Beenden
	 * @return Vom Benutzer vorgegebener Wert
	 */
	public int getNumCloseIndexThreads() {
		return _numCloseIndexThreads;
	}

	/**
	 * Anzahl Threads für Persistenz-Prüfung und Wiederherstellung
	 * @return Vom Benutzer vorgegebener Wert
	 */
	public int getNumCheckPersistenceThreads() {
		return _numCheckPersistenceThreads;
	}

	/**
	 * Anzahl Online-Archivqueues/Buffer
	 * @return Vom Benutzer vorgegebener Wert
	 */
	public int getNumOfArchQueuesOnline() {
		return _numOfArchQueuesOnline;
	}

	/**
	 * Anzahl Threads fürs Löschen
	 * @return Vom Benutzer vorgegebener Wert
	 */
	public int getNumDeleteThreads() {
		return _numDeleteThreads;
	}

	/**
	 * Anzahl Threads fürs Sichern
	 * @return Vom Benutzer vorgegebener Wert
	 */
	public int getNumBackupThreads() {
		return _numBackupThreads;
	}

	/**
	 * Wird von ArchiveConfig ausgeführt, wenn alle Daten initial angemeldet wurden.
	 */
	public void archiveIsReady() {
		logger.info("Das Archivsystem ist bereit für Anfragen.");

		// Archivanfragen erlauben
		getInQueuesMgr().subscribeQueries();

		// Fertigmeldung versenden
		getDavCon().sendApplicationReadyMessage();
	}

	public long getIndexCacheMaxCount() {
		return _indexCacheMaxCount;
	}

	public void setIndexCacheMaxCount(final long indexCacheMaxCount) {
		_indexCacheMaxCount = indexCacheMaxCount;
	}

	public int getIndexCacheMaxSize() {
		return _indexCacheMaxSize;
	}

	public void setIndexCacheMaxSize(final int indexCacheMaxSize) {
		_indexCacheMaxSize = indexCacheMaxSize;
	}

	public boolean isUsingDynamicDeletionTime() {
		return _useDynamicDeletionTime;
	}

	/**
	 * Muss immer vor Zugriffen auf den kritischen Bereich (die Indexe oder Container) der Datenidentifikation aufgerufen werden.
	 * <p>
	 * Wenn ein anderer Task sich im kritischen Bereich befindet, blockiert diese Methode bis der kritische Bereich wieder frei ist.
	 * <p>
	 * Da der Index-Bereich möglicherweise von vielen Tasks benutzt wird und auch Zugriffe beim Archivieren von Daten erfolgen sollte
	 * der Bereich so schnell wie möglich wieder verlassen werden. Es ist zwingend erforderlich, das zurückgegebene Lock-Objekt nach Benutzung zu schließen, sonst können andere Threads blockiert werden.
	 * <p>
	 * Beispiel-Code:
	 * <pre>{@code
	 *     try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(din.getDataIdentification())) {
	 *          // Tue was mit den Indexen
	 *     }
	 * }</pre>
	 * <p>
	 * Das zurückgegebene {@link SyncKey}-Objekt verhält sich wie ein {@link java.util.concurrent.locks.ReentrantLock},
	 * es ist also möglich, ein zweites Lock zu erhalten wenn bereits ein Lock geholt wurde.
	 * <p>
	 *     Dieser Funktionsaufruf ist äquivalent zu {@code getPersMgr().lockIndex(...);}.
	 * @param dataIdentification Datenidentifikation
	 * @return Lock
	 * @throws SynchronizationFailedException Synchronisierung ist fehlgeschlagen (Unterbrochen beim Warten)
	 */
	public SyncKey<IdDataIdentification> lockIndex(IdDataIdentification dataIdentification) throws SynchronizationFailedException {
		return persistenceMgr.lockIndex(dataIdentification);
	}

	/**
	 * Muss immer vor Zugriffen auf den kritischen Bereich (die Indexe oder Container) der Datenidentifikation aufgerufen werden.
	 * <p>
	 * Wenn ein anderer Task sich im kritischen Bereich befindet, blockiert diese Methode bis der kritische Bereich wieder frei ist.
	 * <p>
	 * Da der Index-Bereich möglicherweise von vielen Tasks benutzt wird und auch Zugriffe beim Archivieren von Daten erfolgen sollte
	 * der Bereich so schnell wie möglich wieder verlassen werden. Es ist zwingend erforderlich, das zurückgegebene Lock-Objekt nach Benutzung zu schließen, sonst können andere Threads blockiert werden.
	 * <p>
	 * Beispiel-Code:
	 * <pre>{@code
	 *     try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(din.getDataIdentification())) {
	 *          // Tue was mit den Indexen
	 *     }
	 * }</pre>
	 * <p>
	 * Das zurückgegebene {@link SyncKey}-Objekt verhält sich wie ein {@link java.util.concurrent.locks.ReentrantLock},
	 * es ist also möglich, ein zweites Lock zu erhalten wenn bereits ein Lock geholt wurde.
	 * <p>
	 * Dieser Funktionsaufruf ist äquivalent zu {@code getPersMgr().lockIndex(...);}.     
	 *     
	 * @param dataIdentification Datenidentifikation
	 * @param timeout Maximale Wartezeit
	 * @return Lock
	 * @throws SynchronizationFailedException Synchronisierung ist fehlgeschlagen (Timeout oder Unterbrochen beim Warten)
	 */
	public SyncKey<IdDataIdentification> lockIndex(IdDataIdentification dataIdentification, Duration timeout) throws SynchronizationFailedException {
		return persistenceMgr.lockIndex(dataIdentification, timeout);
	}

	private static class ShutdownHook implements Runnable {
		private final WeakReference<ArchiveManager> _archiveManager;

		public ShutdownHook(final ArchiveManager archiveManager) {
			this._archiveManager = new WeakReference<>(archiveManager);
		}

		@Override
		public void run() {
			ArchiveManager archiveManager = _archiveManager.get();
			if(archiveManager != null) {
				archiveManager.quitErrorNoExit("Terminierungssignal empfangen, Archivsystem wird kontrolliert beendet");
			}
		}
	}

	/**
	 * Implementierung eines UncaughtExceptionHandlers, der bei nicht abgefangenen Exceptions und Errors entsprechende Ausgaben macht und im Falle eines Errors den
	 * Prozess terminiert.
	 */
	@SuppressWarnings("UseOfSystemOutOrSystemErr")
	private class ExceptionHandler implements Thread.UncaughtExceptionHandler {

		/** Speicherreserve, die freigegeben wird, wenn ein Error auftritt, damit die Ausgaben nach einem OutOfMemoryError funktionieren */
		private volatile byte[] _reserve = new byte[20000];

		@Override
		public void uncaughtException(Thread t, Throwable e) {
			if(e instanceof Error && _reserve != null) { // 2. Bedingung, damit nicht doppelt exit(1) aufgerufen wird
				// Speicherreserve freigeben, damit die Ausgaben nach einem OutOfMemoryError funktionieren
				_reserve = null;
				try {
					System.err.println("Schwerwiegender Laufzeitfehler: Ein Thread hat sich wegen eines Errors beendet, Prozess wird terminiert");
					System.err.println(t);
					e.printStackTrace(System.err);
					logger.error("Schwerwiegender Laufzeitfehler: " + t + " hat sich wegen eines Errors beendet, Prozess wird terminiert", e);
				}
				catch(Throwable ignored) {
					// Weitere Fehler während der Ausgaben werden ignoriert, damit folgendes exit() auf jeden Fall ausgeführt wird.
				}
				try {
					if(_shutDownHookThread != null) {
						// Shutdownhook entfernen damit Archiv sich sofort beendet, siehe https://www.nerz-ev.de/nerz-fehlerverwaltung/view.php?id=101#c293
						Runtime.getRuntime().removeShutdownHook(_shutDownHookThread);
					}
				}
				catch(Throwable ignored) {
					// Fehler beim Shutdownhook entfernen ignorieren, damit folgendes exit() auf jeden Fall ausgeführt wird.
				}
				if(!_skipSystemExit) {
					System.exit(5);
				}
			}
			else {
				System.err.println("Laufzeitfehler: Ein Thread hat sich wegen einer Exception beendet:");
				System.err.println(t);
				e.printStackTrace(System.err);
				logger.error("Laufzeitfehler: " + t + " hat sich wegen einer Exception beendet", e);
			}
		}
	}
}
