/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerFileHandle;
import de.bsvrz.ars.ars.persistence.ContainerHeaders;
import de.bsvrz.ars.ars.persistence.ContainerManagementInformation;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.OpenContainerData;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.StandardOpenContainerData;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings.CloseCondition;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import java.util.concurrent.atomic.AtomicLong;
import java.util.zip.Deflater;

/**
 * Task zur Archivierung von Datensätzen. Ist nicht von SingleTask abgeleitet, weil ein spezieller ArchiveRingBuffer verwendet wird, der zwei Objekte
 * (ResultData und Archivzeit) speichern kann. Die work()-Methode hat deswegen zwei Parameter.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveTask extends RepeatingTask  {

	/** Kennzeichen der Betriebsmeldung */
	public static final String MSG_PID_DATAINDEX_ANOMALY = "Datenindex nicht monoton";

	protected final AtomicLong _successCount = new AtomicLong(0);

	protected final AtomicLong _failCount = new AtomicLong(0);

	protected final AtomicLong _closeCount = new AtomicLong(0);

	private final ArchiveRingBuffer inputDataQueue;

	private final ArchiveRingBuffer.PopObject popObject;	// zum übertragen des Paares ResultData/Archivzeit

	/** Letzte verwendete Archivzeit. Wird für die Archivzeitüberwachung benutzt. */
	private static long lastArchiveTime;

	// Kennzeichnungen des zu archivierenden Datensatzes:
	private ResultData rd;

	protected long arsTime;
	protected long dataTime;
	protected long dataIdx;
	protected long openContID;

	protected ArchiveDataKind adk;

	protected DataIdentificationDir didPath;

	protected DataIdentNode din;

	/** Länge des serislisierten Datensatzes inklusive Datensatz-Header */
	protected int dataEntryLen;

	private long lastADataIdx, lastNDataIdx, lastDataIdx, lastDataTime;

	private long lastAContID;

	private boolean lastDataWasGapMarker;

	protected final Deflater compresser = new Deflater();

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// ContainerSettings:
	//
	/** Container-Einstellungen des Archivsystems (aus atg.archivContainer). */
	private static ContainerSettings containerSettings;

	ContainerFileHandle _currentContainer;

	final ArchiveDataSerializer _serializer = new ArchiveDataSerializer();


	static {
		// Standardwerte (keine Default-Werte im Datenkatalog vorgesehen):
		final ContainerSettings settings = new ContainerSettings();
		settings.stdCloseConditions.maxContAnzDS = 5000;
		settings.stdCloseConditions.maxContSize = 20 * 1000 * 1000;	// 20 MB
		settings.stdCloseConditions.maxContTime = 60 * 60 * 24 * 7;	// 1 Woche
		setContainerSettings(settings);
	}

	/**
	 * Setzt die Container-Einstellungen des Archivsystems.
	 *
	 * @param cs Container-Einstellungen
	 */
	public static void setContainerSettings(ContainerSettings cs) {
		containerSettings = cs;
	}

	/**
	 * Liefert die Container-Einstellungen des Archivsystems.
	 *
	 * @return Container-Einstellungen
	 */
	public static ContainerSettings getContainerSettings() {
		return containerSettings;
	}
	//
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Konstruktor des Archiv-Tasks.
	 *
	 * @param archiveMgr Archiv-Manager
	 * @param tskIdx     Task-ID
	 * @param inQueue    Input-Queue
	 */
	public ArchiveTask(ArchiveManager archiveMgr, int tskIdx, ArchiveRingBuffer inQueue) {
		super(archiveMgr);
		setName(getName() + "_" + tskIdx);
		inputDataQueue = inQueue;
		popObject = inQueue.new PopObject();	// Zum übergeben von Paaren (Data, Archivzeit)
	}

	@Override
	public void step() throws InterruptedException {
		inputDataQueue.pop(popObject);		// synchronized, blockiert falls leer
		//noinspection ObjectEquality
		if(popObject.object == TERMINATE_TASK || shouldTerminate()) {
			return;
		}

		work(popObject.object, popObject.time);

		taskStepDone();
	}

	@Override
	public void terminateTask() {
		synchronized(this) {
			_shouldTerminate = true;
			try {
				inputDataQueue.push(TERMINATE_TASK, 0, 0);
			}
			catch(InterruptedException ignored) {
			}
		}
	}

	/** @return Anzahl der seit Systemstart zur Archivierung verarbeiteten Datensätze (erfolgreich und nicht erfolgreich archiviert) */
	public long getProcessedCount() {
		return _successCount.longValue() + _failCount.longValue();
	}

	/** @return Anzahl der seit Systemstart erfolgreich archivierten Datensätze */
	public long getSuccessCount() {
		return _successCount.longValue();
	}

	/** @return Anzahl der seit Systemstart nicht erfolgreich archivierten Datensätze */
	public long getFailedCount() {
		return _failCount.longValue();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} erfolgreich verarbeiteten Close-Container-Datensätze. Kann zur
	 *         Test-Synchronisation verwendet werden.
	 */
	public long getCloseContainerSuccess() {
		return _closeCount.longValue();
	}


	/** Setzt alle Zähler auf 0. */
	public void resetDSCounter() {
		_successCount.set(0);
		_failCount.set(0);
		_closeCount.set(0);
	}

	public static void setLastArchiveTime(long atime) {
		lastArchiveTime = atime;
	}

	public static long getLastArchiveTime() {
		return lastArchiveTime;
	}


	/**
	 * Archiviert den übergebenen Datensatz unter dem angegebenen Archivzeitstempel.
	 *
	 * @param resultData Ergebnisdatensatz
	 * @param archTime   Archivzeitstempel
	 */
	public void work(Object resultData, long archTime) {
		archMgr.getInQueuesMgr().decrOnlineDataInQueues();
		archiveData(resultData, archTime);
	}

	private void archiveData(final Object resultData, final long archTime) {
		// Ist der DS ein vom ArS erzeugter DS (z.B. um Cont.-Abschluss zu erzwingen)?
		// archTime ist dann in Wahrheit die Object-ID
		if(resultData instanceof CloseContainerObject) {
			closeContainer(archTime, (CloseContainerObject)resultData); // Sync'ed selbst auf DIN
			return;
		}

//		logReceivedData((ResultData)resultData);

		// Synchronisierung über den DataIdentNode:
		try {
			presetData((ResultData)resultData, archTime);

			IdDataIdentification dataIdentification = din.getDataIdentification();
			IdContainerFileDir idContainerFileDir = new IdContainerFileDir(dataIdentification, adk);
			IdContainerFileDir idContainerFileDirOnline = new IdContainerFileDir(dataIdentification, ArchiveDataKind.ONLINE);
			IdContainerFileDir idContainerFileDirDelayed = new IdContainerFileDir(dataIdentification, ArchiveDataKind.ONLINE_DELAYED);

			// Indexe zuweisen, falls nicht schon geschehen
			// Es müssen stets beide Indexe (aktuell/nachgeliefert) angelegt werden, da Rückspruenge
			// und Lücken über beide Datensatzarten gehen.

			openContID = persMgr.getOpenContID(idContainerFileDir);

			if(openContID > 0 && din.isFirstDataAfterSubscription(adk)) {
				// Überprüfung des offenen Containers auf Konsistenz bevor Daten angehängt werden
				persMgr.checkContainerFile(dataIdentification, adk, openContID);
			}

			lookupLastDataIdxes(dataIdentification, false);	// Letzte Datenindexe, Datenzeiten nachsehen
			StandardOpenContainerData onlineContainerData = persMgr.getOpenContainerData(idContainerFileDirOnline);
			StandardOpenContainerData delayedContainerData = persMgr.getOpenContainerData(idContainerFileDirDelayed);
			
			if(handleIndexBackstep(onlineContainerData, delayedContainerData)) return;

			// Lücke? (Wenn der letze OA-Datensatz "keine Quelle" war, keine Lücke einfuegen)
			// Nicht extra prüfen, wenn beim letzten Aufruf PotGap eingefuegt wurde und diesmal der echte DS.
			// Aus Performance-Gruenden wird im DIN vermerkt, ob der letzte DS ein "keine Quelle" war. Damit
			// spart man den Dateizugriff.
			boolean insertGapDS = false;
			if(!lastDataWasGapMarker && isDataGap()) {
				boolean b = din.lastDataNoSourceAssigned();
				if(b) {
					if(!din.lastDataWasNoSource()) {
						insertGapDS = true;
					}
				}
				else {
					// Hier muss man leider auf den Index synchronisieren um lesezugriff auf den Container zu haben
					if(!scanOAContForLastNoSrc()) {
						insertGapDS = true;
					}
				}
			}

			if(insertGapDS) {
				// Lücken werden eingefuegt, indem zuerst der Datensatz auf einen Lücken-Datensatz umgesetzt
				// wird und danach work() mit dem urspr. Datensatz nochmal aufgerufen wird. Alles andere bleibt gleich.
				// Lücken werden immer im OA-Cont. eingetragen. Also ggf. Datensatzart wechseln.
				if(!adk.equals(ArchiveDataKind.ONLINE)) {
					adk = ArchiveDataKind.ONLINE;
					didPath = new DataIdentificationDir(persMgr, din);
					openContID = persMgr.getOpenContID(idContainerFileDirOnline);
				}
				// DIdx der Lücke gemäß TAnfArS 5.1.2.4.3.5 setzen
				dataIdx = Util.dIdxSetArSBit(lastDataIdx);
				if(din.isFirstDataAfterSubscription(adk)) {
					// Datenzeit nach einem Neustart gemäß Nerz-Ä-78
					long unsubscriptionTime = din.getUnsubscriptionTime();
					dataTime = unsubscriptionTime == -1 ? lastDataTime : unsubscriptionTime;
				}
				else {
					dataTime = lastArchiveTime;
				}
				dataEntryLen = _serializer.serializeDataGap(dataIdx, dataTime, arsTime, compresser);
			}
			else {
				dataEntryLen = _serializer.serializeData(rd, arsTime, compresser);
			}

			if(storeData()) {
				din.setFirstDataAfterSubscription(adk, false);							// ab jetzt duerfen Datenindexe nicht mehr gleich sein
				din.setLastOAWasNoSource(
						adk.equals(ArchiveDataKind.ONLINE) && rd.isNoSourceAvailable()
				);	// Für nächsten DS merken, damit man nicht in den Cont. schauen muss
				if(rd.isNoSourceAvailable()) {
					if(din.getUnsubscriptionTime() == -1) {
						din.setUnsubscriptionTime(rd.getDataTime());
					}
					din.setValidData(false);
				}
				else {
					din.setUnsubscriptionTime(-1);
					din.setValidData(true);
				}
				if(!insertGapDS) {
					_successCount.incrementAndGet();	// Datenlückensaetze zählen nicht
					if(rd.hasData()) sendAck(rd);	// nicht quittieren, falls kQ/kD oder nur eine Datenlücke eingefuegt wurde
				}
			}
			else {
				_failCount.incrementAndGet();
			}

			// Vermerken, dass als nächstes der echte DS außer der Reihe mit work() eingefuegt wird.
			// Da dieser ebenfalls eine Lücke zum Gap-DS hat, muss verhindert werden, dass wieder eine
			// Datenlücke festgestellt wird.
			lastDataWasGapMarker = insertGapDS;
			if(lastDataWasGapMarker) {
				// Letzter DS war Datenlücke: Einfuegen des echten Datensatzes nachholen.
				// Dazu alles so verlassen, dass die Methode wieder normal ablaufen kann:
				_currentContainer.leaveContainer();
				archiveData(rd, arsTime);
			}
		}
		catch(IndexException | PersistenceException | SynchronizationFailedException e) {
			handleErr(e, Util.rd2Str(rd));
		}
	}

	protected boolean storeData() throws PersistenceException, IndexException, SynchronizationFailedException {
		try {
			IdDataIdentification dataIdentification = din.getDataIdentification();
			if(openContID != -1) {
				_currentContainer = persMgr.accessOpenContainer(dataIdentification, adk, openContID);
				StandardOpenContainerData openContainerData = persMgr.getOpenContainerData(new IdContainerFileDir(dataIdentification, adk));
				if(openContainerData != null && !mustCloseContainer(openContainerData)) {
					// Standard-Fall, an Container anhängen
					
					_currentContainer.appendSerializedData(_serializer.getActualWriteBuf(), _serializer.getTotalWriteDataSize());
					persMgr.updateMaxValues(dataIdentification, adk, dataIdx, arsTime, dataTime);
				}
				else {
					try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(dataIdentification)) {
						_currentContainer.closeContainer(lock, adk);
						_currentContainer.leaveContainer();
						if(tooLargeForNewContainer()) return false;
						openContID = persMgr.nextContainerID();
						_currentContainer = persMgr.accessOpenContainer(dataIdentification, adk, openContID);    // legt notfalls Verzeichnis an
						_currentContainer.createContainer(din.arSParamIsSichern());
						_currentContainer.appendSerializedData(_serializer.getActualWriteBuf(), _serializer.getTotalWriteDataSize());
						persMgr.updateContainerIndex(lock, _currentContainer);
						persMgr.addIndexEntries(lock, adk, dataIdx, arsTime, dataTime, openContID);
					}
				}
			}
			else {
				// Container entweder abgeschlossen oder es gibt keinen, also neuen anlegen
				try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(dataIdentification)) {
					if(tooLargeForNewContainer()) return false;
					openContID = persMgr.nextContainerID();
					_currentContainer = persMgr.accessOpenContainer(dataIdentification, adk, openContID);
					_currentContainer.createContainer(din.arSParamIsSichern());
					_currentContainer.appendSerializedData(_serializer.getActualWriteBuf(), _serializer.getTotalWriteDataSize());
					persMgr.updateContainerIndex(lock, _currentContainer);
					persMgr.addIndexEntries(lock, adk, dataIdx, arsTime, dataTime, openContID);
				}
			}
			createRebuildIndexFlagFile();		// Markieren, dass etwas verändert wurde
			return true;
		}
		finally {
			if(_currentContainer != null) {
				_currentContainer.leaveContainer();
			}
		}
	}

	/**
	 * Setzt Flags, IDs, holt den DataIdentNode, usw.
	 *
	 * @param resultData Aktueller Ergebnisdatensatz
	 * @param archTime   Archivzeitstempel
	 */
	private void presetData(ResultData resultData, long archTime) {
		rd = resultData;
		lastADataIdx = lastNDataIdx = lastDataIdx = lastDataTime = lastAContID = -1;
		arsTime = archTime;
		dataTime = rd.getDataTime();
		dataIdx = rd.getDataIndex();
		adk = rd.isDelayedData() ? ArchiveDataKind.ONLINE_DELAYED : ArchiveDataKind.ONLINE;
		din = didTree.get(resultData);
		didPath = new DataIdentificationDir(persMgr, din);
	}

	/**
	 * Schliesst einen Container ab. Wird durch ein spezielles Objekt ausgeloest, das in die Queue der ArchiveTask gelegt wird. Um Platz zu sparen, wird anstelle
	 * der Archivzeit die Objekt-ID gespeichert. Der Aufrufer dieser Methode muss die Synchronisierung über den DataIdentNode durchfuehren.
	 *
	 * @param closeObjID
	 * @param cco
	 */
	private void closeContainer(long closeObjID, CloseContainerObject cco) {

		IdDataIdentification did = new IdDataIdentification(closeObjID, cco.atgID, cco.aspID, cco.sv);
		// DataIdentNode besorgen:
		din = didTree.get(did);

		try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(did)) {
			for(ArchiveDataKind dataKind : ArchiveDataKindCombination.all()) {
				openContID = persMgr.getOpenContID(new IdContainerFileDir(did, dataKind));
				// Es kann sein, dass Indexe für eine Anfrage angelegt wurden, ohne dass Daten da sind.
				// Drum hier nochmal prüfen
				if(openContID != -1) {
					try {
						_currentContainer = persMgr.accessOpenContainer(did, dataKind, openContID);
						_currentContainer.closeContainer(lock, dataKind);
					}
					catch(PersistenceException e) {
						handleErr(e, cco);
					}
					finally {
						_currentContainer.leaveContainer();
					}
				}
			}
			_closeCount.incrementAndGet();
		}
		catch(Exception e) {
			handleErr(e, cco);
		}
	}

	/**
	 * Durchläuft den aktuellen Container und prüft, ob der letzte Datensatz die Kennung "keine Quelle" hat (langsam!). Dies ist nur notwendig, wenn das erste
	 * mal auf die DID zugegriffen wird. Sonst genuegt {@link DataIdentNode#lastDataWasNoSource()}. Der Aufrufer dieser Methode muss die Synchronisierung über den
	 * DataIdentNode durchfuehren.
	 *
	 * @return Wahr, falls der letzte Datensatz die Kennung "keine Quelle" hat, falsch sonst
	 */
	private boolean scanOAContForLastNoSrc() {
		boolean result = false;
		if(lastAContID == -1) return false;
		try {
			try(ContainerFileHandle containerFileHandle = persMgr.accessOpenContainer(din.getDataIdentification(), ArchiveDataKind.ONLINE, lastAContID);
			    DataIterator iterator = containerFileHandle.iterator()) {
				if(iterator.getContainerManagementData().getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) {
					return false;
				}
				ContainerDataResult dataResult = new ContainerDataResult();
				while(!iterator.isEmpty()) {
					iterator.poll(dataResult);
					result = dataResult.isNoSource();
				}
			}
		}
		catch(Exception e) {
			_debug.warning("Fehler beim Scan über den letzten Container wird ignoriert", e);
			result = false;
		}
		return result;
	}

	/**
	 * Sucht die letzten Datenindexe im Index sowie den Container.
	 *
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	protected void lookupLastDataIdxes(final IdDataIdentification dataIdentification, boolean requested) throws IndexException, SynchronizationFailedException {
		ArchiveDataKind a = requested ? ArchiveDataKind.REQUESTED : ArchiveDataKind.ONLINE;
		ArchiveDataKind n = requested ? ArchiveDataKind.REQUESTED_DELAYED : ArchiveDataKind.ONLINE_DELAYED;
		OpenContainerData acd = persMgr.getOpenContainerData(new IdContainerFileDir(dataIdentification, a));
		OpenContainerData ncd = persMgr.getOpenContainerData(new IdContainerFileDir(dataIdentification, n));
		lastADataIdx =  acd == null ? -1 : acd.getMaxDataIdx();
		lastNDataIdx =  ncd == null ? -1 : ncd.getMaxDataIdx();
		lastAContID = acd == null ? -1 :  acd.getContainerId();
		long lastADTime = acd == null ? -1 :  acd.getMaxDataTime();
		long lastNDTime =  ncd == null ? -1 : ncd.getMaxDataTime();
		if(lastADataIdx == -1 && lastNDataIdx == -1) {
			// nicht in OpenContainerData, aus Containerdatei lesen.
			ContainerHeaders irA;
			ContainerHeaders irN;
			try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(dataIdentification)) {
				irA = persMgr.getLastContainerHeaders(lock, a);
				irN = persMgr.getLastContainerHeaders(lock, n);
			}
			lastADataIdx = irA != null ? irA.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_DATA_IDX_MAX) : -1;
			lastNDataIdx = irN != null ? irN.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_DATA_IDX_MAX) : -1;
			lastADTime = irA != null ? irA.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_DATA_TIME_MAX) : -1;
			lastNDTime = irN != null ? irN.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_DATA_TIME_MAX) : -1;
			lastAContID = irA != null ? irA.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_CONT_ID) : -1;
		}
		lastDataIdx = Math.max(lastADataIdx, lastNDataIdx);
		lastDataTime = Math.max(lastADTime, lastNDTime);
	}

	/**
	 * Prüft, ob eine potentielle Datenlücke vorliegt. Der Aufrufer dieser Methode muss die Synchronisierung über den DataIdentNode durchfuehren.
	 *
	 * @return Wahr, falls eine potentielle Datenlücke vorliegt, falsch sonst.
	 *
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	private boolean isDataGap() {
		if(ArchiveManager.ONLY_DATA) return false;	// nur für Testzwecke
		if(lastDataIdx == -1) {
			return false;	// noch nie vorher was archiviert
		}

		// "63/93-Fehler": Wenn der letzte DS ein Gap-Marker ist, der nächste DS aber nicht geschrieben
		// wurde, wuerde das Setzen des ArSBits von lastDataIdx nichts ändern und einen Indexfehler
		// verursachen. Der neue DS wird nicht geschrieben und das Problem wiederholt sich.
		// Daher kann es durch die folgende Abfrage nie eine Lücke geben, wenn der letzte
		// DS eine Lückenmarkierung ist.
		else if(Util.dIdxArSBit(lastDataIdx) == 1) {
			return false;
		}

		else if(Util.dIdxNoModBits(dataIdx) > Util.dIdxNoModBits(lastDataIdx) + 1) {
			_debug.fine(
					"Potentielle Datenluecke erkannt: letzer archivierter Index (O" + (lastADataIdx > lastNDataIdx ? "A" : "N") + "): " + lastDataIdx + " ["
					+ Util.dIdx2Str(lastDataIdx) + "] [" + Util.dIdx2StrExt(lastDataIdx) + "]" + ", erhaltener Datenindex" + dataIdx + " ["
					+ Util.dIdx2Str(dataIdx) + "] [" + Util.dIdx2StrExt(dataIdx) + "]"
			);
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Prüft, ob der Datenindex plausibel ist und sendet ggf. die entsprechenden Betriebsmeldungen. Prüft, ob der Datenindex kleiner, gleich oder Größer als
	 * der zuletzt archivierte DI ist. Wenn der DI kleiner ist, wird eine BetrMeld verschickt und nichts archiviert. Wenn der DI gleich ist und dies der erste DS
	 * nach einer Neuanmeldung ist, wird der DS ignoriert (weil der DAV möglicherweise einen gepufferten DS geschickt hat). Wenn der DI Größer ist, wird normal
	 * archiviert.
	 *
	 * @return Wahr, falls der Index nicht plausibel ist und der Datensatz nicht archiviert werden darf, falsch sonst.
	 *
	 */
	private boolean handleIndexBackstep(@Nullable final OpenContainerData online, @Nullable final OpenContainerData delayed) {
		if(dataIdx == lastDataIdx && din.isFirstDataAfterSubscription(adk)) {
			_debug.fine("Datenindex nach Neuanmeldung identisch zum letzten archivierten Datenindex. Datensatz ignoriert." + Debug.NEWLINE, Util.rd2Str(rd));
			return true;
		}
		if(dataIdx <= lastDataIdx) {
			DataDescription dd = rd.getDataDescription();
			StringBuilder sb = new StringBuilder(8192);
			sb.append("Es ist ein fehlerhafter Datenindex aufgetreten. Der soeben erhaltene Datenindex ist nicht streng monoton steigend.");
			sb.append("\n\nSoeben erhalten\n===============");
			sb.append("\nObjekt:              ");
			sb.append(rd.getObject());
			sb.append(" [ID=" + rd.getObject().getId() + "])");
			sb.append("\nAttributgruppe:      ");
			sb.append(dd.getAttributeGroup());
			sb.append(" [ID=" + dd.getAttributeGroup().getId() + "]");
			sb.append("\nAspekt:              ");
			sb.append(dd.getAspect());
			sb.append(" [ID=" + dd.getAspect().getId() + "]");
			sb.append("\nSimVar:              ");
			sb.append(dd.getSimulationVariant());
			sb.append("\nDatenindex:          ");
			sb.append(dataIdx);
			sb.append(" [" + Util.dIdx2Str(dataIdx) + "] [" + Util.dIdx2StrExt(dataIdx) + "]");
			sb.append("\nDatenzeit:           ");
			sb.append(Util.timestrMillisFormatted(rd.getDataTime()));
			sb.append("\nerhalten:            ");
			sb.append(Util.timestrMillisFormatted(arsTime));
			sb.append("\nArt:                 ");
			sb.append(rd.getDataKind());
			sb.append("\nkeine Quelle:        ");
			sb.append(Util.bTF(rd.isNoSourceAvailable()));
			sb.append("\nQuelle verfügbar:    ");
			sb.append(Util.bTF(rd.isSourceAvailable()));
			sb.append("\nkeine Daten:         ");
			sb.append(Util.bTF(rd.isNoDataAvailable()));
			sb.append("\nkeine Rechte:        ");
			sb.append(Util.bTF(rd.isNoRightsAvailable()));
			sb.append("\nungültige Anmeldung: ");
			sb.append(Util.bTF(rd.isNoValidSubscription()));
			sb.append("\n\n**Beginn Daten**\n");
			sb.append(rd);
			sb.append("\n**Ende Daten**\n");
			sb.append("\nZuletzt archiviert\n==================");
			sb.append("\nArt:                 ");
			sb.append(ArchiveDataKind.ONLINE);
			if(lastADataIdx < 0) {
				sb.append("\nDatenindex:          -");
				sb.append("\nDatenzeit:           -");
				sb.append("\nArchivzeit:          -");
			}
			else {
				sb.append("\nDatenindex:          ");
				sb.append(lastADataIdx);
				sb.append(" [" + Util.dIdx2Str(lastADataIdx) + "]");
				sb.append(" [" + Util.dIdx2StrExt(lastADataIdx) + "]");
				if(online != null) {
					sb.append("\nDatenzeit:           " + Util.timestrMillisFormatted(online.getMaxDataTime()));
					sb.append("\nArchivzeit:          " + Util.timestrMillisFormatted(online.getMaxArcTime()));
				}
			}
			sb.append("\nArt:                 ");
			sb.append(ArchiveDataKind.ONLINE_DELAYED);
			if(lastNDataIdx < 0) {
				sb.append("\nDatenindex:          -");
				sb.append("\nDatenzeit:           -");
				sb.append("\nArchivzeit:          -");
			}
			else {
				sb.append("\nDatenindex:          ");
				sb.append(lastNDataIdx);
				sb.append(" [" + Util.dIdx2Str(lastNDataIdx) + "]");
				sb.append(" [" + Util.dIdx2StrExt(lastNDataIdx) + "]");
				if(delayed != null) {
					sb.append("\nDatenzeit:           " + Util.timestrMillisFormatted(delayed.getMaxDataTime()));
					sb.append("\nArchivzeit:          " + Util.timestrMillisFormatted(delayed.getMaxArcTime()));
				}
			}
			sb.append("\n");
			_debug.warning("Betriebsmeldung: Datenindex nicht monoton", sb.toString());
			MessageSender.getInstance().sendMessage(
					MSG_PID_DATAINDEX_ANOMALY, MessageType.APPLICATION_DOMAIN, "", MessageGrade.FATAL, MessageState.MESSAGE, sb.toString()
			);
			return true;
		}
		return false;
	}

	/**
	 * Sendet eine Quittung für den angegebenen Datensatz.
	 *
	 * @param rd Datensatz
	 */
	private void sendAck(ResultData rd) {
		// Der Quittierungsaspekt wird von ArchivConfig sofort abgemeldet, sobald eine DID nicht mehr
		// archiviert/parametriert/quittiert wird.
		// Daher kann für solche bereits in der Queue vorhandenen DS keine Quittung gesendet werden.

		if(din.isArSParameterized() && din.arSParamIsArchivieren() && din.arSParamIsQuittierenValid()) {
			long qAspId = din.arSParamGetQuittieren();
			try {
				archMgr.getInQueuesMgr().getDataAckSender().sendAck(rd, qAspId);
				_debug.finest("Quittung unter Aspekt '" + qAspId + "' gesendet fuer", Util.rd2Str(rd));
			}
			catch(DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
				_debug.warning("Senden der Quittung unter Aspekt '" + qAspId + "' fehlgeschlagen: " + e.getMessage() + Debug.NEWLINE + Util.rd2Str(rd));
			}
		}
	}

	protected void handleErr(Exception e, String corpusDelicti) {
		_failCount.incrementAndGet();
		_debug.error(getName() + ": Fehler beim Archivieren eines Datensatzes: " + e.getMessage() + Debug.NEWLINE + corpusDelicti, e);
	}

	protected void handleErr(Exception e, CloseContainerObject cco) {
		_debug.error(getName() + ": Fehler beim Abschliessen eines Containers: " + e.getMessage() + Debug.NEWLINE + cco, e);
	}

	/**
	 * Erzeugt die Kennzeichen-Datei, ob die Indexe verändert wurden, um nach einem Absturz den Wiederanlauf zu beschleunigen.
	 *
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	protected void createRebuildIndexFlagFile() throws PersistenceException {
		if(!din.isRebuildIndexFileFlagCreated(adk)) {
			persMgr.createRebuildIxdFlagFile(didPath, adk);
			din.rebuildIndexFileFlagWasCreated(adk);
		}
	}
//
//	protected void logReceivedData(ResultData rd) {
//		_debug.finer(getName() + ": Datensatz zur Archivierung erhalten: ", rd);
//	}
//
//	protected void logWriteData() throws IndexException {
//		_debug.finest(
//				getName() + ": Datensatz (" + Util.dsKeys2Str(dataIdx, arsTime, dataTime) + ") in Container (ID=" + persMgr.getOpenContID(adk) + ") angefuegt"
//		);
//	}
//
//	protected void logNewContainer() {
//		_debug.finer(getName() + ": neuen Container (ID=" + openContID + ") fuer " + didPath + "angelegt");
//	}
//
//	private void logDataGap() {
//		_debug.fine(
//				"Potentielle Datenluecke vermerkt: letzer archivierter Index (O" + (lastADataIdx > lastNDataIdx ? "A" : "N") + "): " + lastDataIdx + " ["
//				+ Util.dIdx2Str(lastDataIdx) + "] [" + Util.dIdx2StrExt(lastDataIdx) + "]" + ", erhaltener Datenindex" + dataIdx + " [" + Util.dIdx2Str(dataIdx)
//				+ "] [" + Util.dIdx2StrExt(dataIdx) + "]"
//		);
//	}
//
//	protected void logManualCloseContainer(long closeObjID, CloseContainerObject cco) {
//		_debug.finer(
//				getName() + ": Container (ID=" + openContID + ") fuer " + didPath
//				+ " wg. Parameteraenderung abgeschlossen"
//		);
//	}

	/**
	 * Prüft ob ein Container abgeschlossen werden muss. Die Prüfung basiert auf {@link #containerSettings} und Daten des offenen Containers.
	 * Der Aufrufer dieser Methode muss die Synchronisierung über den DataIdentNode durchfuehren.
	 *
	 * @return Wahr, falls Container abgeschlossen werden muss, falsch sonst.
	 * @param openContainerData Daten des offenen Containers
	 */
	private boolean mustCloseContainer(final OpenContainerData openContainerData) throws PersistenceException {
		long previousDataIndex = openContainerData.getMaxDataIdx();
		if(dataIdx < previousDataIndex) {
			_debug.fine(
					getName() + ": Datenindex-Ruecksprung in aktuellen Daten fuer "
							+ didPath + ": " + Util.timestrMillisFormatted(dataTime) + " < "
							+ Util.timestrMillisFormatted(openContainerData.getMaxDataTime())
					);
			return true;
		}
		long previousDataTime;
		if(adk.isDelayed()) {
			// Bei nachgelieferten Daten muss die Datenzeit innerhalb eines Containers monoton sein,
			// damit das Einsortieren nach Datenzeit bei späteren Archivanfragen funktioniert
			previousDataTime =openContainerData.getMaxDataTime();
		}
		else {
			// Bei Online-Daten kann es Datenzeitrücksprünge geben (da dort eh nie einsortiert wird), 
			// aber die Startzeit den Containers darf nicht unterschritten werden.
			previousDataTime = openContainerData.getMinDataTime();
		}
		if(dataTime < previousDataTime) {
			_debug.fine(
					getName() + ": Datenzeit-Ruecksprung in aktuellen Daten fuer "
							+ didPath + ": " + Util.timestrMillisFormatted(dataTime) + " < "
							+ Util.timestrMillisFormatted(previousDataTime)
					);
			return true;
		}
		CloseCondition closeCond = containerSettings.getExceptionSettings(getAtg());
		if(closeCond != null) {
			return checkCloseCondition(closeCond, openContainerData);
		}
		else {
			return checkCloseCondition(containerSettings.stdCloseConditions, openContainerData);
		}
	}

	private AttributeGroup getAtg() {
		return (AttributeGroup) archMgr.getDataModel().getObject(din.getDataIdentification().getAtgId());
	}

	/**
	 * Prüft extra auf ContainerGröße. Damit kann verhindert werden, dass ständig neue Container angelegt werden, wenn bereits mit einem Datum die Größe
	 * überschritten wuerde. Stattdessen wird mit der maximal erlaubten Größe eine Fehlermeldung ausgegeben. Der Aufrufer dieser Methode muss die
	 * Synchronisierung über den DataIdentNode durchfuehren.
	 *
	 * @return wahr, falls die maximale ContainerGröße überschritten ist, falsch sonst
	 */
	private boolean tooLargeForNewContainer() {
		long totalLen = ContainerFile.getHeaderLen() + dataEntryLen;
		CloseCondition closeCond = containerSettings.getExceptionSettings(getAtg());
		if(closeCond != null && totalLen > closeCond.maxContSize) {
			_debug.warning(
					getName() + ": Datensatz (" + dataEntryLen + " Bytes) + Header (" + ContainerFile.getHeaderLen()
					+ " Bytes) ist größer als die maximale Containerlänge (" + closeCond.maxContSize + " Bytes) und wurde nicht archiviert"
			);
			return true;
		}
		else if(totalLen > containerSettings.stdCloseConditions.maxContSize) {
			_debug.warning(
					getName() + ": Datensatz (" + dataEntryLen + " Bytes) + Header (" + ContainerFile.getHeaderLen()
					+ " Bytes) ist größer als die maximale Containerlänge (" + containerSettings.stdCloseConditions.maxContSize
					+ " Bytes) und wurde nicht archiviert"
			);
			return true;
		}
		return false;
	}

	/**
	 * Prüft ob der aktuelle Container die angegebenen Kriterien zum Container-Abschliessen erfuellt.
	 *
	 * @param closeCond Container-Abschluss-kriterien
	 *
	 * @param openContainerData
	 * @return Wahr, wenn der Container abgeschlossen werden muss, falsch sonst.
	 *
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	private boolean checkCloseCondition(CloseCondition closeCond, final OpenContainerData openContainerData) throws PersistenceException {
		return (openContainerData.getNumContainerEntries() >= closeCond.maxContAnzDS) || (dataTime - openContainerData.getMinDataTime() >= closeCond.maxContTime * 1000) || (
				_currentContainer.getContainerSize() + dataEntryLen > closeCond.maxContSize); // erfordert die meiste Zeit
	}


	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Class: CloseContainerObject
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Eine Instanz dieser Klasse wird eingefuegt, wenn ein Container aufgrund einer Parametrierungs-Änderung abgeschlossen werden soll. */
	public static class CloseContainerObject {

		public final long atgID;
		public final long aspID;

		public final short sv;

		public CloseContainerObject(DataDescription dd) {
			atgID = dd.getAttributeGroup().getId();
			aspID = dd.getAspect().getId();
			sv = dd.getSimulationVariant();
		}

		public String toString() {
			return "atgID=" + atgID + ", aspID=" + aspID + ", sv=" + sv;
		}
	}
}
