/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.BackupImplementation;
import de.bsvrz.ars.ars.backup.BackupModul;
import de.bsvrz.ars.ars.backup.BackupModule;
import de.bsvrz.ars.ars.backup.Container;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.ContainerFileHandle;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.ContainerHeaders;
import de.bsvrz.ars.ars.persistence.ContainerManagementInformation;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirWalk;
import de.bsvrz.ars.ars.persistence.walk.internal.StatusPrinter;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Properties;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

/**
 * Task zur Durchführung eines Sicherungslaufs.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class BackupTask extends SingleClientSenderTask {

	/** das verwendete Backupmodul */
	private static BackupImplementation _backupModul;

	/** Name der Datei, in der alle Indexdaten des momentanen Mediums temporär abgelegt werden */
	public static final String INDEXFILE_NAME = "index.dat";

	/** die immer aktuell gehaltene backupMediumID */
	private int _currentBackupMediumID;

	/** Datei, in der die momentan auf dem Medium gesicherten Indexdaten abgelegt sind */
	private File _currentIndexFile;

	/** Temporäres Verzeichnis in dem {@link #_currentIndexFile} abgelegt ist */
	private File _indexFileDir;

	/** Temporäre Datei die zur Erzeugung des temporären Verzeichnisses {@link +_indexFileDir} benutzt wird. */
	private File _tmpFile;

	/** Objekt, dem Informationen zum aktuellen Zustand der Task und dem Fortschritt von langwierigen Arbeiten übergeben werden. */
	private final LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;

	/**
	 * Erstellt einen neuen BackupTask
	 * @param archiveMgr                 Archivmanager
	 * @param inputData                  Eingangswarteschlange
	 * @param longTermTaskStatePublisher siehe {@link LongTermTaskStatePublisher}
	 */
	public BackupTask(
			final ArchiveManager archiveMgr,
			final BlockingQueue<Object> inputData,
			final LongTermTaskStatePublisher longTermTaskStatePublisher) {
		super(archiveMgr, inputData);
		setName("Backup Task");
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Sicherung");
	}

	@Override
	public void execute() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.execute();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	@Override
	protected void work(final ResultData resultData) {
		if(!(resultData instanceof BackupResultData) && (!resultData.hasData() || shouldTerminate())) {
			return; // Externer Auftrag, aber es ist unbekannt, vom wem dieser kam...
		}
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);

		try {
			if(resultData instanceof BackupResultData) {
				_debug.info("Sicherungslauf startet (Automatische Sicherung)");
				runBackup();
			}
			else {
				try {
					_debug.info("Sicherungslauf startet (Manuelle Sicherung)");
					init(resultData.getData());
					runBackup();
					sendSuccessResponse();
				}
				catch(NoSuchVersionException e) {
					_debug.warning("Konnte Sicherunganfrage nicht deserialisieren", e);
					sendErrorResponse("Konnte Sicherunganfrage nicht deserialisieren: " + e.getMessage());
				}
			}
		}
		catch(BackupException e) {
			_debug.error("Sicherungslauf fehlgeschlagen", e);
			if(!(resultData instanceof BackupResultData)) {
				sendErrorResponse("Sicherungslauf fehlgeschlagen: " + e.getMessage());
			}
		}
		catch(RuntimeException e) {
			_debug.warning("Unerwarteter Fehler während der Sicherung", e);
			if(!(resultData instanceof BackupResultData)) {
				sendErrorResponse("Unerwarteter Fehler während der Sicherung: " + e.getMessage());
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}

	/**
	 * Startet den Sicherungslauf
	 *
	 * @throws de.bsvrz.ars.ars.backup.BackupException
	 *          Fehler beim Backup
	 */
	private void runBackup() throws BackupException {
		try {
			if(_backupModul == null) {
				throw new IllegalStateException(
						"Das Sicherungsmodul wurde nicht korrekt eingerichtet. Bitte überprüfen Sie den Startparameter '-backup' des Archivsystems und den Inhalt der angegebenen Einstellungsdatei."
				);
			}
			final String backupRunID = "Backup_" + Util.timestrNow();
			tryRequestBackupMedium(backupRunID);
			try {
				_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
				initIndices();

				final long startTime = Util.startTimer();
				final ArrayList<ContainerEntry> containerList = containerWalk();
				if(_shouldTerminate) {
					_debug.info("Sicherung wurde abgebrochen");
					return;
				}

				_debug.info("Container für die Sicherung indiziert: " + containerList.size());
				int saved = backupContainerList(containerList, backupRunID);

				final int numContainers = containerList.size();
				_debug.info("Sicherung erfolgreich beendet: " + saved + " von " + numContainers + " Container gesichert (" + Util.stopTimer(startTime) + " sec)");
			}
			catch(PersistenceException | IOException e) {
				throw new BackupException(e);
			}
			finally {
				disposeIndices();
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
		}
	}

	private void tryRequestBackupMedium(final String backupRunID) throws BackupException {
		try {
			requestBackupMedium(_currentBackupMediumID, backupRunID);
			_debug.info("Medium mit MediumId=" + _currentBackupMediumID + " geoeffnet.");
		}
		catch(BackupException e) {
			_debug.warning("Medium mit MediumId=" + _currentBackupMediumID + " nicht verfügbar: " + e.getMessage());
			// Betriebsmeldung: Bitte legen Sie Medium XY ein
			MessageSender.getInstance().sendMessage(
					"",
					MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.WARNING,
					MessageState.MESSAGE,
					"Bitte legen Sie Medium " + _currentBackupMediumID + " ein"
			);
			// wir brechen das Backup ab, der User muss danach das Medium einlegen und den Backup-Lauf neu starten.
			throw e;
		}
	}

	/**
	 * Läuft rekursiv durch ein Verzeichnis und laedt die Container in den Index.
	 *
	 * @return Liste mit den gefundenen Containern
	 *
	 * @throws PersistenceException Fehler in der Persistenz
	 */
	private ArrayList<ContainerEntry> containerWalk() throws PersistenceException {

		AtomicLong numContainers = new AtomicLong();

		DataIdentificationDirWalk walk = DataIdentificationDirWalk.allDirectories(persMgr);

		walk.setStatusPrinter(new StatusPrinter(){
			@Override
			public String getStatusMessage(final String actionName, final Duration runtime, final boolean indexFinished, final long indexedContainerDirs, final long visitedContainerDirs, final long visitedContainerFiles) {
				return super.getStatusMessage(actionName, runtime, indexFinished, indexedContainerDirs, visitedContainerDirs, visitedContainerFiles) +
						"\n" + numContainers + " zu sichernde Container gefunden.";
			}

			@Override
			public String getSuccessMessage(final String actionName, final Duration runtime, final long visitedContainerDirs, final long visitedContainerFiles) {
				return super.getSuccessMessage(actionName, runtime, visitedContainerDirs, visitedContainerFiles) +
						"\n" + numContainers + " zu sichernde Container gefunden.";
			}
		});

		return walk.<ContainerEntry>collect(
				"Ermittlung zu sichernder Container-Dateien",
				archMgr.getNumBackupThreads(),
				(dataIdentificationDir, builder) -> {
					try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(dataIdentificationDir.getDataIdentification())) {
						for(ArchiveDataKind archiveDataKind : dataIdentificationDir.getArchiveDataKinds()) {
							IndexResult<ContainerManagementInformation> headers = persMgr.getContainerHeaders(lock, archiveDataKind);
							for(int i = 0; i < headers.size(); i++) {
								final boolean unsaved = headers.getString(i, ContainerManagementInformation.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED);
								final boolean mustSave = headers.getBoolean(i, ContainerManagementInformation.CHP_TO_SAVE);
								final boolean closed = headers.getSignedInt(i, ContainerManagementInformation.CHP_ANZ_DS) != -1;
								if(unsaved && mustSave && closed) {
									numContainers.incrementAndGet();
									builder.add(
											new ContainerEntry(
													headers.get(i, ContainerManagementInformation.CHP_CONT_ID),
													archiveDataKind,
													dataIdentificationDir
											)
									);
								}
							}
						}
					}
				}
		).sorted().collect(Collectors.toCollection(ArrayList::new));
	}

	/**
	 * Sichert alle angegebenen Container.
	 *
	 * @param containerList Liste mit Containern, welche gesichert werden sollen
	 * @param backupRunID   ID des Backuplaufes
	 *
	 * @throws BackupException Fehler beim Backup
	 */
	private int backupContainerList(final ArrayList<ContainerEntry> containerList, final String backupRunID) throws BackupException {
		int countSavedContainer = 0;

		final ArrayList<ContainerEntry> containersInCurrentMedium = new ArrayList<>();
		final int numContainers = containerList.size();
		for(int i = 0, containerListSize = containerList.size(); i < containerListSize; i++) {
			if(_shouldTerminate) {
				_debug.info("Sicherung wurde abgebrochen. Aktuelles Medium wird abgeschlossen.");
				break;
			}

			final ContainerEntry containerEntry = containerList.get(i);
			// Der Backup-Vorgang ist relativ speicherintensiv, also werden nicht mehr gebrauchte Listeneinträge auf null gesetzt,
			// damit der GC diese nach Bedarf entsorgen kann.
			containerList.set(i, null);

			// Prüfen, ob Archivsystem überlastet und Task deshalb angehalten werden soll
			suspendTaskIfNecessary();

			final DataIdentificationDir containerLocation = containerEntry.getContainerFileDir();
			final Path containerFile = containerLocation.getDataIdentificationDir().resolve(PersistenceManager.getDataKindSuffix(containerEntry.getArchiveDataKind())).resolve(PersistenceManager.getContFileName(containerEntry.getContainerID()));
			if(!Files.exists(containerFile)) {
				_debug.warning("Container-Datei " + containerFile + " existiert nicht, wird nicht gesichert.");
				continue;
			}
			// Synchronisierung über DataIdentNode:
			try(SyncKey<IdDataIdentification> indexLock = archMgr.lockIndex(containerLocation.getDataIdentification())) {
				while(!_backupModul.backupContainer(_currentBackupMediumID, new ContainerBackupImpl(containerEntry, containerFile.toFile()))) {
					closeBackup(containersInCurrentMedium);
					containersInCurrentMedium.clear();
					requestBackupMedium(_currentBackupMediumID, backupRunID);
				}
				containersInCurrentMedium.add(containerEntry);
				countSavedContainer++;
			}
			catch(SynchronizationFailedException e) {
				_debug.warning("Sicherung fehlgeschlagen fuer Datenidentifikation: " + containerLocation);
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Sicherung: %1$d von %2$d Containern gesichert", countSavedContainer, numContainers, -1);
		}
		closeBackup(containersInCurrentMedium);

		return countSavedContainer;
	}

	private void closeBackup(final ArrayList<ContainerEntry> containerList) throws BackupException {
		try {
			try(FileOutputStream fos = new FileOutputStream(_currentIndexFile)) {
				for(final ContainerEntry containerEntry : containerList) {
					fos.write(PersistenceManager.getContFileName(containerEntry.getContainerID()).getBytes(StandardCharsets.ISO_8859_1));
					fos.write('\n');
				}
			}
			_backupModul.closeMediumAfterBackup(_currentBackupMediumID, _currentIndexFile);
			if(containerList.size() > 0) {
				for(ContainerEntry containerEntry : containerList) {
					markContainerAsSaved(containerEntry);
				}
				// Medium-ID nur hochzählen, falls mindestens 1 Container gesichert wurde:
				increaseMediumID();
			}
			clearCurrentIndexFile();
		}
		catch(IOException | PersistenceException | SynchronizationFailedException | IndexException e) {
			throw new BackupException(e);
		}
	}

	private void markContainerAsSaved(final ContainerEntry containerEntry) throws PersistenceException, SynchronizationFailedException, IndexException {
		DataIdentificationDir dataIdentificationDir = containerEntry.getContainerFileDir();
		IdDataIdentification dataIdentification = dataIdentificationDir.getDataIdentification();
		try(SyncKey<IdDataIdentification> indexLock = archMgr.lockIndex(dataIdentification)) {
			ContainerHeaders headers = persMgr.getContainerHeaders(indexLock, containerEntry.getArchiveDataKind(), containerEntry.getContainerID());
			headers.setContainerHeaderParam(ContainerManagementInformation.CHP_MEDIUM_ID, formatMediumId(_currentBackupMediumID));
			IdContainerFileDir idContainerFileDir = new IdContainerFileDir(dataIdentification, containerEntry.getArchiveDataKind());
			persMgr.updateContainerIndex(indexLock, headers);
			// Hier wird nicht nur die Indexdatei aktualisiert sondern auch der ContainerHeader auf den selben Stand gebracht (ist so in NÄM-73 
			// gefordert)
			try (ContainerFileHandle containerFileHandle = persMgr.accessContainer(idContainerFileDir, containerEntry.getContainerID())) {
				PersistenceManager.copyHeaders(headers, containerFileHandle);
				containerFileHandle.writeContainerHeader(indexLock);
			}
		}
	}

	private static String formatMediumId(final int mediumID) {
		return Util.leadZero(mediumID, ContainerHdr.CHP_MEDIUM_ID.getValLen());
	}

	private void clearCurrentIndexFile() throws BackupException {
		try {
			final boolean deleted = _currentIndexFile.delete();
			if(!deleted) {
				_debug.warning("Indexdatei existiert nicht: " + _currentIndexFile);
			}
			final boolean created = _currentIndexFile.createNewFile();
			if(!created) {
				throw new BackupException("Konnte neue Indexdatei nicht erstellen: " + _currentIndexFile);
			}
		}
		catch(IOException e) {
			throw new BackupException("Konnte neue Indexdatei nicht erstellen: " + _currentIndexFile, e);
		}
	}

	/**
	 * Mit dieser Methode wird der Operator aufgefordert, ein bestimmtes Medium einzulegen. Die Methode blockiert bis der Operator das Einlegen bestätigt oder
	 * abgebrochen hat.
	 *
	 * @param backupMediumID MediumID
	 * @param backupRunID    ID des Backupvorgangs
	 *
	 * @throws BackupException Medium nicht verfügbar
	 */
	private static void requestBackupMedium(final int backupMediumID, final String backupRunID) throws BackupException {
		int numOfTry = 0;
		while(true) {
			try {
				_backupModul.openMediumBeforeBackup(backupMediumID, backupRunID);
				break;
			}
			catch(MediumNotAccessibleException mex) {
				if(++numOfTry > 3) throw new BackupException("Fehler bei Sicherung: Medium " + backupMediumID + " nicht verfuegbar.", mex);
			}
		}
	}


	/**
	 * Erstellt das temporäre Verzeichnis für die Indexdatei und erstellt eine leere Indexdatei
	 *
	 * @throws IOException IO-Fehler beim Vorgang
	 */
	private void initIndices() throws IOException {
		// Da mehrere Instanzen des Archivsystems nicht die gleichen temporären Dateien verwenden sollen und einige Backupmodule davon ausgehen, dass die
		// Indexdatei den Namen "index.dat" hat, wird an dieser Stelle ein temporäres Verzeichnis mit einem eindeutigen Namen erstellt.
		// Da die Klasse File keine createTempDir-Methode hat wird mit createTempFile eine Datei erzeugt und dann im gleichen Verzeichnis ein Verzeichnis
		// erzeugt, dass den Namen der temporären Datei erweitert um den Text "_dir" trägt.
		boolean createdDir = false;
		while(!createdDir) {
			_tmpFile = File.createTempFile("ArsBackup", "");
			_indexFileDir = new File(_tmpFile.getAbsolutePath() + "_dir");
			if(_indexFileDir.mkdir()) {
				createdDir = true;
			}
			else {
				deleteFile(_tmpFile);
			}
		}

		_currentIndexFile = new File(_indexFileDir, INDEXFILE_NAME);

		final boolean created = _currentIndexFile.createNewFile();
		if(!created) {
			throw new IOException("Konnte neue Indexdatei nicht erstellen: " + _currentIndexFile);
		}
	}

	/** Löscht die Indexdateien und sonstige temporäre Dateien und Verzeichnisse */
	private void disposeIndices() {
		if(_currentIndexFile != null && _currentIndexFile.exists()) {
			deleteFile(_currentIndexFile);
		}

		if(_indexFileDir != null && _indexFileDir.exists()) {
			deleteFile(_indexFileDir);
		}

		if(_tmpFile != null && _tmpFile.exists()) {
			deleteFile(_tmpFile);
		}
	}

	/**
	 * Löscht eine Datei un gibt eine Warnung aus, wenn das Löschen fehlschlägt
	 *
	 * @param file Datei die gelöscht werden soll
	 */
	private static void deleteFile(final File file) {
		final boolean deleted = file.delete();
		if(!deleted) _debug.warning("Konnte Datei nicht löschen: " + file);
	}

	/** Erhöht die MedienID um 1 */
	private void increaseMediumID() {
		_currentBackupMediumID++;
	}

	/**
	 * Setzt das zu verwendende Backup-Modul.
	 *
	 * @param propertiesFile Datei in dem das zu verwendende Modul gespeichert ist.
	 *
	 * @throws BackupException Fehler beim Vorgang
	 */
	public static void setBackupModul(final String propertiesFile) throws BackupException {
		// Wenn keine Properties gesetzt sind, Archiv-System trotzdem starten.
		if(propertiesFile == null || propertiesFile.length() == 0) {
			return;
		}

		final Properties properties = new Properties();
		try {
			try(FileInputStream fis = new FileInputStream(new File(propertiesFile))) {
				properties.load(fis);
			}
		}
		catch(IOException ioe) {
			throw new BackupException("Fehler beim Einlesen der BackupProperties-Datei: ", ioe);
		}

		final String moduleName = properties.getProperty("backupModul");
		if(moduleName == null || moduleName.length() == 0) {
			throw new BackupException("Die BackupProperties-Datei '" + propertiesFile + "' enthält keinen Eintrag 'backupModul'");
		}

		final Class<?> backupClass;
		try {
			backupClass = Class.forName(moduleName);
		}
		catch(ClassNotFoundException e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht gefunden werden. ", e);
		}

		Object instance;
		try {
			instance = backupClass.getDeclaredConstructor().newInstance();
		}
		catch(InstantiationException | IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht instantiiert werden. ", e);
		}

		try {
			initializeBackupModule(properties, instance);
		}
		catch(Exception e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht initialisiert werden. ", e);
		}


	}

	/**
	 * Initialisiert das Backupp-Modul. Wurde eien alte Klasse angegeben, wird diese in einen {@link BackupAdapter} verpackt.
	 * @param properties Properties aus der Datei
	 * @param instance   Instanz der angeforderten Klasse
	 * @throws Exception Exception
	 */
	@SuppressWarnings("deprecation")
	private static void initializeBackupModule(final Properties properties, final Object instance) throws Exception {
		if(instance instanceof BackupModule) {
			BackupModule backupModule = (BackupModule) instance;
			_backupModul = backupModule.getInstance(properties);
		}
		else if(instance instanceof BackupModul) {
			BackupModul backupModul = (BackupModul) instance;
			backupModul.initialize(properties);
			_backupModul = new BackupAdapter(backupModul);
		}
		else {
			throw new BackupException("Das BackupModul ist nicht vom korrekten Typ de.bsvrz.ars.ars.backup.BackupModule oder de.bsvrz.ars.ars.backup.BackupModul: " + instance.getClass());
		}
	}

	/**
	 * Gibt das verwendete Backupmodul zurück
	 *
	 * @return das verwendete Backupmodul
	 */
	public static BackupImplementation getBackupModul() {
		return _backupModul;
	}

	/**
	 * Gibt die zuletzt verwendete MedienID zurück
	 *
	 * @return zuletzt verwendete MedienID
	 */
	public int getLastUsedMediumID() {
		// dadurch, dass currentBackupMediumID bei closeBackup() immer hochgezählt wird,
		// muss die letzte benutzte Medien-ID immer um eins dekrementiert werden.
		return _currentBackupMediumID - 1;
	}

	/**
	 * Setzt die zuletzt verwendete MedienID. Für nach nächste Backup wird der um 1 inkrementierte Wert verwendet
	 *
	 * @param mediumID zuletzt verwendete MedienID
	 */
	public void setLastUsedMediumID(final int mediumID) {
		// dadurch, dass currentBackupMediumID immer die nächste zu verwendende Medien-ID darstellt,
		// muss die letzte benutzte Medien-ID um eins inkrementiert werden.
		_currentBackupMediumID = mediumID + 1;
	}

	/**
	 * Immutable Klasse, die eine ContainerID mit der zugehörigen ContainerFileDir, also dem Speicherort (der Datenidentifikation) verknüpft. Aus
	 * Speicherplatzgründen können mehrere ContainerIDs mit derselben ContainerFileDir-Instanz verknüpft werden
	 */
	private static final class ContainerEntry implements Comparable<ContainerEntry> {

		private final long _containerID;

		private final ArchiveDataKind _archiveDataKind;
		
		private final DataIdentificationDir _containerLocation;

		public ContainerEntry(final long containerID, final ArchiveDataKind archiveDataKind, final DataIdentificationDir containerLocation) {
			this._containerID = containerID;
			_archiveDataKind = archiveDataKind;
			this._containerLocation = containerLocation;
		}

		@Override
		public int compareTo(final ContainerEntry o) {
			return Long.compare(_containerID, o.getContainerID());
		}

		@Override
		public int hashCode() {
			return ((Long)_containerID).hashCode();
		}

		@Override
		public boolean equals(final Object obj) {
			if(null == obj)	return false;
			if(obj instanceof ContainerEntry) {
				final ContainerEntry containerEntry = (ContainerEntry)obj;
				return containerEntry.getContainerID() == _containerID;
			}
			return false;
		}

		public long getContainerID() {
			return _containerID;
		}

		public DataIdentificationDir getContainerFileDir() {
			return _containerLocation;
		}

		public ArchiveDataKind getArchiveDataKind() {
			return _archiveDataKind;
		}

		@Override
		public String toString() {
			return "ContainerEntry{" + "_containerID=" + _containerID + ", _containerLocation=" + _containerLocation + '}';
		}
	}

	/**
	 * Gibt einen {@link BackupResultData}-Dummy zurück
	 *
	 * @return eine leere BackupResultData-Klasse die von ResultData erbt.
	 */
	public static BackupResultData getBackupResultData() {
		return BackupResultDataHolder._backupResultData;
	}

	/** Holder-Objekt, dass eine statische BackupResultData-Instanz (threadsicher) bereitstellt */
	private static class BackupResultDataHolder {

		private static final BackupResultData _backupResultData = new BackupResultData();
	}

	/**
	 * Diese Klasse sieht aus wie ein ResultData, wird aber nur intern verwendet, um einen über den Scheduler angestoßenen Sicherungslauf von einem extern
	 * angestoßenen zu unterscheiden
	 *
	 * @author beck et al. projects GmbH
	 * @author Phil Schrettenbrunner
	 */
	public static class BackupResultData extends ResultData {

		/** 
		 * Erstellt einen neuen BackupResultData
		 */
		public BackupResultData() {
			super(null, null, 0, null);
		}
	}

	private static class ContainerBackupImpl implements Container {
		private final ContainerEntry _containerEntry;
		private final File _containerFile;
		private static final ThreadLocal<byte[]> BUFFER = ThreadLocal.withInitial(() -> new byte[2048]);

		public ContainerBackupImpl(final ContainerEntry containerEntry, final File containerFile) {
			_containerEntry = containerEntry;
			_containerFile = containerFile;
		}

		@Override
		public void backup(final OutputStream stream, final int mediumId) throws IOException {
			// Wir wollen den übergebenen OutputStream nicht schließen.
			@SuppressWarnings("IOResourceOpenedButNotSafelyClosed")
			DataOutputStream output = new DataOutputStream(stream);
			try(DataInputStream inputStream = new DataInputStream(new BufferedInputStream(Files.newInputStream(_containerFile.toPath())))) {
				byte[] buf = ContainerBackupImpl.BUFFER.get();
				inputStream.readFully(buf, 0, 4);
				stream.write(buf, 0, 4);
				int headerLength = inputStream.readInt();

				if(buf.length < headerLength || headerLength <= 0) {
					throw new IOException("Falsche Headergröße: " + headerLength);
				}

				inputStream.readFully(buf, 0, headerLength);
				try {
					ContainerHdr containerHdr = new ContainerHdr();
					containerHdr.readContainerHdr(buf, headerLength);
					containerHdr.setVal(ContainerHdr.CHP_MEDIUM_ID, formatMediumId(mediumId));
					String s = containerHdr.writeContainerHdr();
					output.writeInt(s.length() + 4);
					output.write(s.getBytes(StandardCharsets.ISO_8859_1));
				}
				catch(PersistenceException e) {
					throw new IOException(e);
				}
				output.write(ByteIO.SEPARATOR);
				while(true) {
					int read = inputStream.read(buf);
					if(read < 0) {
						break;
					}
					output.write(buf, 0, read);
				}
			}
			output.flush();
		}

		@Override
		public ArchiveDataKind getArchiveDataKind() {
			return _containerEntry.getArchiveDataKind();
		}

		@Override
		public IdDataIdentification getDataIdentification() {
			return _containerEntry.getContainerFileDir().getDataIdentification();
		}

		@Override
		public long getContainerId() {
			return _containerEntry.getContainerID();
		}

		@Override
		public long getFileSize() {
			return _containerFile.length();
		}

		@Override
		public long lastModified() {
			return _containerFile.lastModified();
		}
	}
}
