/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.ContainerHeaders;
import de.bsvrz.ars.ars.persistence.ContainerManagementInformation;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.DeletedContainerFile;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.walk.ContainerAction;
import de.bsvrz.ars.ars.persistence.walk.ContainerWalk;
import de.bsvrz.ars.ars.persistence.walk.internal.StatusPrinter;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * Task zur Bearbeitung von Loeschzeit-Verlängerungsauftraegen.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class LZVTask extends SingleClientSenderTask {

	private static final long DELETE_IMMEDIATELY = -1;
	private static final long DELETE_LATER = 0;
	/** Um wieviele Monate maximal der Vorhaltezeitraum angepasst wird. */
	private static final int MAX_TIMESHIFTING = 2 * 12; //
	/** Nur für Testzwecke: Löschschutz ignorieren */
	public static final boolean IGNORE_LOESCH_SCHUTZ = false;
	private static long MIN_REM_DSK_CAPA = 2 * 1000 * 1000 * 1000L; // Default 2GB gemäß DatK
	private final LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;
	private final AtomicInteger _numOfDeletedCont = new AtomicInteger();
	private final int _numThreads;

	/**
	 * Erzeugt eine neue Loeschzeit-Verlängerungs-Task.
	 *
	 * @param longTermTaskStatePublisher LongTermTaskStatePublisher
	 * @param numThreads Anzahl Threads, die gleichzeitig das Löschen durchführen.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange.
	 */
	public LZVTask(ArchiveManager archiveMgr, BlockingQueue<Object> inputData, final LongTermTaskStatePublisher longTermTaskStatePublisher, final int numThreads) {
		super(archiveMgr, inputData);
		setName("LZV-Task");
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Löschen");
		_numThreads = numThreads;
	}

	/**
	 * Liefert die Mindest-Restkapazitaet der Festplatte.
	 *
	 * @return Mindestrestkapazitaet in Byte
	 */
	public static long getMinRemainingDiskCapa() {
		return MIN_REM_DSK_CAPA;
	}

	/**
	 * Setzt die Mindest-Restkapazitaet der Festplatte.
	 *
	 * @param minRemainingDiskCapa Neue Mindestrestkapazitaet in Byte.
	 */
	public static void setMinRemainingDiskCapa(long minRemainingDiskCapa) {
		MIN_REM_DSK_CAPA = minRemainingDiskCapa;
	}

	/**
	 * Gibt einen Dummy-Datensatz zurück, der einen Kapazitätscheck (und damit ggf. spontanes Löschen) auslöst
	 * @return Dummy-Datensatz
	 */
	public static CapacityCheckResultData getCapacityCheckResultData() {
		return new CapacityCheckResultData();
	}

	/**
	 * Gibt einen Dummy-Datensatz zurück, der Reguläres (periodisches) Löschen auslöst
	 * @return Dummy-Datensatz
	 */
	public static DeleteRegularResultData getDeleteRegularResultData() {
		return new DeleteRegularResultData();
	}

	/**
	 * Gibt einen Dummy-Datensatz zurück, der endgültiges Löschen auslöst
	 * @return Dummy-Datensatz
	 */
	public static DeletePermanentlyResultData getDeletePermanentlyResultData() {
		return new DeletePermanentlyResultData();
	}

	@Override
	public void execute() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.execute();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	/**
	 * Führt das Löschen, spontanes Löschen oder die Löschzeitverlängerung aus.
	 */
	@Override
	protected void work(ResultData resultData) {
		try {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);
			init(resultData.getData());
			executeAction(resultData);
			if(resultData.hasData()) sendSuccessResponse();
		}
		catch(Exception e) {
			_debug.error("Loeschzeitverlaengerung fehlgeschlagen", e);
			if(resultData.hasData()) sendErrorResponse("Loeschzeitverlaengerung fehlgeschlagen: " + e.getMessage());
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}

	/**
	 * Prüft, welche Operation (Löschen, spontanes Löschen oder die Löschzeitverlängerung) ausgeführt werden soll, und führt diese dann durch.
	 *
	 * @param resultData Datensatz mit Informationen über den Auftrag
	 * @throws Exception Fehler
	 */
	private void executeAction(ResultData resultData) throws Exception {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
		try {
			if(resultData instanceof CapacityCheckResultData) {
				deleteSpontan(new ArchiveDataSpecification[0]);
			}
			else if(resultData instanceof DeleteRegularResultData) {
				// Reguläres Löschen mit Betrachtung des Löschschutzes
				containerWalk(new ArchiveDataSpecification[0], new DeleteRegular());
			}
			else if(resultData instanceof DeletePermanentlyResultData) {
				// Endgültiges Löschen mit Betrachtung des Löschschutzes
				containerWalk(new ArchiveDataSpecification[0], new DeletePermanently());
			}
			else {
				// Externer Auftrag
				if(!resultData.hasData()) return; // Leere Datensätze haben keine changeTime übermittelt

				long changeTime;

				// nur externe Aufträge werden geparst
				changeTime = deserializer.readLong();
				ArchiveDataSpecification[] ads = parseArchiveDataSpec(deserializer);

				if(changeTime == DELETE_IMMEDIATELY) {
					// Sofortiges Löschen der angegebenen Daten ohne Betrachtung des Löschschutzes
					containerWalk(ads, new DeleteNow());
				}
				else if(changeTime == DELETE_LATER) {
					// Löschschutzaufhebung (späteres Löschen) der angegebenen Daten
					containerWalk(ads, new DeleteLater());
				}
				else if(changeTime > 0) {
					// Manuelle Löschschutzverlängerung
					containerWalk(ads, new ChangeTime(changeTime));
				}
				else {
					_debug.warning(getName() + ": Unbekannte Loeschzeit: " + changeTime);
					throw new IllegalArgumentException(getName() + ": Loeschzeitverlaengerung: Unbekannte Loeschzeit: " + changeTime);
				}
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
		}
	}

	/**
	 * Der {@code DeleteSpontanTask} löscht bei knapp werdendem Speicherplatz unwichtige Container.
	 * <p>
	 * Dazu überprüft er in einem parametrierbaren Zyklus, wieviel Speicherplatz dem Archivsystem noch zur Verfuegung steht. Erreicht die vorhandene Kapazität
	 * einen Schwellwert, werden folgende Daten geloescht: <ul> <li>archivierte Daten, deren Loeschschutz abgelaufen ist und die nicht (mehr) gesichert werden
	 * müssen,</li> <li>wiederhergestellte Daten,</li> <li>archivierte Simulationsdaten,</li> <li>archivierte Daten, deren Loeschschutz nicht abgelaufen ist,
	 * die aber schon gesichert wurden.</li> </ul>
	 */
	public void deleteSpontan(ArchiveDataSpecification[] ads) throws Exception {
		long freeDiskSpace = getFreeDiskSpace();
		Calendar compareDate = Calendar.getInstance();
		boolean someDataDeleted = false;

		for(DeleteLevel level : DeleteLevel.values()) {
			if(freeDiskSpace < getMinRemainingDiskCapa()) {
				someDataDeleted = true;
				if(level == DeleteLevel.TIME_SHIFTING) {
					// Verschieben des Vorhaltezeitraums
					for(int i = 0; i < MAX_TIMESHIFTING; i++) {
						compareDate.add(Calendar.MONTH, 1);
						if(freeDiskSpace < getMinRemainingDiskCapa()) {
							long t0 = System.currentTimeMillis();
							_debug.info(
									"Spontanes Löschen, mit " + (i + 1) + " Monaten Verkürzung des Vorhaltezeitraums: "
											+ "Verbleibender Plattenplatz in Bytes: " + freeDiskSpace
							);
							containerWalk(ads, new DeleteSpontan(compareDate, DeleteLevel.TIME_SHIFTING));
							freeDiskSpace = getFreeDiskSpace();
							_debug.info(
									"Nach spontanem Löschen, mit " + (i + 1) + " Monaten Verkürzung des Vorhaltezeitraums: "
											+ "Verbleibender Plattenplatz in Bytes: " + freeDiskSpace + ", " + "Dauer: "
											+ ((System.currentTimeMillis() - t0) / 1000) + " Sekunden"
							);
						}
						else {
							MessageSender.getInstance().sendMessage(
									MessageType.APPLICATION_DOMAIN,
									MessageGrade.INFORMATION,
									"Das spontane Löschen hat Daten gelöscht"
							);
							_debug.info("Das spontane Löschen hat Daten gelöscht");
							return;
						}
					}
				}
				else {
					// Löschung nach den 4 Kriterien in der TAnf 5.1.2.7
					long t0 = System.currentTimeMillis();
					_debug.info("Spontanes Löschen, Level " + level + ": Verbleibender Plattenplatz in Bytes: " + freeDiskSpace);
					containerWalk(ads, new DeleteSpontan(compareDate, level));
					freeDiskSpace = getFreeDiskSpace();
					_debug.info(
							"Nach spontanem Löschen, Level " + level + ": Verbleibender Plattenplatz in Bytes: " + freeDiskSpace + ", " + "Dauer: "
									+ ((System.currentTimeMillis() - t0) / 1000) + " Sekunden"
					);
				}
			}
			else {
				if(someDataDeleted) {
					MessageSender.getInstance().sendMessage(
							de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
							MessageGrade.INFORMATION,
							"Das spontane Löschen hat Daten gelöscht"
					);
					_debug.info("Das spontane Löschen hat Daten gelöscht");
				}
				return; // Es wurde genug PLatz freigegeben -> fertig
			}
		}

		// Wenn jetzt noch immer nicht genug Platz frei ist, belegt u.U. die Verwaltungsinfo
		// alleine schon so viel, dass wir durch Löschen keinen Platz mehr bekommen
		MessageSender.getInstance().sendMessage(
				de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
				MessageGrade.WARNING,
				"Achtung: Der Speicherplatz geht zu Ende! Durch spontanes Löschen konnte nicht so viel Platz freigegeben werden, wie laut Parametrierung mindestens frei sein soll."
		);
		_debug.warning(
				"Achtung: Der Speicherplatz geht zu Ende! Durch spontanes Löschen konnte nicht so viel Platz freigegeben werden, wie laut Parametrierung mindestens frei sein soll."
		);
	}

	/**
	 * Gibt den freien Speicherplatz zurück und veröffentlicht ihn ebenfalls über den Datenverteiler (falls möglich)
	 * @return frier Speicherpaltz in Bytes
	 * @throws Exception
	 */
	public long getFreeDiskSpace() throws Exception {
		long freeDiskSpace = Util.calcFreeDiskSpace(persMgr.getArchivePath().toString());
		archMgr.getInQueuesMgr().setFreeDiskSpace(freeDiskSpace);
		return freeDiskSpace;
	}

	/**
	 * Die generelle Container-Walk-Methode.
	 * <p>
	 * Je nachdem, ob die ArchiveDataSpec gefuellt ist oder nicht, wird ein ContainerWalk über das komplette Persistenz-Verzeichnis (containerWalkAllFromDir)
	 * oder alle spezifizierten Container (containerWalkAllByDataSpec) ausgeführt.<br><br>
	 * <p>
	 * Achtung!<br>Die Methode löscht je nach übergebenen Parametern Archivdaten!
	 *
	 * @param ads
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	public void containerWalk(ArchiveDataSpecification[] ads, ContainerHandler handler) throws Exception {
		_numOfDeletedCont.set(0);
		if(ads.length == 0) {
			doContainerWalk(handler, ContainerWalk.allContainers(persMgr));
		}
		else {
			doContainerWalk(handler, ContainerWalk.fromArchiveDataSpecification(persMgr, Arrays.asList(ads)));
		}
	}

	private void doContainerWalk(final ContainerHandler handler, final ContainerWalk containerWalk) throws PersistenceException {
		containerWalk.setStatusPrinter(new StatusPrinter(){
			@Override
			public String getStatusMessage(final String actionName, final Duration runtime, final boolean indexFinished, final long indexedContainerDirs, final long visitedContainerDirs, final long visitedContainerFiles) {
				return super.getStatusMessage(actionName, runtime, indexFinished, indexedContainerDirs, visitedContainerDirs, visitedContainerFiles)
						+" " + _numOfDeletedCont.get() + " Container bearbeitet.";
			}

			@Override
			public String getSuccessMessage(final String actionName, final Duration runtime, final long visitedContainerDirs, final long visitedContainerFiles) {
				return super.getSuccessMessage(actionName, runtime, visitedContainerDirs, visitedContainerFiles)
						+" " + _numOfDeletedCont.get() + " Container bearbeitet.";
			}
		});
		containerWalk.execute(handler.getActionName(), _numThreads, new ContainerAction() {
			@Override
			public void preVisitDirectory(final DataIdentificationDir dataIdentificationDir) {
				if(shouldTerminate()) containerWalk.terminate();

				//Prüfen, ob Archivsystem überlastet und Task deshalb angehalten werden soll
				//Spontanes Löschen darf aber nicht angehalten werden, sondern nur Löschzeitverlängerung und (nicht-spontanes) Löschen.
				if(handler.maySuspend()) LZVTask.this.suspendTaskIfNecessary();		
			}

			@Override
			public void run(final DataIdentificationDir containerFileDir, final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind, final Collection<ContainerHeaders> headers) throws Exception {
				// Auf Container zugreifen und bearbeiten:
				List<ContainerHeaders> containersToHandle = headers.stream()
						.filter(it -> handler.appliesTo(containerFileDir.getDataIdentification(), it))
						.collect(Collectors.toList());

				if(!containersToHandle.isEmpty()) {
					handler.handleContainers(lock, archiveDataKind, containersToHandle);
				}
			}
		});
	}

	/**
	 * Prüft, ob das ContainerFile geloescht werden muss. Folgende Faktoren werden betrachtet: <ul> <li>archivierte Daten, deren Loeschschutz abgelaufen ist
	 * und die nicht (mehr) gesichert werden müssen,</li> <li>wiederhergestellte Daten,</li> <li>archivierte Simulationsdaten,</li> <li>archivierte Daten,
	 * deren Loeschschutz nicht abgelaufen ist, die aber schon gesichert wurden.</li> </ul>
	 *
	 * @param simVar
	 * @param compareDate
	 * @return boolean
	 */
	private boolean hasToBeDeletedSpontan(final int simVar, Calendar compareDate, DeleteLevel level, final ContainerHeaders header) {
		try {
			// Aktueller, offener Container wird nicht gelöscht
			if(header.getContainerHeaderParamAsInt(ContainerManagementInformation.CHP_ANZ_DS) == -1) return false;

			// schon gelöschte Container nicht nochmal löschen
			if(header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) return false;

			// Kein level gesetzt -> keine weiteren Checks zum Löschen
			if(level == null) return false;

			switch(level) {
				case SAVED_EXPIRED: // Fallthrough
				case TIME_SHIFTING:
					// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelöscht
					if(header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_TO_SAVE)
							&& header.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED)) {
						return false;
					}
					// TAnf 5.1.2.7 erster Absatz: (gesichert || nicht_zu_sichern) && Vorhaltezeitraum_abgelaufen &&  Löschschutz_abgelaufen
					long chpLoeschutz = persMgr.getProtectedDeletionTime(header);

					long delAt = persMgr.getRegularDeletionTime(header);
					long compareDateTimeInMillis = compareDate.getTimeInMillis();
					return delAt < compareDateTimeInMillis && chpLoeschutz < compareDateTimeInMillis;
				case RESTORED:
					return header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_RESTORED);
				case SIMULATION:
					return simVar != 0;
				case SAVED_NOTEXPIRED:
					return !header.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED);
			}
		}
		catch(Exception e) {
			_debug.warning("Problem bei der Überprüfung, ob ein Container spontan gelöscht werden kann", e);
		}

		return false;
	}

	/**
	 * Prüft, ob das ContainerFile geloescht werden muss.
	 *
	 * @return boolean
	 */
	private boolean hasToBeDeletedRegular(final ContainerHeaders header) {
		try {
			// Aktueller, offener Container wird nicht gelöscht
			if(header.getContainerHeaderParamAsInt(ContainerManagementInformation.CHP_ANZ_DS) == -1) return false;

			// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelöscht
			if(header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_TO_SAVE) && header.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID
			).equals(ContainerHdr.MEDIUM_UNSAVED)) {
				return false;
			}

			// schon gelöschte Container nicht nochmal löschen
			if(header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) return false;

			if(IGNORE_LOESCH_SCHUTZ) return true;

			if(!header.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED)
					|| !header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_TO_SAVE)) {

				// TAnf 5.1.2.7 erster Absatz: (gesichert || nicht_zu_sichern) && Vorhaltezeitraum_abgelaufen && Löschschutz_abgelaufen
				long chpLoeschutz = persMgr.getProtectedDeletionTime(header);
				long now = System.currentTimeMillis();
				long delAt = persMgr.getRegularDeletionTime(header);
				if(delAt < now && chpLoeschutz < now) return true;
			}
		}
		catch(Exception e) {
			_debug.warning("Problem bei der Überprüfung, ob ein Container regulär gelöscht werden kann", e);
		}

		// Sonst nichts löschen
		return false;
	}

	/**
	 * Prüft, ob das ContainerFile beim manuellen Löschen gelöscht werden kann.
	 *
	 * @return boolean
	 */
	private boolean hasToBeDeletedExplicit(final ContainerHeaders header) {
		try {
			// Aktueller, offener Container wird nicht gelöscht
			if(header.getContainerHeaderParamAsInt(ContainerManagementInformation.CHP_ANZ_DS) == -1) return false;

			// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelöscht
			boolean toSafe = header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_TO_SAVE);
			boolean isUnsaved = header.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED);
			if(toSafe && isUnsaved) {
				return false;
			}

			// schon gelöschte Container nicht nochmal löschen
			if(header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) return false;

			long now = System.currentTimeMillis();

			long delAt = persMgr.getRegularDeletionTime(header);
			if(delAt < now) return true;

		}
		catch(Exception e) {
			_debug.warning("Problem bei der Überprüfung, ob ein Container manuell gelöscht werden kann", e);
		}

		// Sonst nichts löschen
		return false;
	}

	/**
	 * Für spontantes Löschen: 1. Vorhaltezeitraum abgelaufen, und schon gesichert 2. wiederhergestellte Daten 3. Simulationsdaten 4. Vorhaltezeitraum noch
	 * nicht abgelaufen, aber schon gesichert 5. Löschen durch schrittweises Verschieben des Vorhaltezeitraums
	 * <p>
	 * Die Reihenfolge der Deklaration ist von Bedeutung!
	 */
	private enum DeleteLevel {

		SAVED_EXPIRED("1 (Vorhaltezeitraum abgelaufen, und schon gesichert)"),
		RESTORED("2 (wiederhergestellte Daten)"),
		SIMULATION("3 (Simulationsdaten)"),
		SAVED_NOTEXPIRED("4 (Vorhaltezeitraum noch nicht abgelaufen, aber schon gesichert)"),
		TIME_SHIFTING("5 (Löschen durch schrittweises Verschieben des Vorhaltezeitraums)");

		private final String _name;

		DeleteLevel(final String name) {
			_name = name;
		}


		public String toString() {
			return _name;
		}
	}

	/**
	 * Interface, das angibt, was mit Containern gemacht werden soll, implementiert die Verhaltensunterschiede von spontanem Löschen,
	 * manuellem Löschen, regulärem regelmäßigem Löschen, Löschzeitverlängerung usw.
	 */
	private interface ContainerHandler {

		/**
		 * Bearbeitet einen Container
		 *
		 * @param lock
		 * @param dataKind Datenart
		 * @param headers Header von Containern
		 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
		 */
		void handleContainers(SyncKey<IdDataIdentification> lock, ArchiveDataKind dataKind, Collection<ContainerHeaders> headers) throws PersistenceException, SynchronizationFailedException, IndexException;

		/**
		 * Darf der Task angehalten werden?
		 * @return Darf der Task angehalten werden?
		 */
		boolean maySuspend();

		/**
		 * Gibt an, ob der Container aus dem übergebenen Index behandelt (gelöscht/lz-verlängert) werden soll
		 * @param dataIdentification Datenidentifikation
		 * @param headers Header von einem Container
		 * @return true: Container soll durch {@link #handleContainers(SyncKey, ArchiveDataKind, Collection)} bearbeitet werden, sonst false
		 */
		boolean appliesTo(IdDataIdentification dataIdentification, ContainerHeaders headers);

		/**
		 * Gibt den Anzeigenamen der Aktion zurück
		 * @return Anzeigenamen der Aktion
		 */
		String getActionName();
	}

	private static class CapacityCheckResultData extends ResultData {

		public CapacityCheckResultData() {
			super(null, null, 0, null);
		}
	}

	private static class DeleteRegularResultData extends ResultData {

		public DeleteRegularResultData() {
			super(null, null, 0, null);
		}
	}

	private static class DeletePermanentlyResultData extends ResultData {

		public DeletePermanentlyResultData() {
			super(null, null, 0, null);
		}
	}

	/**
	 * Sofortiges Löschen auf Anfrage. Es werden alle Container gelöscht, die nicht zu sichern sind oder schon gesichert wurden.
	 */
	public class DeleteNow implements ContainerHandler {

		@Override
		public void handleContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind, final Collection<ContainerHeaders> headers) throws PersistenceException, IndexException {
			deleteContainers(lock, dataKind, headers);
		}

		@Override
		public boolean maySuspend() {
			return true;
		}

		@Override
		public boolean appliesTo(final IdDataIdentification dataIdentification, final ContainerHeaders headers) {
			_longTermTaskStatePublisherTask.setTaskProgress("Sofortiges Löschen: %1$d Container gelöscht", _numOfDeletedCont.get(), -1, -1);
			return hasToBeDeletedExplicit(headers);
		}


		@Override
		public String getActionName() {
			return "Sofortiges Löschen";
		}
	}

	/**
	 * Späteres Löschen, wie DeleteNow, nur wird nur der Löschschutz aufgehoben statt den Container zu löschen
	 */
	public class DeleteLater implements ContainerHandler {

		@Override
		public void handleContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind, final Collection<ContainerHeaders> headers) throws PersistenceException, IndexException {
			for(ContainerHeaders header : headers) {
				long regularDeleteTime = header.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_LOESCHEN);
				header.setContainerHeaderParam(ContainerManagementInformation.CHP_LOESCHUTZ, regularDeleteTime);
				persMgr.updateContainerIndex(lock, header);
				_numOfDeletedCont.getAndIncrement();
			}
		}

		@Override
		public boolean maySuspend() {
			return true;
		}

		@Override
		public boolean appliesTo(final IdDataIdentification dataIdentification, final ContainerHeaders headers) {
			_longTermTaskStatePublisherTask.setTaskProgress("Späteres Löschen: %1$d Container markiert", _numOfDeletedCont.get(), -1, -1);
			return hasToBeDeletedExplicit(headers);
		}



		@Override
		public String getActionName() {
			return "Späteres Löschen";
		}
	}

	/**
	 * Löscht einen Container
	 * @param lock Datenidentifikation
	 * @param dataKind Datenart
	 * @throws PersistenceException Fehler beim Löschen
	 */
	private void deleteContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind, Collection<ContainerHeaders> headers) throws PersistenceException, IndexException {
	    IdContainerFileDir containerFileDir = new IdContainerFileDir(lock.getElement(), dataKind);
	    try {
		    long[] containerIds = headers.stream()
				    .mapToLong(it -> it.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_CONT_ID)).toArray();

		    persMgr.deleteContainers(lock, containerFileDir, containerIds);

			_numOfDeletedCont.addAndGet(containerIds.length);
		}
		catch(SynchronizationFailedException e) {
			_debug.warning("Synchronisierung fehlgeschlagen auf " + lock.getElement());
		}
	}

	/**
	 * Reguläres Löschen. Es werden alle Container gelöscht, die nicht zu sichern sind oder die schon gesichert wurden, aber nur sofern kein Löschschutz
	 * besteht.
	 */
	public class DeleteRegular implements ContainerHandler {
		@Override
		public void handleContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind, final Collection<ContainerHeaders> headers) throws PersistenceException, IndexException {
			deleteContainers(lock, dataKind, headers);	
		}

		@Override
		public boolean maySuspend() {
			return true;
		}

		@Override
		public boolean appliesTo(final IdDataIdentification dataIdentification, final ContainerHeaders headers) {
			_longTermTaskStatePublisherTask.setTaskProgress("Reguläres Löschen: %1$d Container gelöscht", _numOfDeletedCont.get(), -1, -1);
			return hasToBeDeletedRegular(headers);	
		}

		@Override
		public String getActionName() {
			return "Reguläres Löschen";
		}
	}

	/**
	 * Spontanes Löschen bei wenig freiem Speicherplatz
	 */
	public class DeleteSpontan implements ContainerHandler {
		private final Calendar _compareDate;
		private final DeleteLevel _deleteLevel;

		/**
		 * Erstellt ein neues DeleteSpontan-Objekt
		 *
		 * @param compareDate            Vergleichsdatum
		 * @param deleteLevel Lösch-Level
		 */
		public DeleteSpontan(final Calendar compareDate, final DeleteLevel deleteLevel) {
			this._compareDate = compareDate;
			this._deleteLevel = deleteLevel;
		}


		@Override
		public void handleContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind, final Collection<ContainerHeaders> headers) throws PersistenceException, IndexException {
			deleteContainers(lock, dataKind, headers);	
		}

		@Override
		public boolean appliesTo(final IdDataIdentification dataIdentification, final ContainerHeaders headers) {
			_longTermTaskStatePublisherTask.setTaskProgress("Spontanes Löschen: %1$d Container gelöscht", _numOfDeletedCont.get(), -1, -1);
			return hasToBeDeletedSpontan(dataIdentification.getSimVariant(), _compareDate, _deleteLevel, headers);	
		}


		@Override
		public boolean maySuspend() {
			return false;
		}

		@Override
		public String getActionName() {
			return "Spontanes Löschen";
		}
	}

	/**
	 * Endgültiges Löschen
	 */
	public class DeletePermanently implements ContainerHandler {

		@Override
		public void handleContainers(final SyncKey<IdDataIdentification> indexLock, final ArchiveDataKind dataKind, final Collection<ContainerHeaders> headers) throws PersistenceException, SynchronizationFailedException, IndexException {
			IdDataIdentification dataIdentification = indexLock.getElement();
			DeletedContainerFile deletedContainerFile = persMgr.openDeletedContainerFile(indexLock, dataKind);

			for(ContainerHeaders header : headers) {

				long containerId = header.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_CONT_ID);

				try {
					persMgr.removeContainerFromIndex(indexLock, dataKind, containerId);
					if(header.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) {
						deletedContainerFile.removeDeletedContainer(containerId);
					}
					else {
						Path contFile = persMgr.getPath(dataIdentification, dataKind).resolve(PersistenceManager.getContFileName(containerId));
						
						Files.delete(contFile);
					}
				}
				catch(IOException e) {
					_debug.warning("Container " + containerId + " in " + dataIdentification.toString() + " konnte nicht endgültig gelöscht werden", e);
					return;
				}

				_numOfDeletedCont.getAndIncrement();
			}
			deletedContainerFile.write();
		}

		@Override
		public boolean appliesTo(final IdDataIdentification dataIdentification, final ContainerHeaders headers) {
			_longTermTaskStatePublisherTask.setTaskProgress("Endgültiges Löschen: %1$d Container gelöscht", _numOfDeletedCont.get(), -1, -1);

			long maxArcTime = headers.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_ARC_TIME_MAX);

			if(maxArcTime == 0) {
				// nicht abgeschlossen
				return false;
			}

			long permanentDeleteTimeSeconds = archMgr.getArchivConfig().getPermanentDeleteTimeSeconds(archMgr.getDataIdentTree().get(dataIdentification));
			if (permanentDeleteTimeSeconds <= 0) return false;

			return maxArcTime + 1000 * permanentDeleteTimeSeconds < System.currentTimeMillis();
		}

		@Override
		public boolean maySuspend() {
			return false;
		}

		@Override
		public String getActionName() {
			return "Endgültiges Löschen";
		}
	}

	/**
	 * Manuelle Löschschutzverlängerung
	 */
	public class ChangeTime implements ContainerHandler {

		private final long _changeTime;

		public ChangeTime(final long changeTime) {
			_changeTime = changeTime;
		}

		@Override
		public void handleContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind, final Collection<ContainerHeaders> headers) throws PersistenceException, IndexException {
			for(ContainerHeaders header : headers) {
				persMgr.extendLoeschschutz(lock, header, _changeTime);
				_numOfDeletedCont.getAndIncrement();
				_longTermTaskStatePublisherTask.setTaskProgress("Löschschutzverlängerung: %1$d Container angepasst", _numOfDeletedCont.get(), -1, -1);
			}
	
		}

		@Override
		public boolean appliesTo(final IdDataIdentification dataIdentification, final ContainerHeaders headers) {
			return true;
		}

		@Override
		public boolean maySuspend() {
			return false;
		}

		@Override
		public String getActionName() {
			return "Löschschutzverlängerung";
		}
	}
}
