/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Objects;

/**
 * Klasse für den Zugriff auf eine Container-Datei. Diese Klasse implementiert {@link ContainerManagementData} um die Header-informationen aus der
 * Container-Datei zu lesen und {@link de.bsvrz.ars.ars.persistence.iter.DataSequence} zum Zugriff auf die Daten.
 * <p>
 * Außerdem implementiert die Klasse {@link AutoCloseable} um sicherzustellen, dass die Containerdatei zeitnah wieder geschlossen wird.
 *
 * @author Kappich Systemberatung
 */
public class ContainerFileHandle extends BasicContainerFileHandle {
	private final PersistenceManager _persMgr;
	private final IdContainerFileDir _directory;

	/**
	 * Erstellt eine neue Instanz zum Zugriff auf die Containerdatei
	 * @param idContainerFileDir {@link DataIdentNode Datenidentifikations-Knoten} zum synchronisierten Zugriff auf die Datei
	 * @param containerId Container-ID, die geöffnet werden soll
	 * @param readonly Schreibgeschützter Zugriff? (Benötigt weniger Synchronisierung)
	 * @param persMgr Persistenzmanager (zur Ermittlung des Wurzelverzeichnisses)
	 * @throws PersistenceException Fehler beim Zugriff
	 */
	public ContainerFileHandle(final IdContainerFileDir idContainerFileDir, final long containerId, final boolean readonly, final PersistenceManager persMgr) throws PersistenceException {
		Objects.requireNonNull(idContainerFileDir, "idContainerFileDir == null");
		Objects.requireNonNull(persMgr, "persMgr == null");
		_directory = idContainerFileDir;
		_persMgr = persMgr;
		File contDir = createDirectoriesIfAbsent();
		getContainerFile().accessContainer(_directory, containerId, contDir, readonly);
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation. Wenn das Verzeichnis nicht existiert, wird es angelegt (lazy-Verhalten).
	 *
	 * @return Archiv-Pfad
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	private File createDirectoriesIfAbsent() throws PersistenceException {
		return createDirectoriesIfAbsent(_directory, _persMgr);
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation. Wenn das Verzeichnis nicht existiert, wird es angelegt (lazy-Verhalten).
	 *
	 * @param idContainerFileDir      Archivdatenart
	 * @param persMgr              Persistenz-Manager
	 * @return Archiv-Pfad
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	static File createDirectoriesIfAbsent(IdContainerFileDir idContainerFileDir, final PersistenceManager persMgr) throws PersistenceException {
		return createDirectoriesIfAbsent(persMgr.getPath(idContainerFileDir));
	}

	/**
	 * Liefert ein File für das angegebene Verzeichnis. Wenn das Verzeichnis nicht existiert, wird es angelegt (lazy-Verhalten).
	 *
	 * @return Archiv-Pfad als File
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	private static File createDirectoriesIfAbsent(Path path) throws PersistenceException {
		
		if(Files.isDirectory(path)) {
			// Verzeichnis existiert schon, also einfach zurückgeben.

			// Dieser vorzeitige Abbruch ist eigentlich nicht nötig, führt aber zu einer deutlichen Performance-Verbesserung.
			// Files.createDirectories ist verhältnismäßig langsam, auch wenn das Verzeichnis schon existiert.
			return path.toFile();
		}
		
		try {
			Files.createDirectories(path);
		}
		catch(IOException e) {
			throw new PersistenceException("Persistenzverzeichnis '" + path.toAbsolutePath() + "' konnte nicht angelegt werden.", e);
		}
		return path.toFile();
	}


	/**
	 * Prüft ob bereits eine Container-Datei existiert.
	 *
	 * @return Kennzeichen
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public boolean existsContainer() throws PersistenceException {
		return getContainerFile().existsContainer();
	}


	/**
	 * Erzeugt eine neue Container-Datei für den mit spezifizierten Container. Schreibt den Container-Header und befuellt diesen mit den vorher gesetzten
	 * Parametern.
	 *
	 * @param toSave Sichern
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void createContainer(boolean toSave) throws PersistenceException {
		getContainerFile().createContainer(toSave);
	}

	/**
	 * Schliesst die Container-Datei. Der Container-Header wird entsprechend aktualisiert. Die Min/Max-Werte und die Anzahl der Datensätze werden dem
	 * DataIdentNode und dessen Indexen entnommen.
	 *
	 * @param adk Datensatzart des aktuellen Containers
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	public void closeContainer(final SyncKey<IdDataIdentification> indexLock, ArchiveDataKind adk) throws PersistenceException, IndexException, SynchronizationFailedException {
		ContainerFile cf = getContainerFile();

		StandardOpenContainerData openContainerData = _persMgr.getOpenContainerData(getLocation());
		if(openContainerData == null) {
			throw new PersistenceException("Kann aktuellen Container nicht schließen, da Container nicht offen ist: " + cf);
		}
		cf.closeContainer(openContainerData, _persMgr.arSParamGetVorhalten(_directory.getDataIdentification()));
		_persMgr.updateContainerIndex(indexLock, this);
		_persMgr.updateStandardIndexes(indexLock, adk, openContainerData);
		_persMgr.removeOpenContainerData(indexLock, adk);
	}


	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Header
	//////////////////////////////////////////////////////////////////////////////////////////////////////////


	/**
	 * Nachdem mit {@link #readContainerHeader()} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines Parameters im
	 * Container-Header geändert werden. Wirft eine {@link PersistenceException}, wenn die Änderung nicht durchgefuehrt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 * @param val   Wert des Container-Header-Parameters
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void setContainerHeaderParam(KeyValParam param, String val) throws PersistenceException {
		getContainerFile().setContHdrParam(param, val);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader()} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines numerischen
	 * Parameters im Container-Header geändert werden. Wirft eine {@link PersistenceException}, wenn die Änderung nicht durchgefuehrt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 * @param val   Numerischer Wert des Container-Header-Parameters
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void setContainerHeaderParam(KeyValParam param, long val) throws PersistenceException {
		getContainerFile().setContHdrParam(param, val);
	}


	/**
	 * Die mit {@link #setContainerHeaderParam(KeyValParam, String)} gesetzten Parameter werden im Container-Header persistiert.
	 *
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 * @param indexLock
	 */
	public void writeContainerHeader(final SyncKey<IdDataIdentification> indexLock) throws PersistenceException, IndexException {
		writeContainerHeaderWithoutUpdatingIndex();
		
		_persMgr.updateContainerIndex(indexLock, this);
	}

	/**
	 * Die mit {@link #setContainerHeaderParam(KeyValParam, String)} gesetzten Parameter werden im Container-Header persistiert.
	 * 
	 * Diese Methode aktualisiert den Verwaltugnsdatenindex nicht und kann z. B. benutzt werden, währen ein Container repariert wird bevor der Index aufgebaut ist.
	 *
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void writeContainerHeaderWithoutUpdatingIndex() throws PersistenceException {
		getContainerFile().writeContainerHeader();
	}


	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Datensätze
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Ermittelt die Gesamtlänge des mit {@code accessContainer()}- im Zugriff befindlichen Containers.
	 *
	 * @return Länge der Container-Datei
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public long getContainerSize() throws PersistenceException {
		return getContainerFile().getContainerSize();
	}

	/**
	 * Haengt den bereits übergebenen und serialisierten Datensatz an die aktuelle Container-Datei an.
	 *
	 * @param actualWriteBuf Schreibpuffer
	 * @param totalWriteDataSize Benutzte Länge im Schreibpuffer
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void appendSerializedData(final byte[] actualWriteBuf, final int totalWriteDataSize) throws PersistenceException {
		getContainerFile().appendSerializedData(actualWriteBuf, totalWriteDataSize);
	}
}
