/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Multimap;
import com.google.common.collect.MultimapBuilder;
import com.google.common.collect.SetMultimap;
import com.google.common.util.concurrent.UncheckedExecutionException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.InQueuesMgr;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.datatree.IndexTree;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.DebuggingSynchronizationManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationManagerImpl;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveTask;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.RestoreTask;
import de.bsvrz.ars.ars.persistence.index.ArchiveTimeIndex;
import de.bsvrz.ars.ars.persistence.index.ArchiveTimeIndexImpl;
import de.bsvrz.ars.ars.persistence.index.ContainerManagementIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndexImpl;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndex;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndexImpl;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexValues;
import de.bsvrz.ars.ars.persistence.index.backend.management.BaseIndex;
import de.bsvrz.ars.ars.persistence.index.backend.management.LongRange;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.ars.ars.persistence.iter.DeletedTreatment;
import de.bsvrz.ars.ars.persistence.walk.internal.StatusPrinter;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.util.BufferedRandomAccessFile;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import de.bsvrz.sys.funclib.losb.util.Util;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.util.zip.Inflater;

/**
 * Zentrale Persistenz-Verwaltung für die Archivierung von Daten auf einem Speichermedium vom Typ A (Festplatte) und die entsprechende Meta-Daten-Verwaltung.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class PersistenceManager {

	/** Maximale Dauer, um die der Loeschschutz automatisch verlängert wird (gemessen ab Ende Vorhaltezeitraum) */
	private static long _tLSMax = 6L * 7 * 24 * 3600; // 6 Wochen in Sekunden

	/**
	 * Soll der Löschzeitpunkt dynamisch aus der Parametrierung ermittelt werden (true) oder beim Abschließen der
	 * Containerdatei dort einmal fest eingetragen werden (klassisches Verhalten, false)
	 */
	private boolean _useDynamicDeletionTime;

	/**
	 * Methode für Tests: Setzt, ob die Löschzeit dynamisch an den parameter angepasst wird oder nicht (altes Verhalten)
	 * @param useDynamicDeletionTime Soll der Löschzeitpunkt dynamisch aus der Parametrierung ermittelt werden (true) oder beim Abschließen der
	 * Containerdatei dort einmal fest eingetragen werden (klassisches Verhalten, false)
	 */
	public void setUseDynamicDeletionTime(final boolean useDynamicDeletionTime) {
		_useDynamicDeletionTime = useDynamicDeletionTime;
	}

	/**
	 * Setzt, ob beim Wiederherstellungslauf beschädigte Indexe einfach nur gelöscht statt wiederhergestellt werden sollen.
	 * @param skipRebuild Wiederherstellung überspringen?
	 */
	public static void setSkipRebuild(final boolean skipRebuild) {
		_skipRebuild = skipRebuild;
	}
	
	private static boolean _skipRebuild;

	/**
	 * Setzt, ob defekte Containerdateien umbenannt werden sollen
	 * @param deleteBrokenContainers
	 */
	public static void setDeleteBrokenContainers(final boolean deleteBrokenContainers) {
		_deleteBrokenContainers = deleteBrokenContainers;
	}

	private static boolean _deleteBrokenContainers;

	private static final Debug _debug = Debug.getLogger();

	private static final String OBJ_ID_PREFIX = "obj";

	private static final String ATG_ID_PREFIX = "atg";

	private static final String ASP_ID_PREFIX = "asp";

	private static final String SV_PREFIX = "sv";

	private static final Pattern OBJ_DIRNAME_PAT = Pattern.compile("^" + OBJ_ID_PREFIX + "(\\d+)$");

	private static final Pattern ATG_DIRNAME_PAT = Pattern.compile("^" + ATG_ID_PREFIX + "(\\d+)$");

	private static final Pattern ASP_DIRNAME_PAT = Pattern.compile("^" + ASP_ID_PREFIX + "(\\d+)$");

	private static final Pattern SV_DIRNAME_PAT = Pattern.compile("^" + SV_PREFIX + "(\\d+)$");

	private static final Pattern NUM_DIR_PAT = Pattern.compile("^\\w+?(\\d+)$");

	/**
	 * Dateiname des Lock-Files des Archivsystems. Die Datei wird angelegt, wenn das Archivsystem gestartet wird. Die wird gelöscht, wenn das Archivsystem
	 * heruntergefahren wird.
	 */
	public static final String ACTIVITY_FLAG_FILE_NAME = "_isActive.flag";

	/**
	 * Eine Datei dieses Namens wird pro Verzeichnis einer Datenidentifikation/Datensatzart angelegt, wenn ein neuer Datensatz archiviert wurde und der Index
	 * möglicherweise (wg. Caching) noch nicht auf die Platte durchgeschrieben wurde. Beim Herunterfahren des Systems werden alle Indexe durchgeschrieben und die
	 * Flag-Dateien geloescht. Existiert diese Datei beim Systemstart noch, ist dies ein Hinweis auf eine unkorrekte Beendigung des Archivsystems. Für jede DId,
	 * für die die Datei existiert, werden die Indexe neu aufgebaut.
	 */
	public static final String REBUILDINDEX_FILE_FLAG_NAME = "_rebuildIndex.flag";

	/**
	 * Dateiname für Datei mit Zeitstempeln je Datenidentifikation für die Ermittlung des Datenlückenanfangs.
	 */
	public static final String RESTART_TIME_FILE_NAME = "_restartTime.property";

	/** Zählt die besuchten Verzeichnisse beim Prüfen des Persistenzverzeichnisses */
	private int directoryCount;

	/** Enthält die zuletzt verwendete ContainerID */
	private AtomicLong nextContainerID = new AtomicLong(1);

	private final ArchiveManager archivMgr;

	private final StartupProperties startupProps;

	private Path archivePath;

	/** Die Datei zeigt an, ob das Persistenzverzeichnis von einem Archivsystem gerade benutzt wird. */
	private File isActive;

	private boolean _preventWriteStartupInfo;
	
	private final DataIdentTree _dataIdentTree;
	
	private final Path _unsubscriptionFile;
	
	private final Map<IdContainerFileDir, StandardOpenContainerData> _openContainerData = new ConcurrentHashMap<>();
	
	private final IndexTree _indexTree;

	private static final StandardOpenContainerData NO_OPEN_CONTAINER = new StandardOpenContainerData(0, 0, 0, 0, 0, 0, 0, 0);

	private final ConcurrentSkipListSet<IdContainerFileDir> _largeDataIdentifications = new ConcurrentSkipListSet<>();

	private final SynchronizationManager<IdDataIdentification> _indexLockManager;


	/**
	 * Erzeugt den Persistenz-Manager.
	 *
	 * @param archMgr  Archiv-Manager
	 * @param archPath Archivierungs-Verzeichnis
	 */
	public PersistenceManager(ArchiveManager archMgr, Path archPath) {
		archivMgr = archMgr;
		int indexCacheMaxSize = archMgr != null ? archMgr.getIndexCacheMaxSize() : 16384;
		_indexTree = new IndexTree(indexCacheMaxSize, this);
		_dataIdentTree = archMgr != null ? archMgr.getDataIdentTree() : new DataIdentTree();
		_useDynamicDeletionTime = archMgr != null && archMgr.isUsingDynamicDeletionTime();
		setPersistenceDir(archPath);
		startupProps = new StartupProperties(archivePath);
		_unsubscriptionFile = getArchivePath().resolve(RESTART_TIME_FILE_NAME);
		if(Files.exists(_unsubscriptionFile)) {
			try {
				loadUnsubscriptionTime();
			}
			catch(IOException e) {
				_debug.warning("Kann " + _unsubscriptionFile + " nicht laden", e);
			}
		}
		_indexLockManager = new DebuggingSynchronizationManager<>(new SynchronizationManagerImpl<>(_indexTree::ensureNoCached,
		                                                                                           _indexTree::closeIndexes));
	}



	/**
	 * Muss immer vor Zugriffen auf den kritischen Bereich (die Indexe oder Container) der Datenidentifikation aufgerufen werden.
	 * <p>
	 * Wenn ein anderer Task sich im kritischen Bereich befindet, blockiert diese Methode bis der kritische Bereich wieder frei ist.
	 * <p>
	 * Da der Index-Bereich möglicherweise von vielen Tasks benutzt wird und auch Zugriffe beim Archivieren von Daten erfolgen sollte
	 * der Bereich so schnell wie möglich wieder verlassen werden. Es ist zwingend erforderlich, das zurückgegebene Lock-Objekt nach Benutzung zu schließen, sonst können andere Threads blockiert werden.
	 * <p>
	 * Beispiel-Code:
	 * <pre>{@code
	 *     try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(din.getDataIdentification())) {
	 *          // Tue was mit den Indexen
	 *     }
	 * }</pre>
	 * <p>
	 * Das zurückgegebene {@link SyncKey}-Objekt verhält sich wie ein {@link java.util.concurrent.locks.ReentrantLock},
	 * es ist also möglich, ein zweites Lock zu erhalten wenn bereits ein Lock geholt wurde.
	 * <p>
	 * @param dataIdentification Datenidentifikation
	 * @return Lock
	 * @throws SynchronizationFailedException Synchronisierung ist fehlgeschlagen (Unterbrochen beim Warten)
	 */
	public SyncKey<IdDataIdentification> lockIndex(final IdDataIdentification dataIdentification) throws SynchronizationFailedException {
		return _indexLockManager.acquireWriteKey(dataIdentification);
	}

	/**
	 * Muss immer vor Zugriffen auf den kritischen Bereich (die Indexe oder Container) der Datenidentifikation aufgerufen werden.
	 * <p>
	 * Wenn ein anderer Task sich im kritischen Bereich befindet, blockiert diese Methode bis der kritische Bereich wieder frei ist.
	 * <p>
	 * Da der Index-Bereich möglicherweise von vielen Tasks benutzt wird und auch Zugriffe beim Archivieren von Daten erfolgen sollte
	 * der Bereich so schnell wie möglich wieder verlassen werden. Es ist zwingend erforderlich, das zurückgegebene Lock-Objekt nach Benutzung zu schließen, sonst können andere Threads blockiert werden.
	 * <p>
	 * Beispiel-Code:
	 * <pre>{@code
	 *     try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(din.getDataIdentification())) {
	 *          // Tue was mit den Indexen
	 *     }
	 * }</pre>
	 * <p>
	 * Das zurückgegebene {@link SyncKey}-Objekt verhält sich wie ein {@link java.util.concurrent.locks.ReentrantLock},
	 * es ist also möglich, ein zweites Lock zu erhalten wenn bereits ein Lock geholt wurde.
	 * <p>
	 * @param dataIdentification Datenidentifikation
	 * @param timeout Maximale Wartezeit
	 * @return Lock
	 * @throws SynchronizationFailedException Synchronisierung ist fehlgeschlagen (Timeout oder Unterbrochen beim Warten)
	 */
	public SyncKey<IdDataIdentification> lockIndex(final IdDataIdentification dataIdentification, Duration timeout) throws SynchronizationFailedException {
		return _indexLockManager.acquireWriteKey(dataIdentification, timeout);
	}

	/**
	 * Gibt alle aktuell genutzten Locks zur Synchronisation auf die Datenidentifikationen zurück.
	 * @return Locks
	 */
	public SetMultimap<IdDataIdentification, SyncKey<IdDataIdentification>> getIndexLocks() {
		return _indexLockManager.getLocks();
	}


	/**
	 * Testmethode fürs Debugging, stellt sicher, dass aktuell keine Locks vom aktuellen Thread gehalten werden.
	 * Dieser Aufruf kann eingefügt werden, wenn eine Methode einen blockierenden Aufruf macht (z. B. auf Netzwerkantwort warten).
	 * Dies sollte aus Deadlock-Gefahr-Gründen nicht gemacht werden, während ein Index gelockt ist. 
	 */
	public void assertNoLocks() {
		assert assertNoLocksInternal(Thread.currentThread());
	}

	@SuppressWarnings("SameReturnValue")
	private boolean assertNoLocksInternal(final Thread currentThread) {
		Collection<SyncKey<IdDataIdentification>> values = getIndexLocks().values();
		for(SyncKey<IdDataIdentification> value : values) {
			if(value.getThread() == currentThread) {
				throw new AssertionError("Thread " + currentThread + " darf keine Locks halten: " + value);
			}
		}
		return true;
	}


	/** 
	 * Setzt die maximale Löschschutzzeit
	 * @param tlsMax Zeit in Sekunden
	 * @see #_tLSMax */
	public static void settLSMax(long tlsMax) {
		_tLSMax = tlsMax;
	}


	/**
	 * Gibt zu der angegebenen Datenart den Pfadnamen zurück, der verwendet wird.
	 * @param archiveDataKind Archivdatenart
	 * @return Suffix
	 */
	public static String getDataKindSuffix(final ArchiveDataKind archiveDataKind) {
		return getDataKindSuffix(Util.getDataKindIndex(archiveDataKind));
	}


	/**
	 * Gibt zu der angegebenen Datenart den Pfadnamen zurück, der verwendet wird.
	 * @param adkIdx Archivdatenart (siehe {@link Util#getDataKindFromIndex(int)})
	 * @return Suffix
	 */
	public static String getDataKindSuffix(final int adkIdx) {
		switch(adkIdx) {
			case 0: return "oa";
			case 1: return "on";
			case 2: return "na";
			case 3: return "nn";
		}
		throw new IllegalArgumentException("adkIdx=" + adkIdx);
	}

	/**
	 * Gibt {@code true} zurück, wenn es sich um einen gültigen Datenart-Ordnernamen handelt
	 * @param dirName Ordnername
	 * @return {@code true}, wenn es sich um einen gültigen Datenart-Ordnernamen handelt, sonst {@code false}
	 */
	public static boolean isValidDataKindSuffix(String dirName) {
		switch(dirName) {
			case "oa":
			case "on":
			case "na":
			case "nn":
				return true;
			default:
				return false;
		}
	}

	/**
	 * Extrahiert die Archivdatenart aus einer Pfadangabe
	 * @param dir Ordner
	 * @return Archivdatenart
	 *
	 * @throws IllegalArgumentException Falls der Ordnername ungültig ist.
	 */
	public static ArchiveDataKind getArchiveDataKindFromDir(File dir) {
		String dirName = dir.getName();

		switch(dirName) {
			case "oa": return ArchiveDataKind.ONLINE;
			case "on": return ArchiveDataKind.ONLINE_DELAYED;
			case "na": return ArchiveDataKind.REQUESTED;
			case "nn": return ArchiveDataKind.REQUESTED_DELAYED;
			default:
				throw new IllegalArgumentException(String.valueOf(dir));
		}
	}
	
	/**
	 * Extrahiert die Archivdatenart aus einer Pfadangabe
	 * @param dir Ordner
	 * @return Archivdatenart
	 *
	 * @throws IllegalArgumentException Falls der Ordnername ungültig ist.
	 */
	public static ArchiveDataKind getArchiveDataKindFromDir(Path dir) {
		String dirName = dir.getFileName().toString();

		switch(dirName) {
			case "oa": return ArchiveDataKind.ONLINE;
			case "on": return ArchiveDataKind.ONLINE_DELAYED;
			case "na": return ArchiveDataKind.REQUESTED;
			case "nn": return ArchiveDataKind.REQUESTED_DELAYED;
			default:
				throw new IllegalArgumentException(String.valueOf(dir));
		}
	}

	/**
	 * Erzeugt ein {@link IdContainerFileDir} aus einem Datenträgerverzeichnis
	 * @param dir Verzeichnis
	 * @return IdContainerFileDir
	 * @throws PersistenceException Falls der Pfad kein gültiges Containerverzeichnis war.
	 */
	public static IdContainerFileDir getContainerLocationFromDir(final Path dir) throws PersistenceException {
		return new IdContainerFileDir(getDataIdentificationFromPath(dir), getArchiveDataKindFromDir(dir));
	}

	/**
	 * Erzeugt ein {@link IdContainerFileDir} aus einem Datenträgerverzeichnis.
	 * @param dir Verzeichnis
	 * @return IdContainerFileDir
	 * @throws PersistenceException Falls der Pfad kein gültiges Containerverzeichnis war.
	 */
	public static IdContainerFileDir getContainerLocationFromDir(final File dir) throws PersistenceException {
		return new IdContainerFileDir(getDataIdentificationFromPath(dir), getArchiveDataKindFromDir(dir));
	}

	/**
	 * Markiert die angegebenen Container als gelöscht und trägt sie als gelöscht im Index ein. Diese Methode muss synchronisiert auf
	 * die Container eines {@link DataIdentNode} ausgeführt werden.
	 * 
	 * @param indexLock    Index-Schlüssel
	 * @param idContainerFileDir Container-Schlüssel   
	 * @param containerIds Container-IDs, die gelöscht werden sollen
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void deleteContainers(final SyncKey<IdDataIdentification> indexLock, final IdContainerFileDir idContainerFileDir, final long... containerIds) throws PersistenceException, SynchronizationFailedException, IndexException {
		ArchiveDataKind adk = idContainerFileDir.getArchiveDataKind();

		// Die Datei mit den gelöschten Containern ganz am Anfang öffnen, sonst wird der Container beim Aufbau der Datei evtl. doppelt gelöscht.
		DeletedContainerFile file = openDeletedContainerFile(indexLock, adk);

		for(Long containerID : containerIds) {
			ContainerFileHandle containerFileHandle = accessContainer(idContainerFileDir, containerID);
			ContainerHeaders actualHeaders = getContainerHeaders(indexLock, adk, containerID);
			actualHeaders.setContainerHeaderParam(ContainerManagementInformation.CHP_DELETED, true);

			updateContainerIndex(actualHeaders, _indexTree.getContainerManagementIndex(indexLock, adk));

			File originalFile = containerFileHandle.getFile();
			file.addDeletedContainer(actualHeaders, containerFileHandle);
			containerFileHandle.close();

			if(!originalFile.delete()) {
				throw new PersistenceException("Kann Containerdatei nicht löschen: " + originalFile);
			}
		}

		file.write();
	}

	/** 
	 * Gibt die Header eines Containers zurück. Die Daten werden aus dem Verwaltungsdatenindex gelesen.
	 * @param indexLock   Indexzugriff
	 * @param adk         Archivdatenart
	 * @param containerID Container-ID, dessen Header glesen wurden
	 * @return die Header eines Containers
	 * @throws IndexException Daten konnten nicht ermittelt werden (z. B. Container nicht vorhanden)
	 */
	public ContainerHeaders getContainerHeaders(final SyncKey<IdDataIdentification> indexLock, final ArchiveDataKind adk, final long containerID) throws IndexException {
		return getHeaders(_indexTree.getContainerManagementIndex(indexLock, adk), containerID, indexLock.getElement(), adk);
	}

	/**
	 * Gibt die Header von alles Cotnainern eines Verzeichnisses zurück. Im Pronzip wird der gesamte Inhlt des Verwaltungsdatenindex zurüückgegeben.
	 * @param indexLock   Indexzugriff
	 * @param adk         Archivdatenart
	 * @return die Header als {@link IndexResult}
	 * @throws IndexException Daten konnten nicht ermittelt werden
	 */
	public IndexResult<ContainerManagementInformation> getContainerHeaders(final SyncKey<IdDataIdentification> indexLock, final ArchiveDataKind adk) throws IndexException {
		return _indexTree.getContainerManagementIndex(indexLock, adk).query();
	}

	/**
	 * Trägt Headerdaten in eine Containerdatei ein
	 * @param from daten
	 * @param to   Handle auf einen Container
	 * @throws PersistenceException Schreibfehler
	 */
	public static void copyHeaders(final ContainerManagementData from, final ContainerFileHandle to) throws PersistenceException {
		to.ensureHeaderRead();
		for(ContainerManagementInformation information : ContainerManagementInformation.values()) {
			if(information.isNumeric()) {
				to.setContainerHeaderParam(information.delegate(), from.getContainerHeaderParamAsLong(information));
			}
			else {
				to.setContainerHeaderParam(information.delegate(), from.getContainerHeaderParamAsString(information));
			}
		}
	}

	/**
	 * Setzt persistenceDir von diesem PersistenceManager-Objekt.
	 *
	 * @param archPath neuer Wert
	 *
	 */
	private void setPersistenceDir(Path archPath) {
		archivePath = archPath;
		boolean exists = Files.isDirectory(archPath);
		if(exists) {
			return;
		}
		_debug.info("Persistenzverzeichnis '" + archPath + "' existiert nicht, wird neu angelegt.");
		try {
			Files.createDirectories(archPath);
		}
		catch(IOException e) {
			_debug.warning("Kann Verzeichnis nicht anlegen: " + archPath, e);
			archivMgr.quitError("Persistenzverzeichnis '" + archPath + "' existiert nicht und konnte nicht angelegt werden.");
		}
	}

	/**
	 * Liefert Kennzeichen, ob die angegebene Datei eine Container-Datei ist.
	 *
	 * @param file zu prüfende Datei
	 *
	 * @return Kennzeichen
	 */
	public static boolean isContainerFile(File file) {
		return ContainerFile.isContainerFile(file);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer Indexdatei ist
	 */
	private static boolean isIndexFile(String name) {
		return name.equals(DataTimeIndexImpl.IDX_FILENAME) 
				|| name.equals(ArchiveTimeIndexImpl.IDX_FILENAME)
				|| name.equals(ContainerManagementIndex.IDX_FILENAME)
				|| name.equals(DataIndexIndexImpl.IDX_FILENAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer _rebuildIndex.flag-Datei ist
	 */
	private static boolean isRebuildIndexFlagFile(String name) {
		return name.equals(REBUILDINDEX_FILE_FLAG_NAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer _isActive.flag-Datei ist
	 */
	private static boolean isActivityFlagFile(String name) {
		return name.equals(ACTIVITY_FLAG_FILE_NAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es ein zulaessiger Dateiname des Archivsystems ist
	 */
	private static boolean isValidStorageArSFile(String name) {
		return isValidIDFolderName(name) || ContainerFile.isContainerFileName(name) || isIndexFile(name) || isRebuildIndexFlagFile(name);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es ein zulaessiger Dateiname im obersten Persistenzverzeichnis des Archivsystems ist
	 */
	private static boolean isValidTopLevelArSFile(String name) {
		return isValidIDFolderName(name) || RestoreTask.isUnavailMediumIDsFile(name) || StartupProperties.isStartupInfoFile(name) || isActivityFlagFile(name);
	}


	/**
	 * Liefert die Container-ID zur angegebenen Container-Datei, indem die im Container-Dateinamen enthaltene Container-ID ausgelesen wird. Alternativ kann die
	 * Container-ID aus den Container-Header-Parametern gelesen werden.
	 *
	 * @param file Container-Datei
	 *
	 * @return Container-ID
	 */
	public static long getContID(File file) {
		return ContainerFile.getContID(file);
	}

	/**
	 * Liefert den Dateinamen des Daten-Containers mit der angegebenen Container-ID.
	 *
	 * @param containerId Container-ID
	 *
	 * @return Dateiname
	 */
	public static String getContFileName(long containerId) {
		return ContainerFile.getContainerFileName(containerId);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Verzeichnis
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt und mit einem
	 * Praefix versehen werden. Pro Dreiergruppe und pro SimVarwird ein Verzeichnis verwendet.
	 *
	 * @param objId Objekt-ID
	 * @param atgId Attributgruppen-ID
	 * @param aspId Aspekt-ID
	 * @param sv    Simulationsvariante
	 *
	 * @return Archiv-Pfad
	 */
	public Path getPath(long objId, long atgId, long aspId, int sv) {
		return getPath(archivePath, objId, atgId, aspId, sv);
	}


	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und den Wurzelpfad des Archivsystems. Der Pfad wird gebildet indem die IDs und die SV in
	 * Dreiergruppen zerlegt und mit einem Praefix versehen werden. Pro Dreiergruppe und pro SimVarwird ein Verzeichnis verwendet.
	 *
	 * @param basePath Wurzelpfad des Archivsystems
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 *
	 * @return Archiv-Pfad
	 */
	public static Path getPath(Path basePath, long objId, long atgId, long aspId, int sv) {
		return appendPathName(basePath, objId, atgId, aspId, sv);
	}

	/**
	 * Ergänzt das übergebene StringBuilder-Objekt um den Archiv-Pfad für die angegebene Datenidentifikation und den Wurzelpfad des Archivsystems.
	 * Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt und mit einem Praefix versehen werden.
	 * Pro Dreiergruppe und pro SimVar wird ein Verzeichnis verwendet.
	 *
	 * @param basePath Wurzelpfad des Archivsystems
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 */
	private static Path appendPathName(final Path basePath, final long objId, final long atgId, final long aspId, final int sv) {
		Path tmp = basePath;
		tmp = appendIdElements(tmp, OBJ_ID_PREFIX, objId);
		tmp = appendIdElements(tmp, ATG_ID_PREFIX, atgId);
		tmp = appendIdElements(tmp, ASP_ID_PREFIX, aspId);
		return tmp.resolve(SV_PREFIX + sv);
	}

	/**
	 * Zerlegt die numerische ID in Gruppen von jeweils 3 Zeichen in Dezimaldarstellung und ergänzt das übergebene Path-Objekt,
	 * sodass die übergebene Separtor/Prefixsequenz hinzugefügt wird. Beispiel: mit "obj" im Parameter prefix führt die ID 
	 * 12345678 dazu, dass der Pfad "/obj123/obj456/obj78" ergänzt wird.
	 *
	 * @param basePath Basispfad
	 * @param prefix Prefix, das vor jeder Zeichengruppe eingefügt werden soll.
	 * @param id Numerische ID.
	 */
	public static Path appendIdElements(Path basePath, String prefix, long id) {
		Path tmp = basePath;
		final String idString = Long.toString(id);
		final int end = idString.length();
		int start = 0;
		StringBuilder substringBuilder = new StringBuilder(prefix.length() + 3);
		substringBuilder.append(prefix);
		while(start < end) {
			int newStart = start + 3;
			if(newStart > end) newStart = end;
			substringBuilder.setLength(prefix.length());
			substringBuilder.append(idString, start, newStart);
			tmp = tmp.resolve(substringBuilder.toString());
			start = newStart;
		}
		return tmp;
	}
	
	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 * @param dataKind Datensatzart
	 *
	 * @return Archiv-Pfad
	 */
	public Path getPath(long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind) {
		return getPath(objId, atgId, aspId, sv, Util.getDataKindIndex(dataKind));
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 * @param dataKind Datensatzart, 0 für OA, 1 für ON, 2 für NA und 3 für NN
	 *
	 * @return Archiv-Pfad
	 */
	public Path getPath(long objId, long atgId, long aspId, int sv, int dataKind) {
		return getPath(archivePath, objId, atgId, aspId, sv).resolve(getDataKindSuffix(dataKind));
	}
	
	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param ads Spezifikation des Archivdatensatzes (enthält objID, atgID, aspID, sv)
	 * @param archiveDataKind Datensatzart
	 *
	 * @return Name des Archiv-Pfads
	 */
	public Path getPath(ArchiveDataSpecification ads, ArchiveDataKind archiveDataKind) {
		DataDescription dataDescription = ads.getDataDescription();
		return getPath(ads.getObject().getId(), dataDescription.getAttributeGroup().getId(), dataDescription.getAspect().getId(), dataDescription.getSimulationVariant(), archiveDataKind);
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param dataIdentification objID, atgID, aspID, sv
	 * @param archiveDataKind Datensatzart
	 *
	 * @return Name des Archiv-Pfads
	 */
	public Path getPath(IdDataIdentification dataIdentification, ArchiveDataKind archiveDataKind) {
		return getPath(dataIdentification.getObjectId(), dataIdentification.getAtgId(), dataIdentification.getAspectId(), dataIdentification.getSimVariant(), archiveDataKind);
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe und pro SimVar wird ein Verzeichnis verwendet.
	 *
	 * @param dataIdentification objID, atgID, aspID, sv
	 *
	 * @return Name des Archiv-Pfads
	 */
	public Path getPath(IdDataIdentification dataIdentification) {
		return getPath(dataIdentification.getObjectId(), dataIdentification.getAtgId(), dataIdentification.getAspectId(), dataIdentification.getSimVariant());
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene DataIdentNode und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param dataIdentification objID, atgID, aspID, sv
	 * @param archiveDataKind Datensatzart
	 *
	 * @return Name des Archiv-Pfads
	 */
	public Path getPath(DataIdentNode dataIdentification, ArchiveDataKind archiveDataKind) {
		return getPath(dataIdentification.getDataIdentification(), archiveDataKind);
	}


	/**
	 * Diese Methode loescht das komplette Verzeichnis eines {@link DataIdentNode} mit allen Unterverzeichnissen der Datensatzarten und aller darin befindlichen
	 * Datencontainern; damit werden alle archivierten Datensätze einer Simulationsvariante geloescht. Der {@link DataIdentNode} muss in der Verwaltung aus dem
	 * {@link DataIdentTree} geloescht werden. Falls die Simulationsvariante gleich Null ist, oder das Verzeichnis nicht geloescht werden konnte, wird eine {@link
	 * PersistenceException} geworfen.
	 *
	 * @param syncKey    Schlüssel für Zugriff auf Indexknoten
	 * 
	 * @throws PersistenceException Schreibfehler im Persistenzverzeichnis
	 */
	public void deleteSimVar(final SyncKey<IdDataIdentification> syncKey) throws PersistenceException {

		if(syncKey.getElement().getSimVariant() == 0) throw new PersistenceException("Simulations-Variante '0' (Echtdaten) darf nicht geloescht werden!");

		try {
			final CacheManager cacheManager = CacheManager.getInstance();
			cacheManager.forgetCache(getOpenContID(new IdContainerFileDir(syncKey.getElement(), ArchiveDataKind.ONLINE)));
			cacheManager.forgetCache(getOpenContID(new IdContainerFileDir(syncKey.getElement(), ArchiveDataKind.ONLINE_DELAYED)));
			cacheManager.forgetCache(getOpenContID(new IdContainerFileDir(syncKey.getElement(), ArchiveDataKind.REQUESTED)));
			cacheManager.forgetCache(getOpenContID(new IdContainerFileDir(syncKey.getElement(), ArchiveDataKind.REQUESTED_DELAYED)));
			
			for(ArchiveDataKind archiveDataKind : ArchiveDataKindCombination.all()) {
				// Indexe schließen
				_indexTree.closeIndexes(syncKey, archiveDataKind);
				_openContainerData.remove(new IdContainerFileDir(syncKey.getElement(), archiveDataKind));
			}
		}
		catch(Exception e) {
			_debug.warning("Fehler beim Vergessen des Caches", e);
		}
		Path path = getPath(syncKey.getElement());
		try {
			deletePath(path);
		}
		catch(IOException e) {
			throw new PersistenceException("Simulations-Varianten-Verzeichnis konnte nicht geloescht werden: " + path, e);
		}
	}

	/**
	 * Löscht ein Verzeichnis inklusive enthaltener Dateien.
	 * @param path Verzeichnis
	 * @throws IOException Mögliche Exception
	 */
	private static void deletePath(final Path path) throws IOException {
		if(!Files.exists(path)) return;
		Files.walkFileTree(path, new SimpleFileVisitor<Path>()
		{
			@Override
			public FileVisitResult visitFile(Path file, BasicFileAttributes attrs)
					throws IOException
			{
				Files.deleteIfExists(file);
				return FileVisitResult.CONTINUE;
			}

			@Override
			public FileVisitResult visitFileFailed(Path file, IOException exc) {
				return FileVisitResult.CONTINUE;
			}

			@Override
			public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException
			{
				Files.deleteIfExists(dir);
				return FileVisitResult.CONTINUE;
			}
		});
	}
	

	/**
	 * Legt im Verzeichnis der Datenidentifikation ein {@link #REBUILDINDEX_FILE_FLAG_NAME}-Datei an.
	 *
	 * @param didPath Pfad der Datenidentifikation
	 *
	 * @param adk Archivdatenart
	 *               
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 * @see #REBUILDINDEX_FILE_FLAG_NAME
	 */
	public void createRebuildIxdFlagFile(final DataIdentificationDir didPath, final ArchiveDataKind adk) throws PersistenceException {
		Path idxFile = getRebuildIdxFile(didPath, adk);
		try {
			if(!Files.exists(idxFile)) {
				Files.createFile(idxFile);
			}
			else {
				_debug.fine(idxFile + ": Datei existiert bereits.");
			}
		}
		catch(IOException e) {
			throw new PersistenceException("Fehler beim Anlegen der '" + idxFile + "'", e);
		}
	}

	/**
	 * Die Methode getRebuildIdxFile ermittelt den Speicherort für die _rebuildIndex.flag
	 *
	 * @param didPath von Typ DataIdentificationDir  
	 * @param adk von Typ ArchiveDataKind  
	 * @return Path
	 */
	@NotNull
	public Path getRebuildIdxFile(final DataIdentificationDir didPath, final ArchiveDataKind adk) {
		return didPath.getDataIdentificationDir().resolve(getDataKindSuffix(adk)).resolve(REBUILDINDEX_FILE_FLAG_NAME);
	}

	/**
	 * Loescht die {@link #REBUILDINDEX_FILE_FLAG_NAME}-Datei.
	 *
	 * @param didPath Containerverzeichnis
	 */
	private static void removeRebuildIxdFlagFile(Path didPath) throws IndexException {
		Path file = didPath.resolve(REBUILDINDEX_FILE_FLAG_NAME);
		try {
			Files.deleteIfExists(file);
		}
		catch(IOException e) {
			throw new IndexException("Datei konnte nicht gelöscht werden", file, e);
		}
	}

	/**
	 * Überprüft eine Containerdatei darauf, ob diese mit "##\n" endet. Falls nicht wird das ungültige Ende der Datei abgeschnitten (NÄM-79)
	 *
	 * @param dataIdentification Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 * @param containerId   Container-ID
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void checkContainerFile(final IdDataIdentification dataIdentification, final ArchiveDataKind archiveDataKind, final long containerId) throws PersistenceException {
		File persDir = ContainerFileHandle.createDirectoriesIfAbsent(new IdContainerFileDir(dataIdentification, archiveDataKind), this);
		File contFile = new File(persDir, getContFileName(containerId));
		try(BufferedRandomAccessFile file = new BufferedRandomAccessFile(contFile)) {

			boolean isBroken;
			if(file.length() < 384) {
				// Container zu kurz
				isBroken = true;
			}
			else {
				file.seek(file.length() - 3);
				byte[] tmp = new byte[3];
				file.readFully(tmp);
				isBroken = !Arrays.equals(tmp, ByteIO.SEPARATOR);
			}
			if(isBroken) {
				// Defekte Containerdatei!
				_debug.warning("Container " + contFile + " ist beschädigt und wird repariert.");
				fixContainerFile(dataIdentification, archiveDataKind, contFile, file);
			}
		}
		catch(IOException e) {
			throw new PersistenceException("Konnte Container " + contFile + " nicht auf Fehler überprüfen.", e);
		}
	}

	/**
	 * Verkürzt die Containerdatei auf den gültigen Inhalt
	 * @param idDataIdentification Datenideentifikation
	 * @param archiveDataKind    ADK
	 * @param containerFile      Container-Datei
	 * @param container          RandomAccessFile der Cotnainerdatei
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	private void fixContainerFile(final IdDataIdentification idDataIdentification, final ArchiveDataKind archiveDataKind, final File containerFile, final BufferedRandomAccessFile container) throws PersistenceException {
		try {
			if(container.length() < 384) {
				// Wir wollen direkt in den catch-Block unten springen statt die Containerdatei zu verlängern
				throw new IOException("Container hat keinen Header, Länge ist nur " + container.length() + " Bytes.");
			}
			container.seek(384);

			long lastValidPosition = 384;
			final long oldLength = container.length();
			while(container.getFilePointer() < oldLength) {
				try {
					lastValidPosition = container.getFilePointer();
					int length = container.readInt();

					if(length <= 0) {
						break;
					}

					byte[] tmp = new byte[length];
					container.readFully(tmp);

					if(tmp[tmp.length - 3] != '#' || tmp[tmp.length - 2] != '#' || tmp[tmp.length - 1] != '\n') {
						// Jeder Datensatz muss auf "##\n" enden
						break;
					}
				}
				catch(EOFException ignored) {
					break;
				}
			}
			if(lastValidPosition == container.length()) {
				_debug.warning(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: Es wurde kein Fehler zum reparieren gefunden");
			}
			else {
				container.setLength(lastValidPosition);
				_debug.warning(containerFile.toString() + ": Länge wurde von " + oldLength + " auf " + lastValidPosition + " Bytes verkürzt.");
			}
		}
		catch(Exception e) {
			try {
				container.close();
			}
			catch(IOException ignored) {
			}
			final Path contFile = containerFile.toPath();
			renameContainerFile(contFile);
		}

		try {
			try(SyncKey<IdDataIdentification> lock = lockIndex(idDataIdentification)) {
				// Neue Indexe zuweisen um Inkonsistenzen zu korrigieren
				rebuildIndexes(lock, archiveDataKind, null);
			}
		}
		catch(Exception e) {
			throw new PersistenceException(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: ", e);
		}

	}

	/**
	 * Greift auf einen Container zu.
	 * @param idContainerFileDir        Synchronisierung auf Containerverzeichnis
	 * @param containerId Id
	 * @return Cotnainerhandle
	 * @throws PersistenceException Lesefehler
	 */
	@NotNull
	public ContainerFileHandle accessContainer(final IdContainerFileDir idContainerFileDir, final long containerId) throws PersistenceException {
		return new ContainerFileHandle(idContainerFileDir, containerId, false, this);
	}

	/**
	 * Greift auf den offenen Container zu. Die Methode darf nicht für geschlossene Container verwendet werden, da dann die Synchronisation nicht sichergestellt ist.
	 * @param dataIdentification Datenidentifikation
	 * @param dataKind Archivdatenart
	 * @param openContID Id des offenen Containers.
	 * @return Containerhandle
	 * @throws PersistenceException Lesefehler
	 */
	public ContainerFileHandle accessOpenContainer(final IdDataIdentification dataIdentification, final ArchiveDataKind dataKind, final long openContID) throws PersistenceException {
		return new ContainerFileHandle(new IdContainerFileDir(dataIdentification, dataKind), openContID, false, this);
	}

	/**
	 * Greift auf eine Containerdatei unsynchronisiert zu. Diese Methode darf nur für Container außerhalb des Persistenzverzeichisses benutzt werden. 
	 * @param file Containerdatei
	 * @return Handle
	 * @throws PersistenceException Lesefehler
	 */
	public BasicContainerFileHandle accessContainerUnsynchronized(final File file) throws PersistenceException {
		return new BasicContainerFileHandle(file);
	}

	/**
	 * Iteriert über die Daten eines Containers.
	 *
	 * @param adk              Archivdatenart
	 * @param containerId      Container-ID
	 * @param deletedTreatment Behandlung gelöschter Containerdaten
	 * @param inflater         Unzip-Klasse zum dekomprimieren (darf null sein, wenn die Dateninhalte nicht interessant sind)
	 * @return Iterator
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public DataIterator iterator(final ArchiveDataKind adk, final long containerId, DeletedTreatment deletedTreatment, @Nullable final Inflater inflater, final SyncKey<IdDataIdentification> indexLock) throws PersistenceException, SynchronizationFailedException {
		try {
			ContainerHeaders headers = getContainerHeaders(indexLock, adk, containerId);

			if(headers.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) {
				switch(deletedTreatment) {
					case Ignore:
						return DataIterator.empty();
					case MinMax:
						return new DeletedMinMaxIterator(headers, adk, indexLock.getElement());
					case Full:
						DeletedContainerFile deletedContainerFile;
						try {
							deletedContainerFile = openDeletedContainerFile(indexLock, adk);
						}
						catch(SynchronizationFailedException e) {
							throw new AssertionError("ContainerLock besteht bereits", e);
						}
						return deletedContainerFile.dataIterator(containerId);
				}
				throw new IllegalArgumentException();
			}
			else{
				ContainerFileHandle handle = new ContainerFileHandle(new IdContainerFileDir(indexLock.getElement(), adk), containerId, false, this);
				handle.readContainerHeader();
				return handle.iterator(inflater);
			}
		}
		catch(IndexException e) {
			throw new PersistenceException(e);
		}
	}

	/**
	 * Die Methode prepareShutdown wird beim Herunterfahren des Archivsystems ausgeführt.
	 */
	public void prepareShutdown() throws InterruptedException {
		
		// Beenden darf nicht gestartet werden, wenn vom gleichen Thread Locks gehalten werden
		// (DeadLock mit IndexCloserWorker möglich)
		assertNoLocks();
		
		CacheManager.getInstance().close();

		ExecutorService threadPool = Executors.newFixedThreadPool(archivMgr.getNumCloseIndexThreads());

		Instant start = Instant.now();
		long indexed = 0;
		final AtomicLong run = new AtomicLong();
		// Indexdateien in mehreren Threads schließen
		// OpenContainerData nach Datenidentifikation ummodeln, damit jede Datenidentifikation zusammen geschlossen wird
		// Da ArchiveDataKind kein Enum ist, wird hier statt einer EnumMap ein Array genutzt, die Indizes entsprechen `ArchiveDataKind.getCode() - 1`
		final Map<IdDataIdentification, StandardOpenContainerData[]> tmpMap = new HashMap<>();
		for (Map.Entry<IdContainerFileDir, StandardOpenContainerData> entry : _openContainerData.entrySet()) {
			IdDataIdentification dataIdentification = entry.getKey().getDataIdentification();
			ArchiveDataKind archiveDataKind = entry.getKey().getArchiveDataKind();
			StandardOpenContainerData openContainerData = entry.getValue();
			//noinspection ObjectEquality
			if(openContainerData == NO_OPEN_CONTAINER) {
				openContainerData = null;
			}
			StandardOpenContainerData[] data =
				tmpMap.computeIfAbsent(dataIdentification, it -> new StandardOpenContainerData[4]);
			data[archiveDataKind.getCode()-1] = openContainerData;
		}
		
		for(Map.Entry<IdDataIdentification, StandardOpenContainerData[]> entry : tmpMap.entrySet()) {
			StandardOpenContainerData[] value = entry.getValue();
			threadPool.submit(new IndexCloserWorker(run, entry.getKey(), value));
			indexed++;
		}

		// Abwarten, bis alle Threads fertig sind.
		threadPool.shutdown();
		while(!threadPool.awaitTermination(60, TimeUnit.SECONDS)) {
			_debug.info("Indexe werden geschlossen: " + run.get() + " von " + indexed + " Datenidentifikationen\n" +
					            StatusPrinter.estimateRuntime(
							            Duration.between(start, Instant.now()),
							            run.get(),
							            indexed
					            ));
		}
		_debug.info("Alle Indexe geschlossen: " + run.get());
		if(_preventWriteStartupInfo) {
			_debug.warning("Herunterfahren: Startup-Info wird unterdrückt, weil beim Wiederherstellungslauf ein Fehler aufgetreten ist");
			return;
		}
		try {
			BackupTask backupTask = (BackupTask)archivMgr.getInQueuesMgr().getSingleTask(InQueuesMgr.ST_BACKUP_IDX);

			startupProps.setVal(StartupProperties.STUP_MAX_CONT_ID, getLastContainerID());
			startupProps.setVal(StartupProperties.STUP_LAST_ATIME, ArchiveTask.getLastArchiveTime());
			startupProps.setVal(StartupProperties.STUP_LAST_USED_MEDIUM_ID,backupTask.getLastUsedMediumID());
			startupProps.writeStartUpProperties();
		}
		catch(PersistenceException e) {
			_debug.error("Herunterfahren: Fehler beim Erzeugen der Startup-Info: " + e.getMessage());
		}
	}

	/**
	 * Schließt die Indexe der Datensatzart des DataIndentNode.
	 *
	 * @param lock Objekt, das den synchronisierten Zugriff auf den Index erlaubt
	 * @param data Daten
	 * @param archiveDataKind    Datensatzart
	 * @return Wahr, falls der Index geschlossen werden musste, sonst false.
	 *
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	private boolean closeIndex(final SyncKey<IdDataIdentification> lock, @Nullable StandardOpenContainerData data, ArchiveDataKind archiveDataKind) throws IndexException {
		boolean result = false;
		if(data != null) {
			updateStandardIndexes(lock, archiveDataKind, data);
			result = true;
		}
		_indexTree.closeIndexes(lock, archiveDataKind);
		DataIdentNode din = _dataIdentTree.get(lock.getElement());
		if(din.isRebuildIndexFileFlagCreated(archiveDataKind)) {
			removeRebuildIxdFlagFile(getPath(lock.getElement(), archiveDataKind));
		}
		return result;
	}

	/**
	 * Versucht, die StartUp-Properties-Datei einzulesen. Wenn das nicht gelingt, wird das gesamte Persistenzverzeichnis ({@link RestorePersDirTsk}) durchlaufen
	 * und versucht, einen gueltigen Startpunkt wiederherzustellen. Das Persistenzverzeichnis wird in {@code InQueuesMgr.NUM_OF_ARCH_QUEUES_ONLINE} Teile
	 * aufgeteilt, die jeweils von einem Thread bearbeitet werden (zwengs Performance). Am Schluss wird die StartUp-Properties-Datei geloescht. Beim Herunterfahren
	 * wird in {@link #prepareShutdown()} die Datei neu geschrieben. Daran kann das ArS beim nächsten Start erkennen, ob es ordnungsgemäß
	 * heruntergefahren wurde oder ob ein Wiederherstellungslauf erforderlich ist.
	 *
	 * @return Wahr, falls der Durchlauf des Persistenzverzeichnisses erfolgreich war und das Archivsystem starten kann, falsch sonst.
	 */
	public boolean startupProcedure() {
		long mxAT;
		int lastUsedMediumID;
		try {
			startupProps.readStartUpProperties();
			nextContainerID.set(startupProps.getValAsLong(StartupProperties.STUP_MAX_CONT_ID));
			mxAT = startupProps.getValAsLong(StartupProperties.STUP_LAST_ATIME);
			lastUsedMediumID = startupProps.getValAsInt(StartupProperties.STUP_LAST_USED_MEDIUM_ID);
		}
		catch(PersistenceException e) {
			_preventWriteStartupInfo = true;
			_debug.info(e.getMessage() + Debug.NEWLINE + "Starte Wiederherstellungslauf...");
			RestorePersDirTsk restorePersDirTsk = new RestorePersDirTsk(archivMgr);
			while(!restorePersDirTsk.isTerminated()) {
				try {
					if(archivMgr.wasTerminated()) {		// Archivsystem per Telnet beendet?
						restorePersDirTsk.terminateTask();
						return false;
					}
					//noinspection BusyWait
					Thread.sleep(100);
				}
				catch(InterruptedException ie) {
					_debug.info("Wiederherstellungslauf abgebrochen.", ie);
					return false;
				}
			}
			if(!restorePersDirTsk.success()) {
				_debug.error("Wiederherstellungslauf fehlgeschlagen.");
				return false;
			}
			_preventWriteStartupInfo = false;
			nextContainerID = new AtomicLong(restorePersDirTsk.getMaxContID());
			mxAT = restorePersDirTsk.getMaxArchiveTime();
			lastUsedMediumID = restorePersDirTsk.getMaxMedienID();
		}

		long now = ArchiveManager.archiveTime();
		ArchiveTask.setLastArchiveTime(mxAT > 0 ? mxAT : now);	// gemäß TAnfArS 5.1.2.4.3.3
		BackupTask backupTask = (BackupTask)archivMgr.getInQueuesMgr().getSingleTask(InQueuesMgr.ST_BACKUP_IDX);
		backupTask.setLastUsedMediumID(lastUsedMediumID);

		_debug.info(
				"StartupInfo: max. ContainerID=" + nextContainerID + ", letzte MedienID = " + lastUsedMediumID + ", max. Archivzeit="
						+ Util.timestrMillisFormatted(mxAT) + (mxAT <= 0 ? " (nicht ermittelbar, wurde auf " + Util.timestrMillisFormatted(now) + " gesetzt)" : "")
		);

		try {
			startupProps.deleteStartupPropsFile();
		}
		catch(IOException e) {
			// Gibt zwar ein Problem beim Beenden des Archivsystems, sollte aber den Start nicht verhindern
			_debug.warning("StartUp-Properties konnten nicht geloescht werden: " + startupProps, e);
		}

		return true;
	}

	/**
	 * Führt eine Aktion auf einem Container auf und benennt den Container um, falls die Aktion eine {@link PersistenceException} geworfen hat und der Benutzer {@link #_deleteBrokenContainers} gesetzt hat.
	 * <p>Falls der Benutzer {@link #_deleteBrokenContainers} nicht gesetzt hat, werden Exceptions weitergegeben.</p>
	 * 
	 * @param contFile Containerdatei
	 * @param action   Aktion
	 * @throws PersistenceException Beliebiger Grund
	 */
	private static void executeOnContainer(final Path contFile, ContainerCall action) throws PersistenceException, SynchronizationFailedException {
		try {
			action.call();
		}
		catch(PersistenceException e) {
			if(_deleteBrokenContainers) {
				renameContainerFile(contFile);
			}
			else {
				_debug.error("Container " + contFile + " konnte nicht gelesen werden. Aufrufargument \"-defekteDateienVerschieben\" setzen, um defekte Dateien automatisch zu entfernen.");
			}
			throw new PersistenceException(e);
		}
	}

	private static void renameContainerFile(final Path contFile) throws PersistenceException {
		Path target = contFile.resolveSibling(contFile.getFileName() + ".defect");
		try {
			Files.move(contFile, target);
			_debug.fine("Container " + contFile + " konnte nicht gelesen werden und wurde in " + target.getFileName() + " umbenannt.");
		}
		catch(IOException e1) {
			throw new PersistenceException("Umbenennen der defekten Containerdatei fehlgeschlagen: " + contFile, e1);
		}
	}

	/**
	 * Die Methode getPathName gibt den Speicherort für ein Containerverzeichnis zurück
	 *
	 * @param containerFileDir von Typ IdContainerFileDir  
	 * @return String
	 */
	public Path getPath(final IdContainerFileDir containerFileDir) {
		return getPath(containerFileDir.getObjectId(), containerFileDir.getAtgId(), containerFileDir.getAspectId(), containerFileDir.getSimVariant(), containerFileDir.getArchiveDataKind());
	}
	/**
	 * Die Methode rebuildStandardIndex erstellt einen Standard-Index aus dem Verwaltungsdatenindex
	 *
	 * @param managementIndex Verwaltungsdatenindex
	 * @param containerFileDir Speicherort
	 * @param index Zu erzeugender Index (leer)
	 * @throws IndexException problem beim Index-Zugriff
	 */
	public void rebuildStandardIndex(final ContainerManagementIndex managementIndex, final IdContainerFileDir containerFileDir, final BaseIndex<IndexValues> index) throws IndexException {
		_debug.fine("Baue Index neu auf", index.toString());
		IndexResult<ContainerManagementInformation> result = managementIndex.query();
		for(int i = 0; i < result.size(); i++) {
			int num = result.getSignedInt(i, ContainerManagementInformation.CHP_ANZ_DS);
			long contId = result.get(i, ContainerManagementInformation.CHP_CONT_ID);
			
			index.setInsertValue(IndexValues.ContainerId, contId);

			if(num == ContainerHdr.CONT_UNCLOSED) {
				// Offener Container steht nicht im Index
				StandardOpenContainerData openContainerData = getOpenContainerData(containerFileDir, managementIndex);
				if(openContainerData == null || openContainerData.getContainerId() != contId) {
					throw new IndexException("Inkonsistente Indexe, offener Container ist nicht eindeutig definiert", managementIndex.getFile());
				}
				
				index.setInsertValue(IndexValues.DataIndexMin, openContainerData.getMinDataIdx());
				index.setInsertValue(IndexValues.DataIndexMax, openContainerData.getMaxDataIdx());
				index.setInsertValue(IndexValues.DataTimeMin, openContainerData.getMinDataTime());
				index.setInsertValue(IndexValues.DataTimeMax, openContainerData.getMaxDataTime());
				index.setInsertValue(IndexValues.ArchiveTimeMin, openContainerData.getMinArcTime());
				index.setInsertValue(IndexValues.ArchiveTimeMax, openContainerData.getMaxArcTime());
			}
			else {
				index.setInsertValue(IndexValues.DataIndexMin, result.get(i, ContainerManagementInformation.CHP_DATA_IDX_MIN));
				index.setInsertValue(IndexValues.DataIndexMax, result.get(i, ContainerManagementInformation.CHP_DATA_IDX_MAX));
				index.setInsertValue(IndexValues.DataTimeMin, result.get(i, ContainerManagementInformation.CHP_DATA_TIME_MIN));
				index.setInsertValue(IndexValues.DataTimeMax, result.get(i, ContainerManagementInformation.CHP_DATA_TIME_MAX));
				index.setInsertValue(IndexValues.ArchiveTimeMin, result.get(i, ContainerManagementInformation.CHP_ARC_TIME_MIN));
				index.setInsertValue(IndexValues.ArchiveTimeMax, result.get(i, ContainerManagementInformation.CHP_ARC_TIME_MAX));
			}
			index.insert();
		}
	}

	/**
	 * Erzeugt einen neuen ContainerHeaderIndex basierend aus den Containerdateien in einem Verzeichnis
	 *
	 * @param containerFileDir     ID des Containerverzeichnisses
	 * @param containerHeaderIndex Container-Header-Index (leer)
	 * @param rebuildResult        Statistik über die Wiederherstellung
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	public void rebuildContainerHeaderIndex(final IdContainerFileDir containerFileDir, final ContainerManagementIndex containerHeaderIndex, @Nullable final RebuildResult rebuildResult) throws IndexException {

		_debug.fine("Baue Verwaltungsdatenindex neu auf", containerHeaderIndex);
		
		Path dir = getPath(containerFileDir);
		if(!Files.exists(dir)) {
			return;
		}
		
		// Zeit messen, wie lange Wiederherstellung dauert, um Benutzer bei großen Verzeichnissen zu warnen.
		Duration nextWarnTimeLimit = Duration.ofMinutes(2);
		Instant startTime = Instant.now();

		try {
			final Map<Long, Long> containerIndexToMinDataIndex = new HashMap<>();
			DeletedContainerFile deletedContainerFile = new DeletedContainerFile(containerFileDir, this);
    			if(deletedContainerFile.exists()) {
				try {
					deletedContainerFile.read();
				}
				catch(PersistenceException e) {
					_debug.warning("Fehler beim Einlesen der Datei, die gelöschte Container enthält: " + deletedContainerFile.getFile() + ". Ungültige Einträge werden ignoriert.", e);
				}
				for(long containerId : deletedContainerFile.containers()) {
					try {
						ContainerManagementData headers = deletedContainerFile.headers(containerId);
						updateContainerIndex(headers, containerHeaderIndex);
						containerIndexToMinDataIndex.put(containerId, headers.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_DATA_IDX_MIN));
					}
					catch(PersistenceException e) {
						throw new IndexException("Fehler beim Indizieren der Daten eines gelöschten Containers (" + containerId + ")", deletedContainerFile.getFile(), e);
					}
				}
			}
			try(DirectoryStream<Path> paths = Files.newDirectoryStream(dir)) {
				for (Path path : paths) {
					Duration durationOfRestore = Duration.between(startTime, Instant.now());
					if (durationOfRestore.compareTo(nextWarnTimeLimit) > 0) {
						try {
							nextWarnTimeLimit = nextWarnTimeLimit.multipliedBy(2);
						} catch (ArithmeticException ignored) {
						}
						long numEntries = containerHeaderIndex.numEntries();
						if (numEntries > IndexTree.ENTRIES_WARN_LIMIT) {
							warnAboutSlowRestore(containerFileDir, numEntries, durationOfRestore);
						}
					}
					File file = path.toFile();
					if(ContainerFile.CONT_FILENAME_FILTER.accept(file.getParentFile(), file.getName())) {
						try {PersistenceManager.executeOnContainer(path, () -> {
							try(BasicContainerFileHandle handle = new BasicContainerFileHandle(file)) {
								handle.readContainerHeader();
								updateContainerIndex(handle, containerHeaderIndex);
								if(rebuildResult != null) {
									rebuildResult.containerRead++;
								}
								containerIndexToMinDataIndex.put(handle.getContainerId(), getMinimumDataIndex(handle));
							}
						});}
						catch(PersistenceException e) {
							_debug.warning("Container " + file + " konnte nicht gelesen werden und wird ignoriert.");
						}
					}
				}
			}

			HashSet<Long> knownMinDataIndexes = new HashSet<>((int) containerHeaderIndex.numEntries());

			// Inhalte überprüfen, Fehler wie doppelte offene Container erkennen und beheben
			IndexResult<ContainerManagementInformation> contents = containerHeaderIndex.query();
			for(int i = 0 ; i < contents.size(); i++) {
				if(hasError(contents,
				            i,
				            knownMinDataIndexes,
				            containerIndexToMinDataIndex,
				            !containerFileDir.getArchiveDataKind().isRequested())) {
					
					long containerId = contents.get(i, ContainerManagementInformation.CHP_CONT_ID);
					
					Path contFile = getPath(containerFileDir).resolve(getContFileName(containerId));

					// bekanntes Problem aus alten Archivversionen, höheres Warnlevel würde vermutlich Anwender irritieren 
					_debug.fine("Container \"" + contFile + "\" enthält inkonsistente Daten und wird ignoriert.");

					if(_deleteBrokenContainers) {
						renameContainerFile(contFile);
					}
					if(rebuildResult != null) {
						rebuildResult.containerIgnored++;
					}

					containerHeaderIndex.removeAll(ImmutableMap.of(ContainerManagementInformation.CHP_CONT_ID, new LongRange(containerId, containerId)));
				}
			}
		}
		catch(Exception e) {
			throw new IndexException("IO-Fehler beim Erstellen des Verwaltungsdaten-Index", containerHeaderIndex.getFile(), e);
		}
	}

	/**
	 * Die Methode getMinimumDataIndex bestimmt den minimalen Datenindex aus einer Containerdatei
	 *
	 * @param containerFileHandle von Typ BasicContainerFileHandle  
	 * @return long Datenindex
	 * @throws PersistenceException wenn ein Fehler beim Dateizugriff auftritt  
	 * @throws SynchronizationFailedException wenn die Synchronisierung fehlgeschlagen ist (sollte nicht auftreten)  
	 */
	private static long getMinimumDataIndex(final BasicContainerFileHandle containerFileHandle) throws PersistenceException, SynchronizationFailedException {
		int numDs = containerFileHandle.getContainerHeaderParamAsInt(ContainerManagementInformation.CHP_ANZ_DS);
		if(numDs == ContainerHdr.CONT_UNCLOSED) {
			// Im Container nachsehen
			try(DataIterator iterator = containerFileHandle.iterator()) {
				return iterator.peekDataIndex();
			}
		}
		return containerFileHandle.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_DATA_IDX_MIN);
	}

	/**
	 * Bestimmt, ob ein Eintrag im Verwaltungsdatenindex fehlerhaft ist
	 * @param contents Indexinhalte
	 * @param i                   Zu prüfende Zeile im Index
	 * @param knownMinDataIndexes Bereits bekannte Minimum-Datenindex-Werte im Container (Duplikate sind ein Fehler), dieses Set wird von der Methode ergänzt.
	 * @param contIdToMinDataIndex Bekannte Minimum-Datenindex-Werte je Container-ID
	 * @param checkIndexMonotonic  Soll geprüft werden ob der Datenindex-Minimum-Wert monoton ist?
	 * @return true falls Fehler vorhanden, sonst false
	 */
	public boolean hasError(final IndexResult<ContainerManagementInformation> contents, final int i, final HashSet<Long> knownMinDataIndexes,
	                        final Map<Long, Long> contIdToMinDataIndex, boolean checkIndexMonotonic) {
		
		// Container ist offen, obwohl es nicht der letzte Container ist -> Fehler
		if(i != contents.size() - 1 && contents.getSignedInt(i, ContainerManagementInformation.CHP_ANZ_DS) == ContainerHdr.CONT_UNCLOSED) {
			return true;
		}

		// oder Container hat einen doppelten DataIndexMin-Wert (darf nicht vorkommen)
		long myContId = contents.get(i, ContainerManagementInformation.CHP_CONT_ID);
		final Long dataIndexMinValue = contIdToMinDataIndex.get(myContId);
		
		if(dataIndexMinValue == null) {
			// Minimum-Datenindex nicht bekannt, Container gelöscht?
			return true;
		}
		
		// Datenindex nicht monoton?
		if(checkIndexMonotonic && i > 0) {
			long prevContId = contents.get(i - 1, ContainerManagementInformation.CHP_CONT_ID);
			final Long prevDataIndexMinValue = contIdToMinDataIndex.get(prevContId);
			if(prevDataIndexMinValue > 0 && dataIndexMinValue <= prevDataIndexMinValue) {
				// Datenindex ist kleiner als vorhergehender -> Fehler
				return true;
			}
		}

		return !knownMinDataIndexes.add(dataIndexMinValue);
	}

	/**
	 * Die Methode updateContainerIndex aktualisiert den Verwaltungsdatenindex
	 *
	 * @param containerFile Containerdaten, die in den Index eingefügt werden sollen (ersetzt ggf. vorhandene Einträge mit gleicher CintainerID)  
	 * @param index Verwaltungsdatenindex
	 * @throws PersistenceException wenn ein Fehler auftritt  
	 */
	private static void updateContainerIndex(final ContainerManagementData containerFile, final ContainerManagementIndex index) throws PersistenceException {
		for(ContainerManagementInformation information : ContainerManagementInformation.values()) {
			if(information.isNumeric()) {
				long paramAsLong = containerFile.getContainerHeaderParamAsLong(information);
				index.setInsertValue(information, paramAsLong);
			}
			else {
				String paramAsString = containerFile.getContainerHeaderParamAsString(information);
				index.setInsertValue(information, paramAsString);
			}
		}

		try {
			index.insertOrReplace();
		}
		catch(IndexException e) {
			throw new PersistenceException("Fehler beim Aktualisieren der Header-Informationen in der Indexdatei", e);
		}
	}

	/**
	 * Die Methode getIndexResult führt eine Index-Abfrage durch (für Archivanfragen) und aktualisiert dabei die Daten vom offenen Container im Index
	 *
	 * @param key von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param adk von Typ ArchiveDataKind  
	 * @param ats von Typ ArchiveTimeSpecification  
	 * @return IndexResult&lt;IndexValues&gt;   
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public IndexResult<IndexValues> getIndexResult(final SyncKey<IdDataIdentification> key, final ArchiveDataKind adk, final ArchiveTimeSpecification ats) throws IndexException, SynchronizationFailedException {
		StandardOpenContainerData openContainerData = getOpenContainerData(new IdContainerFileDir(key.getElement(), adk));
		if(openContainerData != null) {
			// Indexe auf den aktuellen Stand bringen, damit die Anfrage aktuelle Daten zurückliefert 
			updateStandardIndexes(key, adk, openContainerData);
		}
		return getIndexResultWithoutUpdate(key, adk, ats);
	}
	
	/**
	 * Die Methode getIndexResultWithoutUpdate führt eine Index-Abfrage durch (für Archivanfragen) und greift dabei direkt auf die Daten aus der Indexdatei zurück. Das heißt, dass die Daten des offenen Containers möglicherweise falsch/verhaltet sind, da der Index nicht ständig aktualisiert wird. 
	 *
	 * @param key von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param adk von Typ ArchiveDataKind  
	 * @param ats von Typ ArchiveTimeSpecification  
	 * @return IndexResult&lt;IndexValues&gt;   
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	private IndexResult<IndexValues> getIndexResultWithoutUpdate(final SyncKey<IdDataIdentification> key, final ArchiveDataKind adk, final ArchiveTimeSpecification ats) throws IndexException {
		if(ats.isStartRelative()) throw new UnsupportedOperationException("Ein relativer Intervallstart wird bei dieser Aktion nicht unterstützt.");
		if(ats.getTimingType().equals(TimingType.DATA_INDEX)) {
			return _indexTree.getDataIndexIndex(key, adk).getContainerIDByDataIndex(ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else if(ats.getTimingType().equals(TimingType.DATA_TIME)) {
			IndexResult<IndexValues> contIDs = _indexTree.getDataTimeIndex(key, adk).getContainerIDByDataTime(ats.getIntervalStart(), ats.getIntervalEnd());
			if(contIDs.isEmpty()) {
				// Falls keine Container ermittelt werden konnte, den ersten Container zurückliefern, damit der erste Datensatz übertragen werden kann
				// Die Anforderungen sind nicht ganz klar, aber der Test de.bsvrz.ars.ars.TestArchiveRelative_Request.testEndeVorKleinstem()
				// deutet darauf hin, dass bei einer Archivanfrage mit DATA-TIME der erste Datensatz zurückgeliefert werden soll, wenn der Anfragebereich
				// vor dem ersten Datensatz liegt.
				return _indexTree.getDataTimeIndex(key,adk).first();
			}
			// Daten, die zwischen Anfang und Ende liegen, aber wegen Zeitrücksprüngen nicht mehr im Anfragebereich liegen, trotzdem mitberücksichtigen,
			// daher hier nochmal anhand des DatenIndex-Indexes die Container bestimmen.
			return _indexTree.getDataIndexIndex(key, adk).getContainerIDByDataIndex(contIDs.getMin(IndexValues.DataIndexMin), contIDs.getMax(IndexValues.DataIndexMax));
		}
		else if(ats.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
			return _indexTree.getArchiveTimeIndex(key, adk).getContainerIDByArchiveTime(ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else {
			return IndexResult.empty();
		}
	}


	/**
	 * Die Methode updateContainerIndex aktualisiert den Verwaltungsdatenindex
	 *
	 * @param managementData Containerdaten, die in den Index eingefügt werden sollen (ersetzt ggf. vorhandene Einträge mit gleicher ContainerID)  
	 * @param lock Schlüssel zum Zugriff auf den Verwaltungsdatenindex
	 * @throws PersistenceException wenn ein Fehler auftritt
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public void updateContainerIndex(final SyncKey<IdDataIdentification> lock, final ContainerManagementData managementData) throws PersistenceException, IndexException {
		updateContainerIndex(managementData, _indexTree.getContainerManagementIndex(lock, managementData.getLocation().getArchiveDataKind()));
	}

	/**
	 * Die Methode maxMedienID ermittelt die maximale Medien-ID in einem Index-Abfrage-Ergebnis
	 *
	 * @param headers von Typ IndexResult&lt;ContainerManagementInformation&gt;  
	 * @return Medien-ID oder -1 falls Index keine Medien-IDs enthält (also z. B. nie etwas gesichert wurde)
	 */
	public int maxMedienID(final IndexResult<ContainerManagementInformation> headers) {
		int result = -1;
		for(int i = 0; i < headers.size(); i++){
			String mediumIdStr = headers.getString(i, ContainerManagementInformation.CHP_MEDIUM_ID);
			if(mediumIdStr.equals(ContainerHdr.MEDIUM_UNSAVED) || mediumIdStr.equals(ContainerHdr.MEDIUM_DESTROYED)) {
				continue;
			}
			result = Math.max(result, Integer.parseInt(mediumIdStr));
		}
		return result;
	}

	/**
	 * Ermittelt maximale Container-ID
	 *
	 * @param indexKey Indexzugriffschlüssel
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 * @return Größte Container-ID im Verzeichnis
	 * @throws IndexException Lesefehler der Indexe (z. B. korrupt)
	 */
	public long maxContainerId(final SyncKey<IdDataIdentification> indexKey, ArchiveDataKind adk) throws IndexException {
		return _indexTree.getContainerManagementIndex(indexKey, adk).getLast(ContainerManagementInformation.CHP_CONT_ID);
	}

	/**
	 * Die Methode rebuildIndexes baut einen Index erneut auf.
	 *
	 * @param lock von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param adk von Typ ArchiveDataKind  
	 * @param rebuildResult Ergebnis für Statistiken
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public void rebuildIndexes(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind adk, @Nullable RebuildResult rebuildResult) throws IndexException {
		if(_skipRebuild) {
			_indexTree.deleteIndex(lock, adk, rebuildResult);
		}
		else {
			_indexTree.recreateIndex(lock, adk, rebuildResult);
		}
	}

	/**
	 * Die Methode addIndexEntries fügt einem Index Werte hinzu.
	 *
	 * @param lock von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param adk von Typ ArchiveDataKind  
	 * @param dataIdx von Typ long  
	 * @param arsTime von Typ long  
	 * @param dataTime von Typ long  
	 * @param openContID von Typ long  
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public void addIndexEntries(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind adk, final long dataIdx, final long arsTime, final long dataTime, final long openContID) throws IndexException {
		final StandardOpenContainerData openContainerData = new StandardOpenContainerData(1, arsTime, arsTime, dataTime, dataTime, dataIdx, dataIdx, openContID);
		setOpenContainerData(lock, adk, openContainerData);
		updateStandardIndexes(lock, adk, openContainerData);
	}

	/**
	 * Die Methode updateStandardIndexes fügt einem Index Werte hinzu oder aktualisiert die vorhandenen Werte.
	 *
	 * @param lock von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param adk von Typ ArchiveDataKind  
	 * @param data von Typ OpenContainerData  
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public void updateStandardIndexes(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind adk, final OpenContainerData data) throws IndexException {
		final List<BaseIndex<IndexValues>> indices = new ArrayList<>(3);
		indices.add(_indexTree.getArchiveTimeIndex(lock, adk));
		indices.add(_indexTree.getDataTimeIndex(lock, adk));
		if(adk.isRequested()) {
			indices.add(_indexTree.getDataIndexIndex(lock, adk));
		}
		for(BaseIndex<IndexValues> index : indices) {
			index.setInsertValue(IndexValues.ContainerId, data.getContainerId());
			index.setInsertValue(IndexValues.DataTimeMin, data.getMinDataTime());
			index.setInsertValue(IndexValues.DataTimeMax, data.getMaxDataTime());
			index.setInsertValue(IndexValues.ArchiveTimeMin, data.getMinArcTime());
			index.setInsertValue(IndexValues.ArchiveTimeMax, data.getMaxArcTime());	
			index.setInsertValue(IndexValues.DataIndexMin, data.getMinDataIdx());
			index.setInsertValue(IndexValues.DataIndexMax, data.getMaxDataIdx());
			index.insertOrReplace(true); // Erstmal prüfen, ob in alle Indexe eingefügt werden kann
		}
		for(BaseIndex<IndexValues> index : indices) {
			index.insertOrReplace(); // Einfügen
		}
	}

	/**
	 * Die Methode speichert im {@link OpenContainerData} neue Maximal-Werte für den offenen Container (Maximal-Datenindex usw.) 
	 *
	 * @param idDataIdentification Datenidentifikation
	 * @param adk von Typ ArchiveDataKind  
	 * @param dataIdx Datenindex
	 * @param arsTime Archivzeit
	 * @param dataTime Datenzeit
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public void updateMaxValues(final IdDataIdentification idDataIdentification, final ArchiveDataKind adk, final long dataIdx, final long arsTime, final long dataTime) throws IndexException, SynchronizationFailedException {
		StandardOpenContainerData openContainerData = getOpenContainerData(new IdContainerFileDir(idDataIdentification, adk));
		if(openContainerData != null) {
			openContainerData.updateMax(dataIdx, arsTime, dataTime);
		}
	}

	/**
	 * Die Methode getOpenContID gibt die ID des offenen Containers für eine Datenidentifikation zurück
	 *
	 * @param idContainerFileDir Containerverzeichnis  
	 * @return long oder -1 falls es keinen offenen Container gibt
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public long getOpenContID(final IdContainerFileDir idContainerFileDir) throws IndexException, SynchronizationFailedException {
		OpenContainerData openContainerData = getOpenContainerData(idContainerFileDir);
		if(openContainerData != null) {
			return openContainerData.getContainerId();
		}
		else {
			return -1;
		}
	}

	/**
	 * Die Methode maxATime gibt die maximale Archivzeit einer datenidentifikation zurück
	 *
	 * @param idContainerFileDir Containerverzeichnis
	 * @return long
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	public long maxATime(final IdContainerFileDir idContainerFileDir) throws IndexException, SynchronizationFailedException {
		OpenContainerData openContainerData = getOpenContainerData(idContainerFileDir);
		if(openContainerData != null) {
			return openContainerData.getMaxArcTime();
		}
		else {
			try(SyncKey<IdDataIdentification> syncKey = lockIndex(idContainerFileDir.getDataIdentification())) {
				return _indexTree.getArchiveTimeIndex(syncKey, idContainerFileDir.getArchiveDataKind()).getLast(IndexValues.ArchiveTimeMax);
			}
		}
	}

	/**
	 * Die Methode getOpenContainerData die zwischengespeicherten daten für den offenen Container zurück
	 *
	 * @param containerFileDir Containerverzeichnis
	 * @return StandardOpenContainerData oder null falls kein offener Container existiert
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	@Nullable
	public StandardOpenContainerData getOpenContainerData(final IdContainerFileDir containerFileDir) throws IndexException, SynchronizationFailedException {
		
		// Der folgende Code sieht unnötig kompliziert aus, es ist aber wichtig, dass wir die initOpenContainerData()-Methode nicht innerhalb
		// von computeIfAbsent ausführen und dass umgekehrt computeIfAbsent nicht ausgeführt wird, während auf den Index synchronisiert wird.
		// Sonst kann sich die ConcurrentHashMap-interne Synchronisierung mit dem Lock hier verhaken.
		
		StandardOpenContainerData data = _openContainerData.get(containerFileDir);
		
		if(data == null) {
			// StandardOpenContainerData muss aus Container berechnet werden
			
			StandardOpenContainerData tmp;
			
			try(SyncKey<IdDataIdentification> syncKey = lockIndex(containerFileDir.getDataIdentification())) {
				ContainerManagementIndex containerManagementIndex = _indexTree.getContainerManagementIndex(syncKey, containerFileDir.getArchiveDataKind());
				tmp = initOpenContainerData(containerFileDir, containerManagementIndex);
			}
			
			data = _openContainerData.computeIfAbsent(containerFileDir, (dir) -> tmp);
		}
		
		//noinspection ObjectEquality
		if(data == NO_OPEN_CONTAINER) {
			return null;
		}
		return data;
	}

	/**
	 * Die Methode getOpenContainerData ermittelt die Daten vom offenen Container
	 *
	 * @param containerFileDir von Typ IdContainerFileDir  
	 * @param containerManagementIndex von Typ ContainerManagementIndex  
	 * @return StandardOpenContainerData
	 */
	@Nullable
	private StandardOpenContainerData getOpenContainerData(final IdContainerFileDir containerFileDir, final ContainerManagementIndex containerManagementIndex) {
		// Der folgende Code sieht unnötig kompliziert aus, es ist aber wichtig, dass wir die initOpenContainerData()-Methode nicht innerhalb
		// von computeIfAbsent ausführen und dass umgekehrt computeIfAbsent nicht ausgeführt wird, während auf den Index synchronisiert wird.
		// Sonst kann sich die ConcurrentHashMap-interne Synchronisierung mit dem Lock hier verhaken.

		StandardOpenContainerData data = _openContainerData.get(containerFileDir);

		if(data == null) {
			// StandardOpenContainerData muss aus Container berechnet werden
			StandardOpenContainerData tmp = initOpenContainerData(containerFileDir, containerManagementIndex);

			data = _openContainerData.computeIfAbsent(containerFileDir, (dir) -> tmp);
		}

		//noinspection ObjectEquality
		if(data == NO_OPEN_CONTAINER) {
			return null;
		}
		return data;
	}

	/**
	 * Die Methode initOpenContainerData initialisiert die Daten eines geöffneten Containers. Da die Header des offenen Containers nicht aussagekräftig sind, muss im offenen Container im Regelfall über die Datensätze iteriert werden. 
	 *
	 * @param dir Containerverzeichnis
	 * @param headerIndex Verwaltungsdatenindex, um zu bestimmen welcher Container der offene ist (oder ob es überhaupt einen offenen gibt)
	 * @return StandardOpenContainerData
	 */
	private StandardOpenContainerData initOpenContainerData(final IdContainerFileDir dir, final ContainerManagementIndex headerIndex) {

		try {
			if(headerIndex.numEntries() == 0) {
				// Diese DID wurde überhaupt noch nie archiviert, daher letzten Eintrag nicht weiterverwenden (welchen auch?)
				return NO_OPEN_CONTAINER;
			}
			long lastContId = headerIndex.getLast(ContainerManagementInformation.CHP_CONT_ID);
			boolean isOpen = (int) headerIndex.getLast(ContainerManagementInformation.CHP_ANZ_DS) == ContainerHdr.CONT_UNCLOSED;

			if(!isOpen) {
				// Container abgeschlossen, letzten Eintrag nicht weiterverwenden
				return NO_OPEN_CONTAINER;
			}
			return getContainerDataFromFile(dir, lastContId);
		}
		catch(Exception e) {
			throw new UncheckedExecutionException(e);
		}
	}

	private StandardOpenContainerData getContainerDataFromFile(final IdContainerFileDir dir, final long containerId) throws SynchronizationFailedException, PersistenceException {
		try (SyncKey<IdDataIdentification> indexLock = lockIndex(dir.getDataIdentification())) {
			// es existiert ein offener Container
			try (ContainerFileHandle handle = accessContainer(dir, containerId)) {
				// Anzahl im Container nachzählen, dabei lastNoSrc-Flag mitnehmen
				int anz = 0;
				long minArchiveTime = Long.MAX_VALUE;
				long maxArchiveTime = -1;
				long minDataTime = Long.MAX_VALUE;
				long maxDataTime = -1;
				long minDataIndex = Long.MAX_VALUE;
				long maxDataIndex = -1;
				boolean lastDataIsNoSrc = false;
				try (DataIterator iter = handle.iterator()) {
					ContainerDataResult result = new ContainerDataResult();
					while (!iter.isEmpty()) {
						anz++;
						iter.poll(result);

						minArchiveTime = Math.min(minArchiveTime, result.getArchiveTime());
						maxArchiveTime = Math.max(maxArchiveTime, result.getArchiveTime());
						minDataTime = Math.min(minDataTime, result.getDataTime());
						maxDataTime = Math.max(maxDataTime, result.getDataTime());
						minDataIndex = Math.min(minDataIndex, result.getDataIndex());
						maxDataIndex = Math.max(maxDataIndex, result.getDataIndex());

						lastDataIsNoSrc = result.isNoSource();
					}
				} catch (PersistenceException ex) {
					_debug.fine("Offener Container kann nicht gelesen werden", ex);
					// Wenn der Container hier nicht vollständig lesbar ist, kann das ignoriert werden
					// der Container (und der zugehörige Index-Eintrag) wird später in #checkContainerFile korrigiert.
				}

				if (anz == 0) {
					return NO_OPEN_CONTAINER;
				}

				if (dir.getArchiveDataKind() == ArchiveDataKind.ONLINE) {
					_dataIdentTree.get(dir.getDataIdentification()).setLastOAWasNoSource(lastDataIsNoSrc);
				}

				// Container nicht abgeschlossen, letzten Eintrag weiterverwenden:
				return new StandardOpenContainerData(anz, minArchiveTime, maxArchiveTime, minDataTime, maxDataTime, minDataIndex, maxDataIndex,
				                                     containerId);
			}
		}
	}

	/**
	 * Setzt die Daten eines offenen Containers. Diese Methode wird aufgerufen, wenn es keinen offenen Container mehr gibt.
	 * @param lock Synchroniserungszugriffsschlüssel
	 * @param adk  Archivdatenart
	 * @param data Neue Daten   
	 */
	private void setOpenContainerData(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind adk, StandardOpenContainerData data) {
		_openContainerData.put(new IdContainerFileDir(lock.getElement(), adk), data);
	}

	/**
	 * Löscht die Daten eines offenen Containers. Diese Methode wird aufgerufen, wenn es keinen offenen Container mehr gibt.
	 * @param lock Synchroniserungszugriffsschlüssel
	 * @param adk  Archivdatenart
	 */
	@Nullable
	public void removeOpenContainerData(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind adk) {
		_openContainerData.put(new IdContainerFileDir(lock.getElement(), adk), NO_OPEN_CONTAINER);
	}

	/**
	 * Die Methode removeContainerFromIndex löscht einen Contaienr aus den Indexen
	 *
	 * @param lock von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param archiveDataKind von Typ ArchiveDataKind  
	 * @param containerId von Typ long  
	 * @throws IndexException wenn der Indexzugriff fehlschlägt  
	 */
	public void removeContainerFromIndex(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind, final long containerId) throws IndexException {
		DataTimeIndex dtIdx = _indexTree.getDataTimeIndex(lock, archiveDataKind);
		if(dtIdx != null) {
			dtIdx.removeAll(ImmutableMap.of(IndexValues.ContainerId, new LongRange(containerId, containerId)));
		}

		if(archiveDataKind.isDelayed()) {
			DataIndexIndex diIdx = _indexTree.getDataIndexIndex(lock, archiveDataKind);
			if(diIdx != null) {
				diIdx.removeAll(ImmutableMap.of(IndexValues.ContainerId, new LongRange(containerId, containerId)));
			}
		}

		ArchiveTimeIndex aTimeMonoIdx = _indexTree.getArchiveTimeIndex(lock, archiveDataKind);
		if(aTimeMonoIdx != null) {
			aTimeMonoIdx.removeAll(ImmutableMap.of(IndexValues.ContainerId, new LongRange(containerId, containerId)));
		}

		ContainerManagementIndex headerIdx = _indexTree.getContainerManagementIndex(lock, archiveDataKind);
		if(headerIdx != null) {
			headerIdx.removeAll(ImmutableMap.of(ContainerManagementInformation.CHP_CONT_ID, new LongRange(containerId, containerId)));
		}
	}

	/**
	 * Die Methode numContainers zählt die Anzahl Einträge im Verwaltungsdatenindex für ein Containerverzeichnis.
	 * Dies ist die Menge der potentiell vorhandenen Container, ausgelagerte Container werden dabei mitgezählt.
	 *
	 * @param lock Synchronisierung auf Datenidentifikation
	 * @param dataKind Archivdatenart  
	 * @return long Anzahl Containerdateien
	 * @throws IndexException wenn der Indexzugriff fehlschlägt  
	 */
	public long numContainers(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind dataKind) throws IndexException {
		return _indexTree.getContainerManagementIndex(lock, dataKind).numEntries();
	}

	/**
	 * Die Methode getLastContainerHeaders ermittelt die Header vom letzten Container (vom Container mti der größten ID)
	 *
	 * @param indexLock Synchronisierung auf Datenidentifikation
	 * @param adk Archivdatenart  
	 * @return ContainerHeaders
	 * @throws IndexException wenn der Indexzugriff fehlschlägt  
	 */
	@Nullable
	public ContainerHeaders getLastContainerHeaders(final SyncKey<IdDataIdentification> indexLock, final ArchiveDataKind adk) throws IndexException {
		ContainerManagementIndex index = _indexTree.getContainerManagementIndex(indexLock, adk);
		if(index.numEntries() == 0) return null;
		long contId = index
				.getLast(ContainerManagementInformation.CHP_CONT_ID);
		return getContainerHeaders(indexLock, adk, contId);
	}

	/**
	 * Die Methode arSParamGetVorhalten ermittelt den parametrierten Vorhaltezeitraum in Sekunden für eine Datenidentifikation.
	 *
	 * @param dataIdentification Datenidentifikation
	 * @return long Vorhaltezeitraum in Sekunden
	 */
	public long arSParamGetVorhalten(final IdDataIdentification dataIdentification) {
		return _dataIdentTree.get(dataIdentification).arSParamGetVorhalten();
	}

	/**
	 * Gibt Statistiken von diesem PersistenceManager-Objekt zurück.
	 *
	 * @return Statistiken (Typ Statistics)
	 */
	public Statistics getStatistics() {
		return new Statistics();
	}

	/**
	 * Die Methode warnAboutHugeContainerDirectory warnt den Benutzer, wenn zu viele Container in einem Containerverzeichnis gespeichert wurden.
	 *
	 * @param location von Typ IdContainerFileDir  
	 * @param numEntries von Typ long  
	 */
	public void warnAboutHugeContainerDirectory(final IdContainerFileDir location, final long numEntries) {

		_largeDataIdentifications.add(location);
		
		_debug.warning("Für die Datenidentifikation " + formatDataIdentification(location)
		 + " im Verzeichnis '" + getPath(location) + "' sind eine große Zahl (" + numEntries + ") Container vorhanden, was zu Problemen und schlechter Performance führen kann. Es sollte geprüft werden, ob eine Software übermäßig viele Daten generiert, ob viele Datenzeitsprünge auftreten oder ob der Vorhaltezeitraum verkürzt werden kann. Falls fehlerhafte Daten archiviert wurden, sollten diese ggf. endgültig gelöscht werden.");

	}
	
	/**
	 * Die Methode warnAboutSlowRestore warnt den Benutzer, wenn die Wiederherstellung eines Verzeichnisses sehr lange dauert.
	 *
	 * @param location von Typ IdContainerFileDir  
	 * @param currentEntries von Typ long  
	 * @param currentDuration von Typ Duration  
	 */
	public void warnAboutSlowRestore(final IdContainerFileDir location, final long currentEntries, Duration currentDuration) {

		_largeDataIdentifications.add(location);
		
		_debug.warning("Die Wiederherstellung des Verwaltungsdatenindexes für die Datenidentifikation " + formatDataIdentification(location)
				               + " im Verzeichnis '" + getPath(location) + "' benötigt länger als " + currentDuration.toMinutes() + " Minuten. Es wurden bereits " + currentEntries + " Containerdateien indiziert. Es sollte geprüft werden, ob eine Software übermäßig viele Daten generiert, ob viele Datenzeitsprünge auftreten oder ob der Vorhaltezeitraum verkürzt werden kann. Falls fehlerhafte Daten archiviert wurden, sollten diese ggf. endgültig gelöscht werden.");

	}

	/**
	 * Gibt ein Set zurück, dass alle bekannten Contaienrverzeichnisse enthält, die "übermäßig viel" Daten enthalten.
	 * Das Set ist threadsafe und änderbar.
	 * @return Set mit Containerverzeichnissen
	 */
	public Set<IdContainerFileDir> getLargeDataIdentifications() {
		return _largeDataIdentifications;
	}

	/**
	 * Die Methode formatDataIdentification gibt eine Datenidentifikation als lesbaren String aus (z. B. ermitteln der Pid falls möglich) 
	 *
	 * @param location Containerverzeichnis-ID
	 * @return String
	 */
	@NotNull
	public String formatDataIdentification(final IdContainerFileDir location) {
		return formatObj(location.getObjectId()) + ":" + formatObj(location.getAtgId()) + ":" + formatObj(location.getAspectId()) + " (" + location.getArchiveDataKind() + ")";
	}
	
	/**
	 * Die Methode formatObj gibt eine Objekt-ID als lesbaren String aus (z. B. ermitteln der Pid falls möglich) 
	 *
	 * @param objId von Typ long  
	 * @return String
	 */
	public String formatObj(final long objId) {
		ArchiveManager am = this.archivMgr;
		if(am != null) {
			ClientDavInterface davCon = am.getDavCon();
			if(davCon != null) {
				DataModel dataModel = davCon.getDataModel();
				if(dataModel != null) {
					SystemObject obj = dataModel.getObject(objId);
					if(obj != null) {
						return obj.getPidOrId();
					}
				}
			}
		}
		return String.valueOf(objId);
	}

	/**
	 * Schreibt den aktuellen Stand der Indexe auf die Festplatte
	 * @param lock Lock für Datenidentifikation
	 * @param adk  Archivdatenart
	 * @throws IndexException Falls ein Indexfehler auftrat.
	 */
	public void flushIndexes(SyncKey<IdDataIdentification> lock, ArchiveDataKind adk) throws IndexException {
		_indexTree.flushIndexes(lock, adk);
	}

	/**
	 * Funktionales Interface für Container-Aktionen
	 */
	@FunctionalInterface
	public interface ContainerCall {
		/**
		 * Methode, die etwas ausführt und dabei Exceptions werfen kann
		 */
		void call() throws PersistenceException, SynchronizationFailedException;
	}

	/**
	 * Die Methode getHeaders gibt die Header eines Containers zurück.
	 *
	 * @param contHdrIdx von Typ ContainerManagementIndex  
	 * @param containerID von Typ long  
	 * @param dataIdentification von Typ IdDataIdentification  
	 * @param adk von Typ ArchiveDataKind  
	 * @return ContainerHeaders
	 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
	 */
	private static ContainerHeaders getHeaders(final ContainerManagementIndex contHdrIdx, final long containerID, final IdDataIdentification dataIdentification, final ArchiveDataKind adk) throws IndexException {
		IndexResult<ContainerManagementInformation> query;
		query = contHdrIdx.query(ImmutableMap.of(ContainerManagementInformation.CHP_CONT_ID, new LongRange(containerID, containerID)));
		// Die Query könnte einen vorherigen Container zurückgeliefert haben
		if(query.size() > 0 && query.get(0, ContainerManagementInformation.CHP_CONT_ID) == containerID) {
			return new ContainerHeaders(query, dataIdentification, adk, 0);
		}
		throw new IllegalArgumentException("Container " + containerID + " ist nicht im Index enthalten.");
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Prüfprozedur für das Persistenzverzeichnis
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Das Logfile {@link #isActive} anlegen. Wird verwendet, um einem Archivsystem anzuzeigen, ob ein anderes Archivsystem dasselbe Persistenzverzeichnis
	 * verwendet
	 *
	 * @return Fehlermeldung, wenn das Erstellen nicht erfolgreich war, null sonst
	 */
	@Nullable
	public String createLockFile() {
		isActive = new File(archivePath.toFile(), ACTIVITY_FLAG_FILE_NAME);
		if(isActive.exists()) {
			return "Die Sperrdatei \"" + isActive.getAbsolutePath() + "\" zeigt an, dass bereits ein Archivsystem im Persistenzverzeichnis aktiv ist.";
		}
		try {
			if(isActive.createNewFile()) {
				return null;
			}
			else {
				return "Sperrdatei \"" + isActive.getAbsolutePath() + "\" konnte nicht erstellt werden.";
			}
		}
		catch(Exception e) {
			return "Fehler beim Erstellen der Sperrdatei \"" + isActive.getAbsolutePath() + "\": " + e.getMessage();
		}
	}

	/**
	 * Löscht das LogFile {@link #isActive}, falls es von diesem ArS angelegt wurde und es existiert und gibt damit das Persistenzverzeichnis wieder frei
	 *
	 */
	public void deleteLockFile() {
		if(isActive == null) {
			return;
		}
		if(!isActive.delete()) {
			_debug.warning(
					"Sperrdatei \"" + isActive.getAbsolutePath() + "\" konnte zur Freigabe des " + "Persistenzverzeichnisses nicht gelöscht werden."
			);
		}
	}

	/**
	 * Prüft das Persistenzverzeichnis auf fremde Dateien und Verzeichnisse und falsche Reihenfolgen von ID-Verzeichnissen. Die Ergebnisse werden über den Logger
	 * protokolliert. Kann durch einen Kommandozeilenparameter beim Systemstart aufgerufen werden.
	 * @param mode Ein Wert von P_CHECK_PERS_DIR_DIR, P_CHECK_PERS_DIR_HDR, P_CHECK_PERS_DIR_ALL
	 */
	public void checkPersistenceDir(String mode) throws InterruptedException {
		AtomicBoolean allValid = new AtomicBoolean(true);
		AtomicInteger numVisited = new AtomicInteger(0);
		directoryCount = 0;
		DataIdentTree countingDit = new DataIdentTree();
		File archiveDir = archivePath.toFile();

		ExecutorService threadPool = Executors.newFixedThreadPool(archivMgr.getNumCheckPersistenceThreads());

		File[] files = archiveDir.listFiles();
		if(files == null) files = new File[0];
		for(File dir : files) {
			boolean valid = isValidTopLevelArSFile(dir.getName());
			if(!valid) _debug.fine("Fremde Datei/Verzeichnis im Persistenzverzeichnis: '" + dir.getAbsolutePath() + "'");
			if(!valid) allValid.set(false);
			if(dir.isDirectory()) {
				checkPersDirNode(dir, countingDit, mode, threadPool, allValid, numVisited);
			}
		}

		threadPool.shutdown();

		while(!threadPool.awaitTermination(10, TimeUnit.SECONDS)) {
			_debug.info(numVisited + " Datenidentifikationen analysiert");
		}

		_debug.info(
				"# " + directoryCount + " durchsuchte Verzeichisse" + Debug.NEWLINE + "# " + countingDit.size()
						+ " Datenidentifikationen gefunden." + Debug.NEWLINE + "# Ergebnis: " + (allValid.get()
						? "Das Archiv ist in einem konsistenten Zustand."
						: "Es sind Fehler aufgetreten (siehe LogLevel FINE, FINER und FINEST für Details).")
		);
	}

	/**
	 * Die Methode checkPersDirNode überprüft einen Knoten im Persistenzverzeichnis
	 *
	 * @param curDir von Typ File  
	 * @param dit von Typ DataIdentTree  
	 * @param mode von Typ String  
	 * @param threadPool von Typ ExecutorService  
	 * @param allValid von Typ AtomicBoolean  
	 * @param numVisited von Typ AtomicInteger  
	 */
	private void checkPersDirNode(File curDir, DataIdentTree dit, String mode, final ExecutorService threadPool, final AtomicBoolean allValid, final AtomicInteger numVisited) {
		if(++directoryCount % 5000 == 0) _debug.info(directoryCount + " Verzeichisse indiziert");

		if(!isValidStorageArSFile(curDir.getName())) {
			_debug.fine("Fremdes Verzeichnis im Persistenzverzeichnis: '" + curDir.getAbsolutePath() + "'");
			allValid.set(false);
		}

		if(curDir.isDirectory()) {
			if(isValidSvFolderName(curDir.getName())) {
				threadPool.submit(new ConsistencyCheckWorker(curDir, dit, mode, allValid, numVisited));
			}
			else {
				File[] content = curDir.listFiles();
				if(content != null) {
					for(File dir : content) {
						checkPersDirNode(dir, dit, mode, threadPool, allValid, numVisited);
					}
				}
			}
		}
	}


	/**
	 * Extrahiert Objekt-ID, Attributgruppen-ID, Aspekt-ID und SimVar von unten beginnend aus einem Verzeichnispfad, der mindestens die Länge 4 haben muss.
	 * Voraußetzung ist, dass alle Verzeichnisnamen entsprechend den Konventionen des PersistenceManagers gebildet sind. Das unterste Verzeichnis kann entweder
	 * die SimVar bezeichnen oder eine Datensatzart sein.
	 *
	 * @param path Pfad bis zum Verzeichnis der Simulationsvariante. Falls ein Datenart-Verzeichnis angegeben wird, wir dieses ignoriert.
	 *
	 * @return Feld bestehend aus [ObjID, AtgID, AspID SimVar] wenn der Pfad ein gueltiges Verzeichnis ist
	 *
	 * @throws PersistenceException wenn der Pfad kein gueltiges Verzeichnis ist
	 * @see PersistenceManager#isValidObjIdFolderName(String)
	 * @see PersistenceManager#isValidAtgIdFolderName(String)
	 * @see PersistenceManager#isValidAspIdFolderName(String)
	 * @see PersistenceManager#isValidSvFolderName(String)
	 */
	public static IdDataIdentification getDataIdentificationFromPath(File path) throws PersistenceException {
		File tmpDir = path;
		if(tmpDir != null) {
			try {
				if(isValidDataKindSuffix(tmpDir.getName())) tmpDir = tmpDir.getParentFile();

				int sv;
				if(isValidSvFolderName(tmpDir.getName())) {
					sv = Integer.parseInt(getIDPart(tmpDir.getName()));
				}
				else {
					throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
				}

				tmpDir = tmpDir.getParentFile();
				StringBuilder aspString = new StringBuilder();
				StringBuilder atgString = new StringBuilder();
				StringBuilder objString = new StringBuilder();
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
					if(isValidAspIdFolderName(tmpDir.getName())) {
						aspString.insert(0, getIDPart(tmpDir.getName()));
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParentFile()) != null);
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
					if(isValidAtgIdFolderName(tmpDir.getName())) {
						atgString.insert(0, getIDPart(tmpDir.getName()));
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParentFile()) != null);
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
					if(isValidObjIdFolderName(tmpDir.getName())) {
						objString.insert(0, getIDPart(tmpDir.getName()));
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParentFile()) != null);

				return new IdDataIdentification(
						Long.parseLong(objString.toString()),
						Long.parseLong(atgString.toString()),
						Long.parseLong(aspString.toString()),
						sv
				);
			}
			catch(NumberFormatException ignored) {
			}
		}
		throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
	}
	
	/**
	 * Extrahiert Objekt-ID, Attributgruppen-ID, Aspekt-ID und SimVar von unten beginnend aus einem Verzeichnispfad, der mindestens die Länge 4 haben muss.
	 * Voraußetzung ist, dass alle Verzeichnisnamen entsprechend den Konventionen des PersistenceManagers gebildet sind. Das unterste Verzeichnis kann entweder
	 * die SimVar bezeichnen oder eine Datensatzart sein.
	 *
	 * @param path Pfad bis zum Verzeichnis der Simulationsvariante. Falls ein Datenart-Verzeichnis angegeben wird, wir dieses ignoriert.
	 *
	 * @return Feld bestehend aus [ObjID, AtgID, AspID SimVar] wenn der Pfad ein gueltiges Verzeichnis ist
	 *
	 * @throws PersistenceException wenn der Pfad kein gueltiges Verzeichnis ist
	 * @see PersistenceManager#isValidObjIdFolderName(String)
	 * @see PersistenceManager#isValidAtgIdFolderName(String)
	 * @see PersistenceManager#isValidAspIdFolderName(String)
	 * @see PersistenceManager#isValidSvFolderName(String)
	 */
	public static IdDataIdentification getDataIdentificationFromPath(Path path) throws PersistenceException {
		Path tmpDir = path.normalize();
		if(tmpDir != null) {
			try {
				String fileName = tmpDir.getFileName().toString();
				if(isValidDataKindSuffix(fileName)) tmpDir = tmpDir.getParent();
				fileName = tmpDir.getFileName().toString();

				int sv;
				if(isValidSvFolderName(fileName)) {
					sv = Integer.parseInt(getIDPart(fileName));
				}
				else {
					throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.toAbsolutePath() + "'");
				}

				tmpDir = tmpDir.getParent();
				StringBuilder aspString = new StringBuilder();
				StringBuilder atgString = new StringBuilder();
				StringBuilder objString = new StringBuilder();
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.toAbsolutePath() + "'");
					fileName = tmpDir.getFileName().toString();
					if(isValidAspIdFolderName(fileName)) {
						aspString.insert(0, getIDPart(fileName));
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParent()) != null);
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.toAbsolutePath() + "'");
					fileName = tmpDir.getFileName().toString();
					if(isValidAtgIdFolderName(fileName)) {
						atgString.insert(0, getIDPart(fileName));
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParent()) != null);
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.toAbsolutePath() + "'");
					fileName = tmpDir.getFileName().toString();
					if(isValidObjIdFolderName(fileName)) {
						objString.insert(0, getIDPart(fileName));
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParent()) != null);

				return new IdDataIdentification(
						Long.parseLong(objString.toString()),
						Long.parseLong(atgString.toString()),
						Long.parseLong(aspString.toString()),
						sv
				);
			}
			catch(NumberFormatException ignored) {
			}
		}
		throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.toAbsolutePath() + "'");
	}

	/**
	 * Prüft, ob der gegebene Verzeichnisname ein gueltiger Name innerhalb eines Pfades ist, der eine Objekt-ID repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#OBJ_DIRNAME_PAT
	 */
	private static boolean isValidObjIdFolderName(String dirName) {
		return OBJ_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Prüft, ob der gegebene Verzeichnisname ein gueltiger Name innerhalb eines Pfades ist, der eine Attributgruppen-ID repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#ATG_DIRNAME_PAT
	 */
	private static boolean isValidAtgIdFolderName(String dirName) {
		return ATG_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Prüft, ob der gegebene Verzeichnisname ein gueltiger Name innerhalb eines Pfades ist, der eine Aspekt-ID repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#ASP_DIRNAME_PAT
	 */
	private static boolean isValidAspIdFolderName(String dirName) {
		return ASP_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Prüft, ob der gegebene Verzeichnisname ein gueltiger Name für ein Verzeichnis ist, das eine Simulationsvariante repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#SV_DIRNAME_PAT
	 */
	public static boolean isValidSvFolderName(String dirName) {
		return SV_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Die Methode isValidIDFolderName früft ob ein Order einen gültigen Namen hat
	 *
	 * @param dirName von Typ String  
	 * @return boolean
	 */
	public static boolean isValidIDFolderName(String dirName) {
		return isValidObjIdFolderName(dirName) || isValidAtgIdFolderName(dirName) || isValidAspIdFolderName(dirName) || isValidSvFolderName(dirName);
	}

	/**
	 * Die Methode getIDPart gibt aus einem Ordnernamen den zahlen_Teil zurück (z. B. 123 aus "obj123")
	 *
	 * @param dirName von Typ String  
	 * @return String
	 */
	private static String getIDPart(String dirName) {
		Matcher m = NUM_DIR_PAT.matcher(dirName);
		if(m.find()) {
			return m.group(1);
		}
		else {
			throw new IllegalArgumentException(dirName);
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Methoden
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liefert den Wurzelpfad des Persistenzverzeichisses, der beim Systemstart in der Kommandozeile übergeben wurde.
	 *
	 * @return Wurzelpfad des Persistenzverzeichisses
	 */
	public Path getArchivePath() {
		return archivePath;
	}

	/**
	 * Erhoeht nextContainerID um 1 und gibt den Wert zurück.
	 *
	 * @return als nächste zu verwendende Container-ID
	 */
	public long nextContainerID() {
		return nextContainerID.incrementAndGet();
	}

	/** @return Letzte vergebe ContainerID */
	public long getLastContainerID() {
		return nextContainerID.longValue();
	}

	/**
	 * Sichert den letzten Zeitpunkt, wo gültige Daten empfangen wurden bzw. beim Beenden wo Daten abgemeldet wurden. Dieser Zeitpunkt wird nach einem Neustart als
	 * Grundlage für die Bildung von potentiellen Datenlücken verwendet.
	 */
	public void saveUnsubscriptionTime() {
		Multimap<Long, IdDataIdentification> didForUnsubscriptionTime = MultimapBuilder.treeKeys().arrayListValues().build();
		final Long now = System.currentTimeMillis();

		for(DataIdentNode node : _dataIdentTree) {
			if(node.hasValidData()) {
				didForUnsubscriptionTime.put(now, node.getDataIdentification());
			}
			else {
				long unsubscriptionTime = node.getUnsubscriptionTime();
				if(unsubscriptionTime != -1) {
					didForUnsubscriptionTime.put(unsubscriptionTime, node.getDataIdentification());
				}
			}
		}

		saveUnsubscriptionTime(didForUnsubscriptionTime, _unsubscriptionFile);
	}

	/**
	 * Sichert den letzten Zeitpunkt, wo gültige Daten empfangen wurden bzw. beim Beenden wo Daten abgemeldet wurden. Dieser Zeitpunkt wird nach einem Neustart als
	 * Grundlage für die Bildung von potentiellen Datenlücken verwendet.
	 * @param didForUnsubscriptionTime Zu schreibende Zeitstempel
	 * @param unsubscriptionFile       Datei, die geschrieben wird.
	 */
	public void saveUnsubscriptionTime(final Multimap<Long, IdDataIdentification> didForUnsubscriptionTime, final Path unsubscriptionFile) {
		try {
			try(DataOutputStream stream = new DataOutputStream(new BufferedOutputStream(new GZIPOutputStream(Files.newOutputStream(unsubscriptionFile))))) {
				Set<Map.Entry<Long, Collection<IdDataIdentification>>> entries = didForUnsubscriptionTime.asMap().entrySet();
				stream.writeInt(entries.size());
				for(Map.Entry<Long, Collection<IdDataIdentification>> entry : entries) {
					stream.writeLong(entry.getKey());
					List<IdDataIdentification> values = (List<IdDataIdentification>) entry.getValue();
					values.sort(null);
					stream.writeInt(values.size());
					for(IdDataIdentification value : values) {
						stream.writeLong(value.getObjectId());
						stream.writeLong(value.getAtgId());
						stream.writeLong(value.getAspectId());
						stream.writeShort(value.getSimVariant());
					}
				}
			}
		}
		catch(IOException e) {
			_debug.error("Fehler beim Schreiben der Datenzeit: " + RESTART_TIME_FILE_NAME + " ist nicht beschreibbar.", e);
		}
	}

	/**
	 * Lädt die mit {@link #saveUnsubscriptionTime()} gespeicherten Zeiten wieder ein und speichert sie im DataIdentTree. Hierbei können überflüssige Knoten entstehen, aber das ist nicht zu verhindern da wir nicht immer wissen,
	 * welche Zeiten wir noch brauchen.
	 * @throws IOException Allgemeiner IO-Fehler
	 */
	private void loadUnsubscriptionTime() throws IOException {
		try(DataInputStream stream = new DataInputStream(new BufferedInputStream(new GZIPInputStream(Files.newInputStream(_unsubscriptionFile))))) {
			int numEntries = stream.readInt();
			for(int i = 0; i < numEntries; i++) {
				long time = stream.readLong();
				int numValues = stream.readInt();
				for(int f = 0; f < numValues; f++) {
					long objId = stream.readLong();
					long atgId = stream.readLong();
					long aspId = stream.readLong();
					int sv = stream.readShort();
					_dataIdentTree.get(new IdDataIdentification(objId, atgId, aspId, sv)).setUnsubscriptionTime(time);
				}
			}
		}
	}

	/**
	 * Die Methode openDeletedContainerFile öffnet die Datei mit den Informationen zu den gelöschten Containern
	 *
	 * @param lock von Typ SyncKey&lt;IdDataIdentification&gt;  
	 * @param archiveDataKind von Typ ArchiveDataKind  
	 * @return DeletedContainerFile
	 * @throws PersistenceException Lesefehler
	 * @throws SynchronizationFailedException Fehler bei Index-Synchronisation
	 */
	public DeletedContainerFile openDeletedContainerFile(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind) throws PersistenceException, SynchronizationFailedException {
		IdContainerFileDir containerFileDir = new IdContainerFileDir(lock.getElement(), archiveDataKind);
		DeletedContainerFile deletedContainerFile = new DeletedContainerFile(containerFileDir, this);
		if(!deletedContainerFile.exists()) {
			try {
				initiallyBuildDeletedContainerFile(lock, containerFileDir, deletedContainerFile);
			}
			catch(IndexException e) {
				throw new PersistenceException(e);
			}
		}
		else {
			try {
				deletedContainerFile.read();
			}
			catch(PersistenceException e) {
				_debug.warning("Fehler beim Einlesen der Datei, die gelöschte Container enthält: " + deletedContainerFile.getFile() + ". Ungültige Einträge werden ignoriert.", e);
			}
		}
		return deletedContainerFile;
	}

	/**
	 * Erstellt erstmalig die Datei, die die gelöschten Container mit Datenindexbereichen enthält. Verkürzte Containerdateien werden dabei konvertiert
	 * und umgewandelt.
	 *
	 * @param lock                 Objekt, das den synchronisierten Zugriff auf den Index erlaubt
	 * @param idContainerFileDir   Containerdatenverzeichnis
	 * @param deletedContainerFile Datei, die erstellt werden soll
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 * @throws IndexException       Lesefehler der Indexe (z. B. korrupt)
	 */
	private void initiallyBuildDeletedContainerFile(final SyncKey<IdDataIdentification> lock, final IdContainerFileDir idContainerFileDir, final DeletedContainerFile deletedContainerFile) throws PersistenceException, IndexException, SynchronizationFailedException {
		IndexResult<ContainerManagementInformation> headers = getContainerHeaders(lock, idContainerFileDir.getArchiveDataKind());
		final List<Long> contsToDelete = new ArrayList<>();
		for(int i = 0; i < headers.size(); i++) {
			if(headers.getBoolean(i, ContainerManagementInformation.CHP_DELETED)) {
				long containerId = headers.get(i, ContainerManagementInformation.CHP_CONT_ID);
				try (ContainerFileHandle handle = accessContainer(idContainerFileDir, containerId)) {
					if(!handle.existsContainer()) continue;
					executeOnContainer(handle.getFile().toPath(), () -> {
						handle.readContainerHeader();
						// Datei öffnen und Datenindexbereiche in neue Datei übernehmen
						deletedContainerFile.addDeletedContainer(handle);
						contsToDelete.add(containerId);
					});
				}
			}
		}
		deletedContainerFile.write();
		for(Long contId : contsToDelete) {
			Path contFile = getPath(idContainerFileDir).resolve(getContFileName(contId));
			try {
				Files.delete(contFile);
			}
			catch(IOException e) {
				_debug.warning("Kann Containerdatei nicht löschen: " + contFile, e);
			}
		}
	}

	/**
	 * Verlängert die Löschzeit eines Containers in Folge einer Archivanfrage oder ArchivInfoAnfrage.
	 *
	 * @param lock  Objekt, das den synchronisierten Zugriff auf den Index erlaubt
	 * @param adk Archivdatenart
	 * @param contID ContainerID
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void extendLoeschschutz(final SyncKey<IdDataIdentification> lock, ArchiveDataKind adk, long contID) throws PersistenceException, IndexException {
		ContainerHeaders handle = getContainerHeaders(lock, adk, contID);
		if(handle.getContainerHeaderParamAsInt(ContainerManagementInformation.CHP_ANZ_DS) >= 0
				&& !handle.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED)) {
			// Container ist geschlossen und noch nicht gelöscht

			extendLoeschschutz(lock, handle, RestoreTask.getLOESSCHUTZ_VERL() * 1000);

		}
	}

	/**
	 * Verlängert die Löschzeit eines Containers.
	 *
	 * @param lock  Objekt, das den synchronisierten Zugriff auf den Index erlaubt
	 * @param handle     Container-Header
	 * @param t_lsMillis Neuer Löschschutz ab aktuellem Zeitpunkt (in Millisekunden)
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public void extendLoeschschutz(final SyncKey<IdDataIdentification> lock, final ContainerHeaders handle, long t_lsMillis) throws PersistenceException, IndexException {
		long now = System.currentTimeMillis();

		long deletionTime = getProtectedDeletionTime(handle);
		
		long newDeletionTime = now + Math.min(t_lsMillis, _tLSMax * 1000); // insgesamt um max. tLSMax Löschzeitverlängern
		
		if(newDeletionTime > deletionTime) handle.setContainerHeaderParam(ContainerManagementInformation.CHP_LOESCHUTZ, newDeletionTime);

		updateContainerIndex(lock, handle);
	}

	
	/**
	 * Gibt den Zeitpunkt zurück, an dem der Löschschutz abläuft.
	 *
	 * @param headers von Typ ContainerHeaders  
	 * @return long (Millisekunden-Zeitstempel)
	 */
	public long getProtectedDeletionTime(final ContainerHeaders headers) {
		return headers.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_LOESCHUTZ);
	}

	/**
	 * Gibt den Zeitpunkt zurück, an dem der Container regulär gelöscht werden kann
	 *
	 * @param headers von Typ ContainerHeaders  
	 * @return long (Millisekunden-Zeitstempel)
	 */
	public long getRegularDeletionTime(final ContainerHeaders headers) {

		if(_useDynamicDeletionTime) {
			long keepDuration = arSParamGetVorhalten(headers.getLocation().getDataIdentification());
			long arsTime = headers.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_ARC_TIME_MAX);
			
			return arsTime + keepDuration;
		}

		return headers.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_LOESCHEN);
	}

	private class IndexCloserWorker implements Runnable {
		private final AtomicLong _run;
		private final IdDataIdentification _dataIdentification;
		
		/** Array mit OpenContainerData je Datenart. Der Index entspricht {@link ArchiveDataKind#getCode()} {@code - 1} */
		private final StandardOpenContainerData[] _data;

		/**
		 * Erzeugt eine neue IndexCloserWorker-Instanz.
		 *  @param run von Typ AtomicLong  
		 * @param dataIdentification von Typ IdContainerFileDir  
		 * @param data von Typ StandardOpenContainerData  
		 */
		public IndexCloserWorker(final AtomicLong run, final IdDataIdentification dataIdentification, @NotNull final StandardOpenContainerData[] data) {
			_run = run;
			_dataIdentification = dataIdentification;
			_data = data;
		}

		/**
		 * Die Methode run schließt Indexe
		 */
		@Override
		public void run() {
			try {
				_run.incrementAndGet();
				for (int i = 0; i < _data.length; i++) {
					ArchiveDataKind adk = ArchiveDataKind.getInstance(i + 1);
					StandardOpenContainerData openContainerData = _data[i];
					if (openContainerData != null) {
						closeIndexHelper(openContainerData, adk);
					}
				}
			}
			catch(IndexException | SynchronizationFailedException e){
				_debug.error("Herunterfahren: Fehler beim Schliessen der Indexe: ", e);
			}
		}

		/**
		 * Die Methode closeIndexHelper schließt einen Index
		 *
		 * @param openContainerData von Typ AtomicInteger  
		 * @param dataKind von Typ ArchiveDataKind  
		 * @throws IndexException wenn ein Problem beim Zugriff auf den Index auftritt
		 */
		private void closeIndexHelper(final StandardOpenContainerData openContainerData, final ArchiveDataKind dataKind) throws IndexException, SynchronizationFailedException {
			try(SyncKey<IdDataIdentification> lockIndex = lockIndex(_dataIdentification)) {
				closeIndex(lockIndex, openContainerData, dataKind);
			}
		}
	}

	private static class ConsistencyCheckWorker implements Runnable {
		private final File _curDir;
		private final DataIdentTree _dit;
		private final String _mode;
		private final AtomicBoolean _allValid;
		private final AtomicInteger _numVisited;

		/**
		 * Erzeugt eine neue ConsistencyCheckWorker-Instanz.
		 *
		 * @param curDir von Typ File  
		 * @param dit von Typ DataIdentTree  
		 * @param mode von Typ String  
		 * @param allValid von Typ AtomicBoolean  
		 * @param numVisited von Typ AtomicInteger  
		 */
		public ConsistencyCheckWorker(final File curDir, final DataIdentTree dit, final String mode, final AtomicBoolean allValid, final AtomicInteger numVisited) {
			_curDir = curDir;
			_dit = dit;
			_mode = mode;
			_allValid = allValid;
			_numVisited = numVisited;
		}

		/**
		 * Die Methode run führt die Konsistenzprüfung durch
		 */
		@Override
		public void run() {
			try {
				IdDataIdentification dataIdentification = getDataIdentificationFromPath(_curDir);
				DataIdentNode din = new DataIdentNode(dataIdentification);
				_dit.putDataIdentNode(din);
				File[] content = _curDir.listFiles();
				if(content == null) {
					return;
				}
				for(File dir : content)         // Ordner der Datensatzarten (oa, on, na, nn) prüfen:
				{
					if(!isValidDataKindSuffix(dir.getName())) {
						_debug.fine("Fremde Datei/Verzeichnis im Persistenzverzeichnis: '" + dir.getAbsolutePath() + "'");
						_allValid.set(false);
						continue;
					}
					int retCode;
					if(_mode.equalsIgnoreCase(ArchiveManager.P_CHECK_PERS_DIR_HDR)) {
						retCode = ConsistencyCheck.simpleCheck(dir, _debug);
					}
					else if(_mode.equalsIgnoreCase(ArchiveManager.P_CHECK_PERS_DIR_ALL)) {
						retCode = ConsistencyCheck.extendedCheck(dir, _debug);
					}
					else {
						continue;
					}
					if((retCode & ConsistencyCheck.WARN) == ConsistencyCheck.WARN) {
						_debug.fine("Es sind Warnungen im Verzeichnis: '" + dir.getAbsolutePath() + "' aufgetreten.");
					}
					if((retCode & ConsistencyCheck.ERROR) == ConsistencyCheck.ERROR) {
						_debug.fine("Es sind Fehler im Verzeichnis: '" + dir.getAbsolutePath() + "' aufgetreten.");
						_allValid.set(false);
					}
				}
			}
			catch(PersistenceException e) {
				_debug.fine("Pfad der Datenidentifikation ungueltig: '" + _curDir.getAbsolutePath() + "'");
				_allValid.set(false);
			}
			finally {
				_numVisited.incrementAndGet();
			}
		}
	}

	/**
	 * Gibt archiveManager von diesem PersistenceManager-Objekt zurück.
	 *
	 *
	 *
	 * @return archiveManager (Typ ArchiveManager)
	 */
	public ArchiveManager getArchiveManager() {
		return archivMgr;
	}
	
	public class Statistics {

		private final long _numNodes = _dataIdentTree.size();
		private final long _numParameterizedNodes = _dataIdentTree.numParameterized();
		private final long _indexLocks = PersistenceManager.this.getIndexLocks().size();
		private final long _numOpenContainerData = _openContainerData.size();
		private final CacheManager.CacheMemoryUsage _cachedMemory = CacheManager.getInstance().getCachedMemory();
		
		/**
		 * Erzeugt eine neue Statistics-Instanz.
		 */
		private Statistics() {
		}

		/**
		 * Gibt numNodes von diesem Statistics-Objekt zurück.
		 *
		 * @return numNodes (Typ long)
		 */
		public long getNumNodes() {
			return _numNodes;
		}

		/**
		 * Gibt numParameterizedNodes von diesem Statistics-Objekt zurück.
		 *
		 * @return numParameterizedNodes (Typ long)
		 */
		public long getNumParameterizedNodes() {
			return _numParameterizedNodes;
		}

		/**
		 * Gibt nodeMemory von diesem Statistics-Objekt zurück.
		 *
		 * @return nodeMemory (Typ long)
		 */
		public long getNodeMemory() {
			return 48 * _numNodes;
		}
		
		/**
		 * Gibt indexLocks von diesem Statistics-Objekt zurück.
		 *
		 * @return indexLocks (Typ long)
		 */
		public long getIndexLocks() {
			return _indexLocks;
		}

		/**
		 * Gibt lockMemory von diesem Statistics-Objekt zurück.
		 *
		 * @return lockMemory (Typ long)
		 */
		public long getLockMemory() {
			return getIndexLocks() * 90; // 90 = ungefährer Speicherverbrauch eines SyncKey mit Referenz usw.
		}

		/**
		 * Gibt numOpenContainerData von diesem Statistics-Objekt zurück.
		 *
		 * @return numOpenContainerData (Typ long)
		 */
		public long getNumOpenContainerData() {
			return _numOpenContainerData;
		}

		/**
		 * Gibt openContainerMemory von diesem Statistics-Objekt zurück.
		 *
		 * @return openContainerMemory (Typ long)
		 */
		public long getOpenContainerMemory() {
			return _numOpenContainerData * 80; // 80 = ungefährer Speicherverbrauch eines OpenContainerData mit Referenz usw.
		}

		/**
		 * Gibt cachedMemory von diesem Statistics-Objekt zurück.
		 *
		 * @return cachedMemory
		 */
		public CacheManager.CacheMemoryUsage getCachedMemory() {
			return _cachedMemory;
		}

		@Override
		public String toString() {
			return "Statistics{" +
					"_numNodes=" + _numNodes +
					", _numParameterizedNodes=" + _numParameterizedNodes +
					", _indexLocks=" + _indexLocks +
					", _openContainerData=" + _numOpenContainerData +
					", _cachedMemory=" + _cachedMemory +
					'}';
		}
	}

	public static class RebuildResult {
		
		public long containerRead;
		public long containerIgnored;
		public long indexDeleted;
		public long indexCreated;

		public RebuildResult() {
		}

		public RebuildResult(final long containerRead, final long containerIgnored, final long indexDeleted, final long indexCreated) {
			this.containerRead = containerRead;
			this.containerIgnored = containerIgnored;
			this.indexDeleted = indexDeleted;
			this.indexCreated = indexCreated;
		}

		/**
		 * Gibt die Zahl der bei der Wiederherstellung gelesenen Containerdateien zurück. 
		 * @return Anzahl Dateien
		 */
		public long getContainerRead() {
			return containerRead;
		}

		/**
		 * Gibt die Zahl der bei der Wiederherstellung ignorierten Containerdateien zurück. 
		 * @return Anzahl Dateien
		 */
		public long getContainerIgnored() {
			return containerIgnored;
		}

		/**
		 * Gibt die Zahl der bei der Wiederherstellung gelöschten Indexdateien zurück. 
		 * @return Anzahl Dateien
		 */
		public long getIndexDeleted() {
			return indexDeleted;
		}

		/**
		 * Gibt die Zahl der bei der Wiederherstellung erstellten Indexdateien zurück. 
		 * @return Anzahl Dateien
		 */
		public long getIndexCreated() {
			return indexCreated;
		}
	}
}
