/*
 * Copyright 2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.util;

import static java.nio.file.FileVisitResult.CONTINUE;


import java.io.IOException;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.StandardCopyOption;
import java.nio.file.attribute.BasicFileAttributes;

/**
 * Hilfsklasse um einen Ordner auf einen Zuielordner zu kopieren (zu synchronisieren) ohne alles erneut zu übertragen.
 *
 * @author Kappich Systemberatung
 */
public class FileSync {

	/**
	 * Synchronisiert zwei Verzeichnisse.
	 * @param from Quellverzeichnis
	 * @param to   Zielverzeichnis
	 * @throws IOException IO-Problem
	 * @return Statistik über kopierte Daten
	 */
	public SyncResult syncDirectories(Path from, Path to) throws IOException {
		SyncVisitor syncVisitor = new SyncVisitor(from, to);
		RemoveVisitor removeVisitor = new RemoveVisitor(from, to);
		Files.walkFileTree(from, syncVisitor);
		Files.walkFileTree(to, removeVisitor);
		return new SyncResult(syncVisitor.getFilesCopied(), removeVisitor.getFilesRemoved(), syncVisitor.getBytesCopied());
	}

	protected boolean shouldIgnore(final String fileName) {
		return false;
	}

	public static boolean isIdentical(final Path sourceFile, final Path targetFile) throws IOException {
		if(!Files.exists(targetFile)) {
			return false;
		}
		try {
			if(Files.isDirectory(targetFile)) return true;
			if(Files.size(sourceFile) != Files.size(targetFile)) return false;
			return Files.getLastModifiedTime(sourceFile).equals(Files.getLastModifiedTime(targetFile));
		}
		catch(NoSuchFileException e) {
			return false;
		}
	}

	private class SyncVisitor extends SimpleFileVisitor<Path> {

		private final Path _source;
		private final Path _target;
		private long _filesCopied;
		private long _bytesCopied;

		public SyncVisitor(final Path source, final Path target) {
			_source = source;
			_target = target;
		}

		@Override
		public FileVisitResult preVisitDirectory(final Path dir, final BasicFileAttributes attrs) throws IOException {
			Path targetDir = _target.resolve(_source.relativize(dir));
			try {
				if(!Files.isDirectory(targetDir)) {
					Path parent = targetDir.getParent();
					if(!Files.isDirectory(parent)) {
						Files.createDirectories(parent);
					}
					Files.copy(dir, targetDir, StandardCopyOption.COPY_ATTRIBUTES);
					_filesCopied++;
				}
			} catch (FileAlreadyExistsException e) {
				if (!Files.isDirectory(targetDir))
					throw e;
			}
			return CONTINUE;		
		}

		@Override
		public FileVisitResult visitFile(final Path file, final BasicFileAttributes attrs) throws IOException {
			Path targetFile = _target.resolve(_source.relativize(file));
			if(shouldIgnore(file.getFileName().toString())) {
				return CONTINUE;
			}
			if(!isIdentical(file, targetFile)) {
				Files.copy(file, targetFile, StandardCopyOption.COPY_ATTRIBUTES, StandardCopyOption.REPLACE_EXISTING);
				if(!isIdentical(file, targetFile)) {
					Files.setLastModifiedTime(targetFile, Files.getLastModifiedTime(file));
					if(!isIdentical(file, targetFile)) {
						Files.setLastModifiedTime(file, Files.getLastModifiedTime(targetFile));
						assert isIdentical(file, targetFile);
					}
				}
				_filesCopied++;
				_bytesCopied += Files.size(targetFile);
			}
			return CONTINUE;
		}

		@Override
		public FileVisitResult visitFileFailed(final Path file, final IOException exc) {
			return CONTINUE;
		}

		public long getFilesCopied() {
			return _filesCopied;
		}

		public long getBytesCopied() {
			return _bytesCopied;
		}
	}
	
	private class RemoveVisitor extends SimpleFileVisitor<Path> {

		private final Path _source;
		private final Path _target;
		private long _filesRemoved;

		public RemoveVisitor(final Path source, final Path target) {
			_source = source;
			_target = target;
		}

		@Override
		public FileVisitResult postVisitDirectory(final Path dir, final IOException exc) throws IOException {
			super.postVisitDirectory(dir, exc);
			Path sourceDirectory = _source.resolve(_target.relativize(dir));
			if(!Files.exists(sourceDirectory)) {
				Files.delete(dir);
				_filesRemoved++;
			}
			return CONTINUE;
		}

		@Override
		public FileVisitResult visitFile(final Path file, final BasicFileAttributes attrs) throws IOException {
			Path sourceFile = _source.resolve(_target.relativize(file));
			if(!shouldIgnore(file.getFileName().toString()) && !Files.exists(sourceFile)) {
				Files.delete(file);
				_filesRemoved++;
			}
			return CONTINUE;
		}


		@Override
		public FileVisitResult visitFileFailed(final Path file, final IOException exc) {
			return CONTINUE;
		}

		public long getFilesRemoved() {
			return _filesRemoved;
		}
	}
}
