/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.walk.internal;

import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Iterator;

/**
 * Iterator über alle Containerverzeichnisse im Archivsystem
 *
 * @author Kappich Systemberatung
 */
public class AllContainerDirs implements DataIdentificationDirGetter {

	private final Deque<DirectoryNode> _stack = new ArrayDeque<>();

	/**
	 * Erstellt eine neue AllContainerDirs-Instanz
	 * @param persistenceManager Persistenzmanager
	 */
	public AllContainerDirs(final PersistenceManager persistenceManager) {
		this(persistenceManager.getArchivePath());
	}

	/**
	 * Erstellt eine neue AllContainerDirs-Instanz
	 * @param rootDir Wurzelverzeichnis
	 */
	public AllContainerDirs(final Path rootDir) {
		try {
			_stack.addLast(new DirectoryNode(rootDir));
		}
		catch(IOException e){
			throw new IllegalArgumentException("Ungültiges Wurzelverzeichnis: " + rootDir, e);
		}
	}

	@Override
	public DataIdentificationDir poll() throws PersistenceException {
		DirectoryNode node = _stack.peekLast();
		while(true) {
			try {
				Path path = node.next();
				while(path == null) {
					_stack.pollLast();
					node = _stack.peekLast();
					if(node == null) {
						// Stack leer, alles abgearbeitet
						return null;
					}
					path = node.next();
				}
				String folderName = path.getFileName().toString();
				if(PersistenceManager.isValidSvFolderName(folderName)) {
					// Es handelt sich um ein Containerdatenverzeichnis
					return new DataIdentificationDir(path);
				}
				else if(PersistenceManager.isValidIDFolderName(folderName)) {
					// Es handelt sich um ein Verzeichnis, das zum Archiv dazugehört
					node = new DirectoryNode(path);
					_stack.addLast(node);
				}
			}
			catch(IOException e) {
				throw new PersistenceException(e);
			}

		}
	}

	private static class DirectoryNode {
		private final DirectoryStream<Path> _stream;

		private final Iterator<Path> _iterator;

		public DirectoryNode(final Path path) throws IOException {
			_stream = Files.newDirectoryStream(path, it -> Files.isDirectory(it) && !Files.isHidden(it));
			_iterator = _stream.iterator();
		}

		@Nullable
		public Path next() throws IOException {
			if(!_iterator.hasNext()) {
				_stream.close();
				return null;
			}
			return _iterator.next();
		}
	}
}
