/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.walk.internal;

import com.google.common.collect.ImmutableList;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirAction;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirCollector;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirWalk;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import java.time.Duration;
import java.time.Instant;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Stream;

/**
 * Standardimplementierung eines ContainerDirWalk
 *
 * @author Kappich Systemberatung
 */
public class StandardContainerDirWalk implements DataIdentificationDirWalk {

	private static final Debug _debug = Debug.getLogger();

	/**
	 * Klasse, die Containerdatei_Verzeichnisse auflistet
	 */
	private final DataIdentificationDirGetter _contGetter;

	/**
	 * Exception, die während des Durchlaufsaufgetreten ist (oder null)
	 */
	private Exception _exception;

	/**
	 * Thread-Pool
	 */
	private ThreadPoolExecutor _executor;

	/**
	 * Anzahl besuchter Container (nur wenn innerhalb eines {@link de.bsvrz.ars.ars.persistence.walk.ContainerWalk} benutzt, sonst 0)
	 */
	private final AtomicLong _visitedContainerFiles = new AtomicLong();

	/**
	 * Anzahl besuchter Containerverz. (threadsave)
	 */
	private final AtomicLong _visitedContainerDirs = new AtomicLong();

	/**
	 * Anzahl gefundener Containerverz. (threadsave)
	 */
	private final AtomicLong _indexedContainerDirs = new AtomicLong();

	/**
	 * Intervall für periodische Debug-Ausgaben
	 */
	private Duration _statusInterval = Duration.ofMinutes(1);

	/**
	 * Letzter Ausgabezeitpunkt
	 */
	private Instant _lastStatus;

	/**
	 * Klasse, die Ausgaben formatiert
	 */
	private StatusPrinter _statusPrinter = new StatusPrinter();

	/**
	 * Startzeitpunkt
	 */
	private Instant _start;

	/**
	 * Soll die Ausführung sofort beendet werden (Aufträge in Warteschlange werden nicht mehr durchgeführt)
	 */
	private volatile boolean _terminateImmediately;
	private LongTermTaskStatePublisher.Task _publisherTask;

	/**
	 * Erstellt einen neuen StandardContainerDirWalk
	 * @param contGetter Klasse, die Cotnainerverzeichnisse auflistet.
	 */
	public StandardContainerDirWalk(final DataIdentificationDirGetter contGetter) {
		_contGetter = contGetter;
	}

	@Override
	public Duration getStatusInterval() {
		return _statusInterval;
	}

	@Override
	public void setStatusInterval(final Duration statusInterval) {
		_statusInterval = statusInterval;
	}

	@Override
	public StatusPrinter getStatusPrinter() {
		return _statusPrinter;
	}

	@Override
	public void setStatusPrinter(final StatusPrinter statusPrinter) {
		_statusPrinter = statusPrinter;
	}

	@Override
	public LongTermTaskStatePublisher.Task getPublisherTask() {
		return _publisherTask;
	}

	@Override
	public void setPublisherTask(final LongTermTaskStatePublisher.Task publisherTask) {
		_publisherTask = publisherTask;
	}

	@Override
	public void execute(final String actionName, final int numThreads, final DataIdentificationDirAction action) throws PersistenceException {
		if(_executor != null) {
			throw new IllegalStateException("Bereits gestartet");
		}
		long id = Thread.currentThread().getId();

		// threadpool für die eigentliche Ausführung
		_executor = new ThreadPoolExecutor(numThreads, numThreads, 0, TimeUnit.SECONDS, new ArrayBlockingQueue<>(1024*1024), new ThreadFactory() {
			private final AtomicInteger _counter = new AtomicInteger(0);

			@Override
			public Thread newThread(@NotNull final Runnable r) {
				return new Thread(r, "ContainerDirWalk-" + id + "-" + actionName + "-" + _counter.getAndIncrement());
			}
		});

		_debug.info("Starte " + actionName);
		_lastStatus = Instant.now();
		_start = _lastStatus;

		// Einzelner Thread, der die Containerverzeichnisse indiziert
		final Runnable indexRunner = () -> {
			try {
				while(!_terminateImmediately) {
					DataIdentificationDir poll = _contGetter.poll();
					if(poll == null) return;

					maybePrintStatus(actionName, false);

					_indexedContainerDirs.incrementAndGet();
					_executor.execute(() -> {
						if(_terminateImmediately) return;
						try {
							action.run(poll, StandardContainerDirWalk.this);
							_visitedContainerDirs.incrementAndGet();
						}
						catch(Exception e) {
							_debug.warning("Fehler beim Ausführen der Aktion", e);
							setException(e);
							_executor.shutdown();
						}
					});
				}
			}
			catch(PersistenceException e) {
				_debug.warning("Fehler beim Auflisten der Verzeichnisse", e);
				setException(e);
			}
			catch(RejectedExecutionException ignored) {
			}

		};
		Thread indexThread = new Thread(indexRunner, "ContainerDirWalk-" + id + "-" + actionName + "-index");
		try {
			indexThread.start();
			indexThread.join();
			_executor.shutdown();
			_debug.fine("ContainerDirWalk-" + id + "-" + actionName + ": Verzeichnisdurchlauf beendet.");
			while(!_executor.awaitTermination(1, TimeUnit.SECONDS)) {
				_debug.fine("ContainerDirWalk-" + id + "-" + actionName + ": Warte auf Beenden des Threadpools...");
				maybePrintStatus(actionName, true);
			}
			_debug.info(_statusPrinter.getSuccessMessage(actionName, Duration.between(_start, Instant.now()), _visitedContainerDirs.get(), _visitedContainerFiles.get())
			);
		}
		catch(InterruptedException e) {
			throw new PersistenceException(e);
		}
		if(_exception != null) throw new PersistenceException(_exception);
	}

	@Override
	public <E> Stream<E> collect(final String actionName, final int numThreads, final DataIdentificationDirCollector<E> collector) throws PersistenceException {

		ConcurrentLinkedQueue<ImmutableList.Builder<E>> builders = new ConcurrentLinkedQueue<>();

		ThreadLocal<ImmutableList.Builder<E>> threadLocalBuilder = ThreadLocal.withInitial(() -> {
			ImmutableList.Builder<E> myBuilder = new ImmutableList.Builder<>();
			builders.add(myBuilder);
			return myBuilder;
		});

		execute(actionName, numThreads, (containerFileDir, standardContainerDirWalk) -> collector.run(containerFileDir, threadLocalBuilder.get()));

		return builders.stream().flatMap(it -> it.build().stream());
	}

	private void maybePrintStatus(final String actionName, final boolean indexFinished) {
		Instant now = Instant.now();
		if(_lastStatus.plus(_statusInterval).isBefore(now)) {
			_debug.info(_statusPrinter.getStatusMessage(actionName, Duration.between(_start, now), indexFinished, _indexedContainerDirs.get(), _visitedContainerDirs.get(), _visitedContainerFiles.get()));
		
			_lastStatus = now;
		}

		if(_publisherTask != null) {
			_publisherTask.setTaskProgress(actionName + ": %1$d von %2$d Datenidentifikationen bearbeitet.", _visitedContainerDirs.get(), _indexedContainerDirs.get(), -1);
		}
	}

	@Override
	public void terminate() {
		if(_executor != null) {
			_executor.shutdown();
		}
		_terminateImmediately = true;
	}

	/**
	 * Setzt die Exception, die beim Auftreten des Auftrags aufgetreten ist. Dadurch wird der Vorgang beendet.
	 *
	 * Die Exception wird nachher im Haupt-Thread geworfen.
	 *
	 * @param exception Exception
	 */
	private void setException(final Exception exception) {
		_terminateImmediately = true;
		if(_exception == null) {
			_exception = exception;
		}
		else {
			_exception.addSuppressed(exception);
		}
	}

	@Override
	public void increaseContCount(final int increment) {
		_visitedContainerFiles.addAndGet(increment);
	}
}
